(* camlp5r *)
(* comp_head.ml.tpl,v *)
(* Copyright (c) INRIA 2007-2017 *)

#load "q_MLast.cmo";
#load "pa_extend.cmo";

module P =
  struct
    value gloc bp strm = Grammar.loc_of_token_interval bp (Stream.count strm);
    value vala symb = parser [: a = symb :] -> <:vala< a >>;
    value list0 symb =
      let rec loop al =
        parser
        [ [: a = symb; a = loop [a :: al] ! :] -> a
        | [: :] -> al ]
      in
      parser [: a = loop [] :] -> List.rev a
    ;
    value list0sep symb sep b =
      if not b then
        let rec kont al =
          parser
          [ [: v = sep; a = symb; a = kont [a :: al] ! :] -> a
          | [: :] -> al ]
        in
        parser
        [ [: a = symb; a = kont [a] ! :] -> List.rev a
        | [: :] -> [] ]
      else failwith "LIST0 _ SEP _ OPT_SEP not yet implemented"
    ;
    value list1 symb =
      let rec loop al =
        parser
        [ [: a = symb; a = loop [a :: al] ! :] -> a
        | [: :] -> al ]
      in
      parser [: a = symb; a = loop [a] ! :] -> List.rev a
    ;
    value list1sep symb sep b =
      if not b then
        let rec kont al =
          parser
          [ [: v = sep; a = symb; a = kont [a :: al] ! :] -> a
          | [: :] -> al ]
        in
        parser [: a = symb; a = kont [a] ! :] -> List.rev a
      else
        let rec kont al =
          parser
          [ [: v = sep; a = symb; a = kont [a :: al] ! :] -> a
          | [: v = sep :] -> al
          | [: :] -> al ]
        in
        parser [: a = symb; a = kont [a] ! :] -> List.rev a
    ;
    value option f =
      parser
      [ [: a = f :] -> Some a
      | [: :] -> None ]
    ;
    value bool f =
      parser
      [ [: _ = f :] -> True
      | [: :] -> False ]
    ;
    value token (p_con, p_prm) =
      if p_prm = "" then parser [: `(con, prm) when con = p_con :] -> prm
      else parser [: `(con, prm) when con = p_con && prm = p_prm :] -> prm
    ;
    value orzero f f0 =
      parser
      [ [: a = f :] -> a
      | [: a = f0 :] -> a ]
    ;
    value error entry prev_symb symb =
      symb ^ " expected" ^
      (if prev_symb = "" then "" else " after " ^ prev_symb) ^ " (in [" ^
      entry ^ "])"
    ;
    value lexer = lazy (Plexer.gmake ());
  end
;

(****************************************)

(* camlp5r *)
(* pa_o.ml,v *)
(* Copyright (c) INRIA 2007-2017 *)


open Asttools;
open Pcaml;
open Mlsyntax.Original;

Pcaml.syntax_name.val := "OCaml";
Pcaml.no_constructors_arity.val := True;

do {
  let odfa = Plexer.dollar_for_antiquotation.val in
  let osrs = Plexer.simplest_raw_strings.val in
  Plexer.dollar_for_antiquotation.val := False;
  Plexer.simplest_raw_strings.val := True;
  Plexer.utf8_lexing.val := True;
  Grammar.Unsafe.gram_reinit gram (Lazy.force P.lexer);
  Plexer.dollar_for_antiquotation.val := odfa;
  Plexer.simplest_raw_strings.val := osrs;
  Grammar.Unsafe.clear_entry attribute_body;
  Grammar.Unsafe.clear_entry interf;
  Grammar.Unsafe.clear_entry implem;
  Grammar.Unsafe.clear_entry top_phrase;
  Grammar.Unsafe.clear_entry use_file;
  Grammar.Unsafe.clear_entry module_type;
  Grammar.Unsafe.clear_entry module_expr;
  Grammar.Unsafe.clear_entry longident;
  Grammar.Unsafe.clear_entry extended_longident;
  Grammar.Unsafe.clear_entry sig_item;
  Grammar.Unsafe.clear_entry str_item;
  Grammar.Unsafe.clear_entry signature;
  Grammar.Unsafe.clear_entry structure;
  Grammar.Unsafe.clear_entry expr;
  Grammar.Unsafe.clear_entry patt;
  Grammar.Unsafe.clear_entry ctyp;
  Grammar.Unsafe.clear_entry let_binding;
  Grammar.Unsafe.clear_entry type_decl;
  Grammar.Unsafe.clear_entry type_extension;
  Grammar.Unsafe.clear_entry extension_constructor;
  Grammar.Unsafe.clear_entry constructor_declaration;
  Grammar.Unsafe.clear_entry label_declaration;
  Grammar.Unsafe.clear_entry match_case;
  Grammar.Unsafe.clear_entry with_constr;
  Grammar.Unsafe.clear_entry poly_variant;
  Grammar.Unsafe.clear_entry class_type;
  Grammar.Unsafe.clear_entry class_expr;
  Grammar.Unsafe.clear_entry class_expr_simple;
  Grammar.Unsafe.clear_entry alg_attribute;
  Grammar.Unsafe.clear_entry alg_attributes;
  Grammar.Unsafe.clear_entry ext_attributes;
  Grammar.Unsafe.clear_entry class_sig_item;
  Grammar.Unsafe.clear_entry class_str_item
};

Pcaml.parse_interf.val := Grammar.Entry.parse interf;
Pcaml.parse_implem.val := Grammar.Entry.parse implem;

value error loc msg = Ploc.raise loc (Failure msg);

value mklistexp loc last =
  loop True where rec loop top =
    fun
    [ [] ->
        match last with
        [ Some e -> e
        | None -> MLast.ExLong loc (MLast.LiUid loc (Ploc.VaVal "[]")) ]
    | [e1 :: el] ->
        let loc = if top then loc else Ploc.encl (MLast.loc_of_expr e1) loc in
        MLast.ExApp loc
          (MLast.ExApp loc
             (MLast.ExLong loc (MLast.LiUid loc (Ploc.VaVal "::"))) e1)
          (loop False el) ]
;

value mklistpat loc last =
  loop True where rec loop top =
    fun
    [ [] ->
        match last with
        [ Some p -> p
        | None ->
            MLast.PaLong loc (MLast.LiUid loc (Ploc.VaVal "[]"))
              (Ploc.VaVal []) ]
    | [p1 :: pl] ->
        let loc = if top then loc else Ploc.encl (MLast.loc_of_patt p1) loc in
        MLast.PaApp loc
          (MLast.PaApp loc
             (MLast.PaLong loc (MLast.LiUid loc (Ploc.VaVal "::"))
                (Ploc.VaVal []))
             p1)
          (loop False pl) ]
;

value operator_rparen_f strm = do {
  let id x = x in
  let app suff s = s ^ suff in
  let trials =
    [(1,
      Right
        (fun
         [ [("LIDENT" | "UIDENT", _) :: _] -> True
         | _ -> False ]));
     (2, Left (is_operator, id, [[("", ")")]]));
     (2, Left (is_letop, id, [[("", ")")]]));
     (2, Left (is_andop, id, [[("", ")")]]));
     (4, Left (is_dotop, app "()", [[("", "("); ("", ")"); ("", ")")]]));
     (4, Left (is_dotop, app "{}", [[("", "{"); ("", "}"); ("", ")")]]));
     (4, Left (is_dotop, app "[]", [[("", "["); ("", "]"); ("", ")")]]));
     (6,
      Left
        (is_dotop, app "(;..)",
         [[("", "("); ("", ";"); ("", ".."); ("", ")"); ("", ")")]]));
     (6,
      Left
        (is_dotop, app "{;..}",
         [[("", "{"); ("", ";"); ("", ".."); ("", "}"); ("", ")")]]));
     (6,
      Left
        (is_dotop, app "[;..]",
         [[("", "["); ("", ";"); ("", ".."); ("", "]"); ("", ")")]]));
     (5,
      Left
        (is_dotop, app "()<-",
         [[("", "("); ("", ")"); ("", "<-"); ("", ")")]]));
     (5,
      Left
        (is_dotop, app "{}<-",
         [[("", "{"); ("", "}"); ("", "<-"); ("", ")")]]));
     (5,
      Left
        (is_dotop, app "[]<-",
         [[("", "["); ("", "]"); ("", "<-"); ("", ")")]]));
     (7,
      Left
        (is_dotop, app "(;..)<-",
         [[("", "("); ("", ";"); ("", ".."); ("", ")"); ("", "<-");
           ("", ")")]]));
     (7,
      Left
        (is_dotop, app "{;..}<-",
         [[("", "{"); ("", ";"); ("", ".."); ("", "}"); ("", "<-");
           ("", ")")]]));
     (7,
      Left
        (is_dotop, app "[;..]<-",
         [[("", "["); ("", ";"); ("", ".."); ("", "]"); ("", "<-");
           ("", ")")]]))]
  in
  let matchers =
    List.map
      (fun
       [ (n, Left (pred, xform, suffixes)) ->
           (n,
            Left
              (fun
               [ [("", s) :: l] when pred s && List.mem l suffixes ->
                   Some (xform s)
               | _ -> None ]))
       | (n, Right f) -> (n, Right f) ])
      trials
  in
  let (n, tok) = check_stream matchers strm in
  for i = 1 to n do { Stream.junk strm };
  tok
};

value operator_rparen =
   operator_rparen_f
;

value check_not_part_of_patt_f strm =
  let matchers =
    [(2,
      fun
      [ [("LIDENT", _); tok :: _] -> Some tok
      | _ -> None ]);
     (4,
      fun
      [ [("", "("); ("", s); ("", ")"); tok :: _] when is_special_op s ->
          Some tok
      | _ -> None ]);
     (6,
      fun
      [ [("", "("); ("", s); ("", "("); ("", ")"); ("", ")"); tok :: _]
        when is_special_op s ->
          Some tok
      | [("", "("); ("", s); ("", "{"); ("", "}"); ("", ")"); tok :: _]
        when is_special_op s ->
          Some tok
      | [("", "("); ("", s); ("", "["); ("", "]"); ("", ")"); tok :: _]
        when is_special_op s ->
          Some tok
      | _ -> None ]);
     (7,
      fun
      [ [("", "("); ("", s); ("", "("); ("", ")"); ("", "<-"); ("", ")");
         tok :: _]
        when is_special_op s ->
          Some tok
      | [("", "("); ("", s); ("", "{"); ("", "}"); ("", "<-"); ("", ")");
         tok :: _]
        when is_special_op s ->
          Some tok
      | [("", "("); ("", s); ("", "["); ("", "]"); ("", "<-"); ("", ")");
         tok :: _]
        when is_special_op s ->
          Some tok
      | _ -> None ]);
     (8,
      fun
      [ [("", "("); ("", s); ("", "("); ("", ";"); ("", ".."); ("", ")");
         ("", ")"); tok :: _]
        when is_special_op s ->
          Some tok
      | [("", "("); ("", s); ("", "{"); ("", ";"); ("", ".."); ("", "}");
         ("", ")"); tok :: _]
        when is_special_op s ->
          Some tok
      | [("", "("); ("", s); ("", "["); ("", ";"); ("", ".."); ("", "]");
         ("", ")"); tok :: _]
        when is_special_op s ->
          Some tok
      | _ -> None ]);
     (9,
      fun
      [ [("", "("); ("", s); ("", "("); ("", ";"); ("", ".."); ("", ")");
         ("", "<-"); ("", ")"); tok :: _]
        when is_special_op s ->
          Some tok
      | [("", "("); ("", s); ("", "{"); ("", ";"); ("", ".."); ("", "}");
         ("", "<-"); ("", ")"); tok :: _]
        when is_special_op s ->
          Some tok
      | [("", "("); ("", s); ("", "["); ("", ";"); ("", ".."); ("", "]");
         ("", "<-"); ("", ")"); tok :: _]
        when is_special_op s ->
          Some tok
      | _ -> None ])]
  in
  let rec crec i =
    fun
    [ [(n, _) :: _] as ml when i < n ->
        let l = stream_npeek i strm in
        let last = fst (sep_last l) in
        if last = ("EOI", "") || last = ("", ";;") then raise Stream.Failure
        else crec (i + 1) ml
    | [(n, f) :: t] ->
        match f (stream_npeek n strm) with
        [ None -> crec (i + 1) t
        | Some tok -> tok ]
    | [] -> raise Stream.Failure ]
  in
  let tok = crec 1 matchers in
  match tok with
  [ ("", "," | "as" | "|" | "::") -> raise Stream.Failure
  | _ -> () ]
;

value check_not_part_of_patt =
  
    check_not_part_of_patt_f
;

value prefixop =
  
    (fun (strm__ : Stream.t _) ->
       match Stream.peek strm__ with
       [ Some ("", x) when is_prefixop x -> do { Stream.junk strm__; x }
       | _ -> raise Stream.Failure ])
;

value infixop0 =
  
    (fun (strm__ : Stream.t _) ->
       match Stream.peek strm__ with
       [ Some ("", x) when is_infixop0 x -> do { Stream.junk strm__; x }
       | _ -> raise Stream.Failure ])
;

value infixop1 =
  
    (fun (strm__ : Stream.t _) ->
       match Stream.peek strm__ with
       [ Some ("", x) when is_infixop1 x -> do { Stream.junk strm__; x }
       | _ -> raise Stream.Failure ])
;

value infixop2 =
  
    (fun (strm__ : Stream.t _) ->
       match Stream.peek strm__ with
       [ Some ("", x) when is_infixop2 x -> do { Stream.junk strm__; x }
       | _ -> raise Stream.Failure ])
;

value infixop3 =
  
    (fun (strm__ : Stream.t _) ->
       match Stream.peek strm__ with
       [ Some ("", x) when is_infixop3 x -> do { Stream.junk strm__; x }
       | _ -> raise Stream.Failure ])
;

value infixop4 =
  
    (fun (strm__ : Stream.t _) ->
       match Stream.peek strm__ with
       [ Some ("", x) when is_infixop4 x -> do { Stream.junk strm__; x }
       | _ -> raise Stream.Failure ])
;

value hashop =
  
    (fun (strm__ : Stream.t _) ->
       match Stream.peek strm__ with
       [ Some ("", x) when is_hashop x -> do { Stream.junk strm__; x }
       | _ -> raise Stream.Failure ])
;

value letop =
  
    (fun (strm__ : Stream.t _) ->
       match Stream.peek strm__ with
       [ Some ("", x) when is_letop x -> do { Stream.junk strm__; x }
       | _ -> raise Stream.Failure ])
;

value andop =
  
    (fun (strm__ : Stream.t _) ->
       match Stream.peek strm__ with
       [ Some ("", x) when is_andop x -> do { Stream.junk strm__; x }
       | _ -> raise Stream.Failure ])
;

value dotop =
  
    (fun (strm__ : Stream.t _) ->
       match Stream.peek strm__ with
       [ Some ("", x) when is_dotop x -> do { Stream.junk strm__; x }
       | _ -> raise Stream.Failure ])
;

value test_constr_decl_f strm =
  match Stream.npeek 1 strm with
  [ [("UIDENT", _)] ->
      match Stream.npeek 2 strm with
      [ [_; ("", ".")] -> raise Stream.Failure
      | [_; ("", "(")] -> raise Stream.Failure
      | [_ :: _] -> ()
      | _ -> raise Stream.Failure ]
  | [("", "true" | "false") :: _] -> ()
  | [("", "|")] -> ()
  | [("", "[")] ->
      match Stream.npeek 2 strm with
      [ [_; ("", "]")] -> ()
      | _ -> raise Stream.Failure ]
  | _ -> raise Stream.Failure ]
;

value test_constr_decl =
   test_constr_decl_f
;

value stream_peek_nth n (strm : Stream.t (string * string)) =
  loop n (Stream.npeek n strm) where rec loop n =
    fun
    [ [] -> None
    | [x] -> if n == 1 then Some (x : (string * string)) else None
    | [_ :: l] -> loop (n - 1) l ]
;

(* horrible hack to be able to parse class_types *)

value test_ctyp_minusgreater =
  
    (fun strm ->
       let rec skip_simple_ctyp n =
         match stream_peek_nth n strm with
         [ Some ("", "->") -> n
         | Some ("", "[" | "[<") ->
             skip_simple_ctyp (ignore_upto "]" (n + 1) + 1)
         | Some ("", "(") -> skip_simple_ctyp (ignore_upto ")" (n + 1) + 1)
         | Some
             ("",
              "as" | "'" | ":" | "*" | "." | "#" | "<" | ">" | ".." | ";" |
              "_") ->
             skip_simple_ctyp (n + 1)
         | Some ("QUESTIONIDENT" | "LIDENT" | "UIDENT", _) ->
             skip_simple_ctyp (n + 1)
         | Some _ | None -> raise Stream.Failure ]
       and ignore_upto end_kwd n =
         match stream_peek_nth n strm with
         [ Some ("", prm) when prm = end_kwd -> n
         | Some ("", "[" | "[<") ->
             ignore_upto end_kwd (ignore_upto "]" (n + 1) + 1)
         | Some ("", "(") -> ignore_upto end_kwd (ignore_upto ")" (n + 1) + 1)
         | Some _ -> ignore_upto end_kwd (n + 1)
         | None -> raise Stream.Failure ]
       in
       match Stream.peek strm with
       [ Some (("", "[") | ("LIDENT" | "UIDENT", _)) -> skip_simple_ctyp 1
       | Some ("", "object") -> raise Stream.Failure
       | _ -> 1 ])
;

value is_lbracket_f strm =
  match Stream.npeek 1 strm with
  [ [("", "[")] -> True
  | _ -> False ]
;

value check_lbracket_f strm =
  if is_lbracket_f strm then () else raise Stream.Failure
;

value check_lbracket =
   check_lbracket_f
;

value is_lbracketbar_f strm =
  match Stream.npeek 1 strm with
  [ [("", "[|")] -> True
  | _ -> False ]
;

value check_lbracketbar_f strm =
  if is_lbracketbar_f strm then () else raise Stream.Failure
;

value check_lbracketbar =
   check_lbracketbar_f
;


value is_lbrace_f strm =
  match Stream.npeek 1 strm with
  [ [("", "{")] -> True
  | _ -> False ]
;

value check_lbrace_f strm =
  if is_lbrace_f strm then () else raise Stream.Failure
;

value check_lbrace =
   check_lbrace_f
;

value is_lident_colon_f strm =
  match Stream.npeek 2 strm with
  [ [("LIDENT", _); ("", ":") :: _] -> True
  | _ -> False ]
;

value check_lident_colon_f strm =
  if is_lident_colon_f strm then () else raise Stream.Failure
;

value check_lident_colon =
   check_lident_colon_f
;

value check_not_lident_colon_f strm =
  if not (is_lident_colon_f strm) then () else raise Stream.Failure
;

value check_not_lident_colon =
  
    check_not_lident_colon_f
;

value test_label_eq =
  
    (test 1 where rec test lev strm =
       match stream_peek_nth lev strm with
       [ Some (("UIDENT", _) | ("LIDENT", _) | ("", ".")) ->
           test (lev + 1) strm
       | Some ("ANTIQUOT_LOC", _) -> ()
       | Some ("", "=" | ";" | "}" | ":") -> ()
       | _ -> raise Stream.Failure ])
;

value test_typevar_list_dot =
  
    (let rec test lev strm =
       match stream_peek_nth lev strm with
       [ Some ("", "'") -> test2 (lev + 1) strm
       | Some ("", ".") -> ()
       | _ -> raise Stream.Failure ]
     and test2 lev strm =
       match stream_peek_nth lev strm with
       [ Some ("UIDENT" | "LIDENT", _) -> test (lev + 1) strm
       | _ -> raise Stream.Failure ]
     in
     test 1)
;

value e_phony =
  
    (fun (strm__ : Stream.t _) -> raise Stream.Failure)
;
value p_phony =
  
    (fun (strm__ : Stream.t _) -> raise Stream.Failure)
;

value constr_arity = ref [("Some", 1); ("Match_Failure", 1)];

value rec is_expr_constr_call =
  fun
  [ MLast.ExLong _ _ -> True
  | MLast.ExApp _ e _ -> is_expr_constr_call e
  | _ -> False ]
;

value rec constr_expr_arity loc =
  fun
  [ MLast.ExLong _ (MLast.LiUid _ (Ploc.VaVal c)) |
    MLast.ExLong _ (MLast.LiAcc _ _ (Ploc.VaVal c)) ->
      try List.assoc c constr_arity.val with [ Not_found -> 0 ]
  | _ -> 1 ]
;

value rec constr_patt_arity loc =
  fun
  [ MLast.PaLong _ (MLast.LiUid _ (Ploc.VaVal c)) (Ploc.VaVal []) ->
      try List.assoc c constr_arity.val with [ Not_found -> 0 ]
  | MLast.PaPfx _ (MLast.LiUid _ (Ploc.VaVal _)) p -> constr_patt_arity loc p
  | MLast.PaLong _ (MLast.LiAcc _ _ (Ploc.VaVal c)) (Ploc.VaVal []) ->
      try List.assoc c constr_arity.val with [ Not_found -> 0 ]
  | _ -> 1 ]
;

value get_seq =
  fun
  [ MLast.ExSeq _ (Ploc.VaVal el) -> el
  | e -> [e] ]
;

value mem_tvar s tpl =
  List.exists (fun (t, _) -> Pcaml.unvala t = Some s) tpl
;

value choose_tvar tpl =
  let rec find_alpha v =
    let s = String.make 1 v in
    if mem_tvar s tpl then
      if v = 'z' then None else find_alpha (Char.chr (Char.code v + 1))
    else Some (String.make 1 v)
  in
  let rec make_n n =
    let v = "a" ^ string_of_int n in
    if mem_tvar v tpl then make_n (succ n) else v
  in
  match find_alpha 'a' with
  [ Some x -> x
  | None -> make_n 1 ]
;

value quotation_content s =
  loop 0 where rec loop i =
    if i = String.length s then ("", s)
    else if s.[i] = ':' || s.[i] = '@' then
      let i = i + 1 in
      (String.sub s 0 i, String.sub s i (String.length s - i))
    else loop (i + 1)
;

value concat_comm loc e =
  let loc =
    Ploc.with_comment loc
      (Ploc.comment loc ^ Ploc.comment (MLast.loc_of_expr e))
  in
  let floc =
    let first = ref True in
    fun loc1 -> if first.val then do { first.val := False; loc } else loc1
  in
  Reloc.expr floc 0 e
;

value rec expr_of_patt p =
  let loc = MLast.loc_of_patt p in
  match p with
  [ MLast.PaLid _ (Ploc.VaVal x) -> MLast.ExLid loc (Ploc.VaVal x)
  | MLast.PaPfx _ (MLast.LiUid _ (Ploc.VaVal _)) p -> expr_of_patt p
  | _ -> Ploc.raise loc (Stream.Error "identifier expected") ]
;

value build_letop_binder loc letop b l e =
  let (argpat, argexp) =
    List.fold_left
      (fun (argpat, argexp) (andop, (pat, exp)) ->
         (MLast.PaTup loc (Ploc.VaVal [argpat; pat]),
          MLast.ExApp loc
            (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal andop)) argexp)
            exp))
      b l
  in
  MLast.ExApp loc
    (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal letop)) argexp)
    (MLast.ExFun loc (Ploc.VaVal [(argpat, Ploc.VaVal None, e)]))
;

value is_let_exception_f strm =
  Stream.npeek 1 strm = [("", "let")] &&
  (match Stream.npeek 2 strm with
   [ [("", "let"); ("", "exception")] -> True
   | _ -> False ])
;

value check_let_exception_f strm =
  if is_let_exception_f strm then () else raise Stream.Failure
;

value check_let_exception =
   check_let_exception_f
;
value is_let_not_exception_f strm =
  Stream.npeek 1 strm = [("", "let")] &&
  (match Stream.npeek 2 strm with
   [ [("", "let"); ("", "exception")] -> False
   | _ -> True ])
;


value check_let_not_exception_f strm =
  if is_let_not_exception_f strm then () else raise Stream.Failure
;

value check_let_not_exception =
  
    check_let_not_exception_f
;

(* returns True if the stream is a type-decl, and not an extension.
   returns False if the stream is an extension and not a type-decl.
   Since a type-decl might not have an "=" (if it's a list of decls)
   the default is "type-decl".
*)
value word_keywordp s =
  let rec wrec (strm__ : Stream.t _) =
    match Stream.peek strm__ with
    [ Some ('a'..'z' | 'A'..'Z' | '_' | '0'..'9') -> do {
        Stream.junk strm__;
        let strm = strm__ in
        wrec strm
      }
    | _ -> do {
        let strm = strm__ in
        Stream.empty strm;
        True
      } ]
  in
  let check (strm__ : Stream.t _) =
    match Stream.peek strm__ with
    [ Some ('a'..'z' | 'A'..'Z' | '_') -> do {
        Stream.junk strm__;
        let strm = strm__ in
        wrec strm
      }
    | _ -> False ]
  in
  try check (Stream.of_string s) && s <> "_" with
  [ Stream.Failure | Stream.Error _ -> False ]
;

value is_type_decl_not_extension strm =
  wrec 1 where rec wrec n =
    match stream_peek_nth n strm with
    [ None -> assert False
    | Some (("", "=") | ("", ":=") | ("", ";") | ("", ";;")) -> True
    | Some ("", s) when word_keywordp s -> True
    | Some ("EOI", "") -> True
    | Some ("", "+=") -> False
    | Some
        (("", _) | ("UIDENT", _) | ("LIDENT", _) | ("GIDENT", _) |
         ("ANTIQUOT", _)) ->
        wrec (n + 1)
    | Some (a, b) ->
        raise
          (Stream.Error
             (Printf.sprintf
                "unexpected tokens in a type-decl/extension: (\"%s\",\"%s\")"
                a b)) ]
;

value check_type_decl_f strm =
  if is_type_decl_not_extension strm then () else raise Stream.Failure
;

value check_type_decl =
   check_type_decl_f
;

value check_type_extension_f strm =
  if not (is_type_decl_not_extension strm) then () else raise Stream.Failure
;

value check_type_extension =
   check_type_extension_f
;

value check_module_alias_f strm =
  match Stream.npeek 2 strm with
  [ [("UIDENT", _); ("", "=")] -> ()
  | _ -> raise Stream.Failure ]
;

value check_module_alias =
   check_module_alias_f
;


value check_module_not_alias_f strm =
  match Stream.npeek 2 strm with
  [ [("UIDENT", _); ("", "=")] -> raise Stream.Failure
  | _ -> () ]
;

value check_module_not_alias =
  
    check_module_not_alias_f
;

value merge_left_auxiliary_attrs ~{nonterm_name} ~{left_name} ~{right_name}
    left_attrs right_attrs =
  match (left_attrs, right_attrs) with
  [ (l1, Ploc.VaVal l2) -> Ploc.VaVal (l1 @ l2)
  | ([], Ploc.VaAnt _) -> right_attrs
  | _ ->
      failwith
        (Printf.sprintf "%s: cannot specify both %s AND %s antiquotation"
           nonterm_name left_name right_name) ]
;

value merge_right_auxiliary_attrs ~{nonterm_name} ~{left_name} ~{right_name}
    left_attrs right_attrs =
  match (left_attrs, right_attrs) with
  [ (Ploc.VaVal l1, l2) -> Ploc.VaVal (l1 @ l2)
  | (Ploc.VaAnt _, []) -> left_attrs
  | _ ->
      failwith
        (Printf.sprintf "%s: cannot specify both %s antiquotation AND %s"
           nonterm_name left_name right_name) ]
;

value check_dot_uid_f strm =
  crec 1 where rec crec n =
    match stream_npeek n strm with
    [ [(_, tok) :: _] when tok <> "." -> raise Stream.Failure
    | [("", ".")] -> crec (n + 1)
    | [("", "."); ("UIDENT", _)] -> ()
    | [("", "."); ("ANTIQUOT_LOC", s)]
      when
        match Plexer.parse_antiloc s with
        [ Some (_, "uid" | "_uid", _) -> True
        | _ -> False ] ->
        ()
    | [("", "."); ("", "$")] -> crec (n + 1)
    | [("", "."); ("", "$"); ("LIDENT", "uid" | "_uid")] -> crec (n + 1)
    | [("", "."); ("", "$"); ("LIDENT", "uid" | "_uid"); ("", ":")] -> ()
    | _ -> raise Stream.Failure ]
;

value check_dot_uid =
   check_dot_uid_f
;

value is_new_type_extended strm =
  isrec 4 where rec isrec n =
    let l = stream_npeek n strm in
    if l = [] then False
    else
      match l with
      [ [("", "("); ("", "type") :: l] ->
          match sep_last l with
          [ (("", ")"), l) ->
              l <> [] &&
              List.for_all
                (fun
                 [ ("LIDENT", _) -> True
                 | _ -> False ])
                l
          | (("LIDENT", _), _) -> isrec (n + 1)
          | _ -> False ]
      | _ -> False ]
;

value check_new_type_extended_f strm =
  if is_new_type_extended strm then () else raise Stream.Failure
;

value check_new_type_extended =
  
    check_new_type_extended_f
;

value check_not_new_type_extended_f strm =
  if not (is_new_type_extended strm) then () else raise Stream.Failure
;

value check_not_new_type_extended =
  
    check_not_new_type_extended_f
;

value check_uident_coloneq_f strm =
  match stream_npeek 2 strm with
  [ [("UIDENT", _); ("", ":=")] -> ()
  | [("ANTIQUOT", qs); ("", ":=")]
    when prefix_eq "uid:" qs || prefix_eq "_uid:" qs ->
      ()
  | _ -> raise Stream.Failure ]
;

value check_uident_coloneq =
   check_uident_coloneq_f
;

value check_colon_f strm =
  match stream_npeek 1 strm with
  [ [("", ":")] -> ()
  | _ -> raise Stream.Failure ]
;

value check_colon =  check_colon_f;

value check_not_colon_f strm =
  match stream_npeek 1 strm with
  [ [("", ":")] -> raise Stream.Failure
  | _ -> () ]
;

value check_not_colon =
   check_not_colon_f
;

value check_and_or_in_f strm =
  match stream_npeek 1 strm with
  [ [("", "and" | "in")] -> ()
  | _ -> raise Stream.Failure ]
;

value check_and_or_in =
   check_and_or_in_f
;

value uident_True_True_ =
  fun
  [ "True" -> "True_"
  | "False" -> "False_"
  | x -> x ]
;

value make_string_extension loc s =
  let colonpos = String.index s ':' in
  let attrid = String.sub s 0 colonpos in
  let strpayload =
    String.sub s (colonpos + 1) (String.length s - (colonpos + 1))
  in
  (Ploc.VaVal (loc, attrid),
   MLast.StAttr loc
     (Ploc.VaVal
        [MLast.StExp loc (MLast.ExStr loc (Ploc.VaVal strpayload))
           (Ploc.VaVal [])]))
;

value is_lparen_f strm =
  match Stream.npeek 1 strm with
  [ [("", "(")] -> True
  | _ -> False ]
;

value is_lparen_type_f strm =
  is_lparen_f strm &&
  (match Stream.npeek 2 strm with
   [ [("", "("); ("", "type")] -> True
   | _ -> False ])
;

value check_lparen_type_f strm =
  if is_lparen_type_f strm then () else raise Stream.Failure
;

value check_lparen_type =
   check_lparen_type_f
;

value is_type_binder_f strm = check_fsm type_binder_fsm strm;

value check_type_binder_f strm =
  if is_type_binder_f strm then () else raise Stream.Failure
;

value check_type_binder =
   check_type_binder_f
;


(* Main entry points *)


Pcaml.add_option "-no_quot" (Arg.Set Plexer.no_quotations)
  "Don't parse quotations, allowing to use, e.g. \"<:>\" as token";
(* camlp5r *)
(* pa_op.ml,v *)
(* Copyright (c) INRIA 2007-2017 *)


open Asttools;
open Exparser;
open Pcaml;

(* Syntax extensions in Ocaml grammar *)

value rec interf_0 =
  parser bp
    [: :] ->
      match Stream.npeek 2 strm__ with
      [ [("", "#"); ("LIDENT", n)] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          let dp =
            try P.option expr_0 strm__ with
            [ Stream.Failure -> raise (Stream.Error "") ]
          in
          match strm__ with parser
          [ [: `("", ";;"); loc = P.gloc bp ! :] ->
              ([(MLast.SgDir loc (Ploc.VaVal n) (Ploc.VaVal dp), loc)], None)
          | [: :] -> raise (Stream.Error (P.error "interf" "" "';;'")) ]
        }
      | _ ->
          match strm__ with parser
          [ [: `("EOI", _); loc = P.gloc bp ! :] -> ([], Some loc)
          | [: si = sig_item_semi_0;
               (sil, stopped) = interf_0
                 ? P.error "interf" "[sig_item_semi]" "[interf]" :] ->
              ([si :: sil], stopped) ] ]
and implem_0 =
  parser bp
    [: :] ->
      match Stream.npeek 2 strm__ with
      [ [("", "#"); ("LIDENT", n)] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          let dp =
            try P.option expr_0 strm__ with
            [ Stream.Failure -> raise (Stream.Error "") ]
          in
          match strm__ with parser
          [ [: `("", ";;"); loc = P.gloc bp ! :] ->
              ([(MLast.StDir loc (Ploc.VaVal n) (Ploc.VaVal dp), loc)], None)
          | [: :] -> raise (Stream.Error (P.error "implem" "" "';;'")) ]
        }
      | _ ->
          match strm__ with parser
          [ [: `("EOI", _); loc = P.gloc bp ! :] -> ([], Some loc)
          | [: si = str_item_semi_0;
               (sil, stopped) = implem_0
                 ? P.error "implem" "[str_item_semi]" "[implem]" :] ->
              ([si :: sil], stopped) ] ]
and top_phrase_0 =
  parser bp
  [ [: `("EOI", _) :] -> None
  | [: ph = phrase_0;
       `("", ";;") ? P.error "top_phrase" "[phrase]" "';;'" :] ->
      Some ph ]
and use_file_0 =
  parser bp
    [: :] ->
      match Stream.npeek 2 strm__ with
      [ [("", "#"); ("LIDENT", n)] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          let dp =
            try P.option expr_0 strm__ with
            [ Stream.Failure -> raise (Stream.Error "") ]
          in
          match strm__ with parser
          [ [: `("", ";;"); loc = P.gloc bp ! :] ->
              ([MLast.StDir loc (Ploc.VaVal n) (Ploc.VaVal dp)], True)
          | [: :] -> raise (Stream.Error (P.error "use_file" "" "';;'")) ]
        }
      | _ ->
          match strm__ with parser
          [ [: `("EOI", _) :] -> ([], False)
          | [: si = str_item_0; _ = P.option (P.token ("", ";;"));
               (sil, stopped) = use_file_0
                 ? P.error "use_file" "" "[use_file]" :] ->
              ([si :: sil], stopped) ] ]
and phrase_0 =
  parser bp
    [: :] ->
      match Stream.npeek 2 strm__ with
      [ [("", "#"); ("LIDENT", n)] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          let dp =
            try P.option expr_0 strm__ with
            [ Stream.Failure -> raise (Stream.Error "") ]
          in
          let loc = P.gloc bp strm__ in
          MLast.StDir loc (Ploc.VaVal n) (Ploc.VaVal dp)
        }
      | _ ->
          match
            try Some (str_item_0 strm__) with [ Stream.Failure -> None ]
          with
          [ Some sti -> sti
          | _ -> raise Stream.Failure ] ]
and sig_item_semi_0 =
  parser bp
    [: si = sig_item_0; _ = P.option (P.token ("", ";;"));
       loc = P.gloc bp ! :] ->
      (si, loc)
and expr_longident_0 =
  parser bp
    [: li = longident_0;
       a =
         parser
         [ [: `("", ".");
              a =
                parser
                [ [: `("", "(");
                     a =
                       parser
                       [ [: op = operator_rparen; loc = P.gloc bp ! :] ->
                           if op = "::" then
                             MLast.ExLong loc
                               (MLast.LiAcc loc li (Ploc.VaVal op))
                           else
                             MLast.ExFle loc (MLast.ExLong loc li)
                               (Ploc.VaVal (None, Ploc.VaVal op))
                       | [: e = expr_0;
                            `("", ")")
                              ? P.error "expr_longident" "[expr]" "')'";
                            loc = P.gloc bp ! :] ->
                           MLast.ExOpen loc li e ]
                       ? P.error "expr_longident" "'('"
                           "[operator_rparen] or [expr]" :] ->
                    a
                | [: id = P.vala (P.token ("LIDENT", ""));
                     loc = P.gloc bp ! :] ->
                    MLast.ExFle loc (MLast.ExLong loc li)
                      (Ploc.VaVal (None, id))
                | [: _ = check_lbracket;
                     e = expr_18
                       ? P.error "expr_longident" "[check_lbracket]"
                           "[expr level simple]";
                     loc = P.gloc bp ! :] ->
                    MLast.ExOpen loc li e
                | [: _ = check_lbrace;
                     e = expr_18
                       ? P.error "expr_longident" "[check_lbrace]"
                           "[expr level simple]";
                     loc = P.gloc bp ! :] ->
                    MLast.ExOpen loc li e
                | [: _ = check_lbracketbar;
                     e = expr_18
                       ? P.error "expr_longident" "[check_lbracketbar]"
                           "[expr level simple]";
                     loc = P.gloc bp ! :] ->
                    MLast.ExOpen loc li e ]
                ? P.error "expr_longident" "'.'"
                    "'(' or lowercase identifier or [check_lbracket] or [check_lbrace] or [check_lbracketbar]" :] ->
             a
         | [: loc = P.gloc bp :] -> MLast.ExLong loc li ]
         ? P.error "expr_longident" "[longident]" "'.'" :] ->
      a
and lbl_expr_0 =
  parser bp
    [: i = patt_label_ident_0;
       tycon =
         P.option
           (parser
              [: `("", ":");
                 c = ctyp_0 ? P.error "implem" "':'" "[ctyp]" :] ->
                c);
       e =
         P.option
           (parser
              [: `("", "=");
                 e = expr_1
                   ? P.error "implem" "'='" "[expr level expr1]" :] ->
                e);
       loc = P.gloc bp ! :] ->
      let last =
        match i with
        [ MLast.PaPfx _ _ (MLast.PaLid _ (Ploc.VaVal j)) -> j
        | MLast.PaLid _ (Ploc.VaVal j) -> j
        | _ -> failwith "internal error: lbl_expr" ]
      in
      let rhs =
        match e with
        [ None -> MLast.ExLid loc (Ploc.VaVal last)
        | Some e -> e ]
      in
      let rhs =
        match tycon with
        [ None -> rhs
        | Some ty -> MLast.ExTyc loc rhs ty ]
      in
      (i, rhs)
and lbl_expr_list_0 =
  parser bp
    [: le = lbl_expr_0;
       a =
         parser
         [ [: `("", ";");
              a =
                parser
                [ [: lel = lbl_expr_list_0 :] -> [le :: lel]
                | [: :] -> [le] ]
                ? P.error "lbl_expr_list" "';'" "[lbl_expr_list]" :] ->
             a
         | [: :] -> [le] ]
         ? P.error "lbl_expr_list" "[lbl_expr]" "';'" :] ->
      a
and expr1_semi_list_0 =
  parser bp [: el = P.list1sep expr_1 (P.token ("", ";")) True :] -> el
and label_0 = parser bp [: `("LIDENT", i) :] -> i
and field_expr_list_0 =
  parser bp
    [: l = label_0; `("", "=") ? P.error "field_expr_list" "[label]" "'='";
       e = expr_1 ? P.error "field_expr_list" "'='" "[expr level expr1]";
       a =
         parser
         [ [: `("", ";");
              a =
                parser
                [ [: fel = field_expr_list_0 :] -> [(l, e) :: fel]
                | [: :] -> [(l, e)] ]
                ? P.error "field_expr_list" "';'" "[field_expr_list]" :] ->
             a
         | [: :] -> [(l, e)] ]
         ? P.error "field_expr_list" "[expr level expr1]" "';'" :] ->
      a
and stream_expr_comp_0 =
  parser bp
  [ [: `("", "'");
       e = expr_1 ? P.error "stream_expr_comp" "'''" "[expr level expr1]";
       loc = P.gloc bp ! :] ->
      SeTrm loc e
  | [: e = expr_1; loc = P.gloc bp ! :] -> SeNtr loc e ]
and stream_expr_comp_list_0 =
  parser bp
    [: se = stream_expr_comp_0;
       a =
         parser
         [ [: `("", ";");
              a =
                parser
                [ [: sel = stream_expr_comp_list_0 :] -> [se :: sel]
                | [: :] -> [se] ]
                ? P.error "stream_expr_comp_list" "';'"
                    "[stream_expr_comp_list]" :] ->
             a
         | [: :] -> [se] ]
         ? P.error "stream_expr_comp_list" "[stream_expr_comp]" "';'" :] ->
      a
and andop_binding_0 =
  parser bp
    [: op = andop;
       b = letop_binding_0
         ? P.error "andop_binding" "[andop]" "[letop_binding]" :] ->
      (op, b)
and letop_binding_0 =
  parser bp
  [ [: p = val_ident_0;
       a =
         parser
         [ [: e = fun_binding_0 :] -> (p, e)
         | [: loc = P.gloc bp :] ->
             match p with
             [ MLast.PaLid _ (Ploc.VaVal s) ->
                 (p, MLast.ExLid loc (Ploc.VaVal s))
             | _ ->
                 failwith
                   "letop punning: binder must be a lowercase ident (variable)" ] ]
         ? P.error "letop_binding" "[val_ident]" "[fun_binding]" :] ->
      a
  | [: p = patt_0;
       a =
         parser
         [ [: `("", "=");
              e = expr_0 ? P.error "letop_binding" "'='" "[expr]" :] ->
             (p, e)
         | [: `("", ":");
              t = poly_type_0 ? P.error "letop_binding" "':'" "[poly_type]";
              `("", "=") ? P.error "letop_binding" "[poly_type]" "'='";
              e = expr_0 ? P.error "letop_binding" "'='" "[expr]";
              loc = P.gloc bp ! :] ->
             (MLast.PaTyc loc p t, e) ]
         ? P.error "letop_binding" "[patt]" "'=' or ':'" :] ->
      a ]
and direction_flag_0 =
  parser bp
  [ [: `("", "to") :] -> True
  | [: `("", "downto") :] -> False ]
and fun_def_0 =
  parser bp
  [ [: p = labeled_patt_0;
       (eo, e) = fun_def_0 ? P.error "fun_def" "[labeled_patt]" "[fun_def]";
       loc = P.gloc bp ! :] ->
      (None, MLast.ExFun loc (Ploc.VaVal [(p, Ploc.VaVal eo, e)]))
  | [: p = patt_8;
       (eo, e) = fun_def_0
         ? P.error "fun_def" "[patt level simple]" "[fun_def]";
       loc = P.gloc bp ! :] ->
      (None, MLast.ExFun loc (Ploc.VaVal [(p, Ploc.VaVal eo, e)]))
  | [: eo =
         P.option
           (parser
              [: `("", "when");
                 e = expr_0 ? P.error "implem" "'when'" "[expr]" :] ->
                e);
       tyo =
         P.option
           (parser
              [: `("", ":");
                 ty = ctyp_6
                   ? P.error "implem" "':'" "[ctyp level apply]" :] ->
                ty);
       `("", "->") ? P.error "fun_def" "" "'->'";
       e = expr_0 ? P.error "fun_def" "'->'" "[expr]"; loc = P.gloc bp ! :] ->
      let e =
        match tyo with
        [ None -> e
        | Some ty -> MLast.ExTyc loc e ty ]
      in
      (eo, e)
  | [: :] -> raise (Stream.Error "") ]
and expr_or_dot_0 =
  parser bp
  [ [: `("", "."); loc = P.gloc bp ! :] -> MLast.ExUnr loc
  | [: e = expr_0 :] -> e ]
and match_case_0 =
  parser bp
    [: x1 = patt_0;
       w =
         P.vala
           (P.option
              (parser
                 [: `("", "when");
                    e = expr_0 ? P.error "implem" "'when'" "[expr]" :] ->
                   e))
         ? P.error "match_case" "[patt]" "???";
       `("", "->") ? P.error "match_case" "???" "'->'";
       e = expr_or_dot_0 ? P.error "match_case" "'->'" "[expr_or_dot]" :] ->
      (x1, w, e)
and stream_patt_let_0 =
  parser bp
    [: `("", "let");
       p = ipatt_0 ? P.error "stream_patt_let" "'let'" "[ipatt]";
       `("", "=") ? P.error "stream_patt_let" "[ipatt]" "'='";
       e = expr_0 ? P.error "stream_patt_let" "'='" "[expr]";
       `("", "in") ? P.error "stream_patt_let" "[expr]" "'in'";
       loc = P.gloc bp ! :] ->
      (SpLet loc p e, SpoNoth)
and stream_patt_comp_err_0 =
  parser bp
    [: spc = stream_patt_comp_0;
       a =
         parser
         [ [: `("", "??");
              e = expr_1
                ? P.error "stream_patt_comp_err" "'??'"
                    "[expr level expr1]" :] ->
             (spc, SpoQues e)
         | [: `("", "?!") :] -> (spc, SpoBang)
         | [: :] -> (spc, SpoNoth) ]
         ? P.error "stream_patt_comp_err" "[stream_patt_comp]"
             "'??' or '?!'" :] ->
      a
and stream_patt_kont_0 =
  parser bp
  [ [: spc = stream_patt_comp_err_0;
       a =
         parser
         [ [: `("", ";");
              a =
                parser
                [ [: sp = stream_patt_kont_0 :] -> [spc :: sp]
                | [: :] -> [spc] ]
                ? P.error "stream_patt_kont" "';'" "[stream_patt_kont]" :] ->
             a
         | [: :] -> [spc] ]
         ? P.error "stream_patt_kont" "[stream_patt_comp_err]" "';'" :] ->
      a
  | [: spc = stream_patt_let_0;
       sp = stream_patt_kont_0
         ? P.error "stream_patt_kont" "[stream_patt_let]"
             "[stream_patt_kont]" :] ->
      [spc :: sp] ]
and lookahead_0 =
  parser bp
    [: `("", "[");
       pl = P.list1sep patt_0 (P.token ("", ";")) False
         ? P.error "lookahead" "'['" "[patt]";
       `("", "]") ? P.error "lookahead" "" "[patt] or ']'" :] ->
      pl
and patt_ident_0 =
  parser bp
  [ [: s = P.vala (P.token ("LIDENT", "")); loc = P.gloc bp ! :] ->
      MLast.PaLid loc s
  | [: s = P.vala (P.token ("GIDENT", "")); loc = P.gloc bp ! :] ->
      MLast.PaLid loc s
  | [: li = longident_0;
       a =
         parser
         [ [: `("", ".");
              p = patt_8 ? P.error "patt_ident" "'.'" "[patt level simple]";
              loc = P.gloc bp ! :] ->
             match p with
             [ MLast.PaLong _ (MLast.LiUid _ (Ploc.VaVal i))
                 (Ploc.VaVal []) ->
                 let i = uident_True_True_ i in
                 let li = MLast.LiAcc loc li (Ploc.VaVal i) in
                 MLast.PaLong loc li (Ploc.VaVal [])
             | _ -> MLast.PaPfx loc li p ]
         | [: _ = check_lparen_type :] ->
             match Stream.npeek 2 strm__ with
             [ [("", "("); ("", "type")] -> do {
                 Stream.junk strm__;
                 Stream.junk strm__;
                 match
                   try
                     Some
                       (P.vala
                          (P.list1
                             (parser
                                [: `("LIDENT", s); loc = P.gloc bp ! :] ->
                                  (loc, s)))
                          strm__)
                   with
                   [ Stream.Failure -> None ]
                 with
                 [ Some loc_ids ->
                     match strm__ with parser
                     [ [: `("", ")"); loc = P.gloc bp ! :] ->
                         MLast.PaLong loc li loc_ids
                     | [: :] ->
                         raise
                           (Stream.Error (P.error "patt_ident" "???" "')'")) ]
                 | _ ->
                     raise (Stream.Error (P.error "patt_ident" "'('" "???")) ]
               }
             | _ ->
                 raise
                   (Stream.Error
                      (P.error "patt_ident" "[check_lparen_type]" "'('")) ]
         | [: loc = P.gloc bp :] -> MLast.PaLong loc li (Ploc.VaVal []) ]
         ? P.error "patt_ident" "[longident]"
             "'.' or [check_lparen_type]" :] ->
      a ]
and patt_label_ident_0 =
  parser bp
  [ [: `("LIDENT", i); loc = P.gloc bp ! :] -> MLast.PaLid loc (Ploc.VaVal i)
  | [: li = longident_0;
       `("", ".") ? P.error "patt_label_ident" "[longident]" "'.'";
       p2 = patt_label_ident_0
         ? P.error "patt_label_ident" "'.'" "[patt_label_ident]";
       loc = P.gloc bp ! :] ->
      MLast.PaPfx loc li p2 ]
and lbl_patt_0 =
  parser bp
  [ [: `("", "_"); loc = P.gloc bp ! :] -> (MLast.PaAny loc, MLast.PaAny loc)
  | [: i = patt_label_ident_0;
       tycon =
         P.option
           (parser
              [: `("", ":");
                 c = ctyp_0 ? P.error "implem" "':'" "[ctyp]" :] ->
                c);
       p =
         P.option
           (parser
              [: `("", "=");
                 p = patt_0 ? P.error "implem" "'='" "[patt]" :] ->
                p);
       loc = P.gloc bp ! :] ->
      let last =
        match i with
        [ MLast.PaPfx _ _ (MLast.PaLid _ (Ploc.VaVal j)) -> j
        | MLast.PaLid _ (Ploc.VaVal j) -> j
        | _ -> failwith "internal error: lbl_patt" ]
      in
      let rhs =
        match p with
        [ None -> MLast.PaLid loc (Ploc.VaVal last)
        | Some p -> p ]
      in
      let rhs =
        match tycon with
        [ None -> rhs
        | Some ty -> MLast.PaTyc loc rhs ty ]
      in
      (i, rhs) ]
and lbl_patt_list_0 =
  parser bp
    [: le = lbl_patt_0;
       a =
         parser
         [ [: `("", ";");
              a =
                parser
                [ [: lel = lbl_patt_list_0 :] -> [le :: lel]
                | [: :] -> [le] ]
                ? P.error "lbl_patt_list" "';'" "[lbl_patt_list]" :] ->
             a
         | [: :] -> [le] ]
         ? P.error "lbl_patt_list" "[lbl_patt]" "';'" :] ->
      a
and patt_semi_list_0 =
  parser bp
    [: p = patt_0;
       a =
         parser
         [ [: `("", ";");
              a =
                parser
                [ [: pl = patt_semi_list_0 :] -> [p :: pl]
                | [: :] -> [p] ]
                ? P.error "patt_semi_list" "';'" "[patt_semi_list]" :] ->
             a
         | [: :] -> [p] ]
         ? P.error "patt_semi_list" "[patt]" "';'" :] ->
      a
and attribute_structure_0 =
  parser bp
    [: st =
         P.vala
           (P.list1
              (parser
                 [: s = str_item_0; _ = P.option (P.token ("", ";;")) :] ->
                   s)) :] ->
      st
and rest_mod_decl_binding_0 =
  parser bp
    [: `("", "and");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "rest_mod_decl_binding" "'and'" "[alg_attributes_no_anti]";
       i = uidopt_no_anti_0
         ? P.error "rest_mod_decl_binding" "[alg_attributes_no_anti]"
             "[uidopt_no_anti]";
       mt = module_declaration_0
         ? P.error "rest_mod_decl_binding" "[uidopt_no_anti]"
             "[module_declaration]";
       item_attrs = item_attributes_0
         ? P.error "rest_mod_decl_binding" "[module_declaration]"
             "[item_attributes]";
       loc = P.gloc bp ! :] ->
      let attrs =
        merge_left_auxiliary_attrs ~{nonterm_name = "mod_decl_binding"}
          ~{left_name = "algebraic attributes"}
          ~{right_name = "item attributes"} alg_attrs item_attrs
      in
      (Ploc.VaVal i, mt, attrs)
and module_declaration_0 =
  parser bp
  [ [: `("", ":");
       mt = module_type_0
         ? P.error "module_declaration" "':'" "[module_type]" :] ->
      mt
  | [: arg = P.vala functor_parameter_0;
       mt = module_declaration_0
         ? P.error "module_declaration" "[functor_parameter]"
             "[module_declaration]";
       loc = P.gloc bp ! :] ->
      MLast.MtFun loc arg mt ]
and uidopt_no_anti_0 =
  parser bp
  [ [: `("", "_") :] -> None
  | [: `("UIDENT", m); loc = P.gloc bp ! :] -> Some (Ploc.VaVal m) ]
and first_mod_decl_binding_0 =
  parser bp
    [: i = uidopt_no_anti_0;
       mt = module_declaration_0
         ? P.error "first_mod_decl_binding" "[uidopt_no_anti]"
             "[module_declaration]";
       attrs = item_attributes_0
         ? P.error "first_mod_decl_binding" "[module_declaration]"
             "[item_attributes]";
       loc = P.gloc bp ! :] ->
      (Ploc.VaVal i, mt, attrs)
and class_description_0 =
  parser bp
    [: alg_attrs = alg_attributes_no_anti_0;
       vf = P.vala (P.bool (P.token ("", "virtual")))
         ? P.error "class_description" "[alg_attributes_no_anti]" "???";
       ctp = class_type_parameters_0
         ? P.error "class_description" "???" "[class_type_parameters]" :] ->
      match Stream.npeek 2 strm__ with
      [ [("LIDENT", n); ("", ":")] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          let n = Ploc.VaVal n in
          match
            try Some (class_type_0 strm__) with [ Stream.Failure -> None ]
          with
          [ Some ct ->
              match
                try Some (item_attributes_0 strm__) with
                [ Stream.Failure -> None ]
              with
              [ Some item_attrs ->
                  let loc = P.gloc bp strm__ in
                  let attrs =
                    merge_left_auxiliary_attrs
                      ~{nonterm_name = "class-description"}
                      ~{left_name = "algebraic attributes"}
                      ~{right_name = "item attributes"} alg_attrs item_attrs
                  in
                  {MLast.ciLoc = loc; MLast.ciVir = vf; MLast.ciPrm = ctp;
                   MLast.ciNam = n; MLast.ciExp = ct;
                   MLast.ciAttributes = attrs}
              | _ ->
                  raise
                    (Stream.Error
                       (P.error "class_description" "[class_type]"
                          "[item_attributes]")) ]
          | _ ->
              raise
                (Stream.Error
                   (P.error "class_description" "lowercase identifier"
                      "[class_type]")) ]
        }
      | _ ->
          raise
            (Stream.Error
               (P.error "class_description" "[class_type_parameters]"
                  "lowercase identifier")) ]
and class_sig_item_0 =
  parser bp
  [ [: `("", "inherit");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "class_sig_item" "'inherit'" "[alg_attributes_no_anti]";
       cs = class_signature_0
         ? P.error "class_sig_item" "[alg_attributes_no_anti]"
             "[class_signature]";
       item_attrs = item_attributes_0
         ? P.error "class_sig_item" "[class_signature]" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      let attrs =
        merge_left_auxiliary_attrs ~{nonterm_name = "csig-inherit"}
          ~{left_name = "algebraic attributes"}
          ~{right_name = "item attributes"} alg_attrs item_attrs
      in
      MLast.CgInh loc cs attrs
  | [: `("", "val");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "class_sig_item" "'val'" "[alg_attributes_no_anti]";
       (mf, vf) = mut_virt_0
         ? P.error "class_sig_item" "[alg_attributes_no_anti]"
             "[mut_virt]" :] ->
      match Stream.npeek 2 strm__ with
      [ [("LIDENT", l); ("", ":")] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          let l = Ploc.VaVal l in
          match try Some (ctyp_0 strm__) with [ Stream.Failure -> None ] with
          [ Some t ->
              match
                try Some (item_attributes_0 strm__) with
                [ Stream.Failure -> None ]
              with
              [ Some item_attrs ->
                  let loc = P.gloc bp strm__ in
                  let attrs =
                    merge_left_auxiliary_attrs
                      ~{nonterm_name = "csig-inherit"}
                      ~{left_name = "algebraic attributes"}
                      ~{right_name = "item attributes"} alg_attrs item_attrs
                  in
                  MLast.CgVal loc (Ploc.VaVal mf) (Ploc.VaVal vf) l t attrs
              | _ ->
                  raise
                    (Stream.Error
                       (P.error "class_sig_item" "[ctyp]"
                          "[item_attributes]")) ]
          | _ ->
              raise
                (Stream.Error
                   (P.error "class_sig_item" "lowercase identifier"
                      "[ctyp]")) ]
        }
      | _ ->
          raise
            (Stream.Error
               (P.error "class_sig_item" "[mut_virt]"
                  "lowercase identifier")) ]
  | [: `("", "method");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "class_sig_item" "'method'" "[alg_attributes_no_anti]";
       a =
         parser
         [ [: `("", "virtual");
              a =
                parser
                  [: :] ->
                    match Stream.npeek 3 strm__ with
                    [ [("", "private"); ("LIDENT", l); ("", ":")] -> do {
                        Stream.junk strm__;
                        Stream.junk strm__;
                        Stream.junk strm__;
                        let l = Ploc.VaVal l in
                        match
                          try
                            Some (poly_type_below_alg_attribute_0 strm__)
                          with
                          [ Stream.Failure -> None ]
                        with
                        [ Some t ->
                            match
                              try Some (item_attributes_0 strm__) with
                              [ Stream.Failure -> None ]
                            with
                            [ Some item_attrs ->
                                let loc = P.gloc bp strm__ in
                                let attrs =
                                  merge_left_auxiliary_attrs
                                    ~{nonterm_name = "csig-inherit"}
                                    ~{left_name = "algebraic attributes"}
                                    ~{right_name = "item attributes"}
                                    alg_attrs item_attrs
                                in
                                MLast.CgVir loc (Ploc.VaVal True) l t attrs
                            | _ ->
                                raise
                                  (Stream.Error
                                     (P.error "class_sig_item"
                                        "[poly_type_below_alg_attribute]"
                                        "[item_attributes]")) ]
                        | _ ->
                            raise
                              (Stream.Error
                                 (P.error "class_sig_item" "'private'"
                                    "[poly_type_below_alg_attribute]")) ]
                      }
                    | _ ->
                        match Stream.npeek 2 strm__ with
                        [ [("LIDENT", l); ("", ":")] -> do {
                            Stream.junk strm__;
                            Stream.junk strm__;
                            let l = Ploc.VaVal l in
                            match
                              try
                                Some (poly_type_below_alg_attribute_0 strm__)
                              with
                              [ Stream.Failure -> None ]
                            with
                            [ Some t ->
                                match
                                  try Some (item_attributes_0 strm__) with
                                  [ Stream.Failure -> None ]
                                with
                                [ Some item_attrs ->
                                    let loc = P.gloc bp strm__ in
                                    let attrs =
                                      merge_left_auxiliary_attrs
                                        ~{nonterm_name = "csig-inherit"}
                                        ~{left_name = "algebraic attributes"}
                                        ~{right_name = "item attributes"}
                                        alg_attrs item_attrs
                                    in
                                    MLast.CgVir loc (Ploc.VaVal False) l t
                                      attrs
                                | _ ->
                                    raise
                                      (Stream.Error
                                         (P.error "class_sig_item"
                                            "[poly_type_below_alg_attribute]"
                                            "[item_attributes]")) ]
                            | _ ->
                                raise
                                  (Stream.Error
                                     (P.error "class_sig_item"
                                        "lowercase identifier"
                                        "[poly_type_below_alg_attribute]")) ]
                          }
                        | _ -> raise Stream.Failure ] ]
                ? P.error "class_sig_item" "'virtual'"
                    "'private' or lowercase identifier" :] ->
             a
         | [: `("", "private");
              a =
                parser
                  [: :] ->
                    match Stream.npeek 3 strm__ with
                    [ [("", "virtual"); ("LIDENT", l); ("", ":")] -> do {
                        Stream.junk strm__;
                        Stream.junk strm__;
                        Stream.junk strm__;
                        let l = Ploc.VaVal l in
                        match
                          try
                            Some (poly_type_below_alg_attribute_0 strm__)
                          with
                          [ Stream.Failure -> None ]
                        with
                        [ Some t ->
                            match
                              try Some (item_attributes_0 strm__) with
                              [ Stream.Failure -> None ]
                            with
                            [ Some item_attrs ->
                                let loc = P.gloc bp strm__ in
                                let attrs =
                                  merge_left_auxiliary_attrs
                                    ~{nonterm_name = "csig-inherit"}
                                    ~{left_name = "algebraic attributes"}
                                    ~{right_name = "item attributes"}
                                    alg_attrs item_attrs
                                in
                                MLast.CgVir loc (Ploc.VaVal True) l t attrs
                            | _ ->
                                raise
                                  (Stream.Error
                                     (P.error "class_sig_item"
                                        "[poly_type_below_alg_attribute]"
                                        "[item_attributes]")) ]
                        | _ ->
                            raise
                              (Stream.Error
                                 (P.error "class_sig_item" "'virtual'"
                                    "[poly_type_below_alg_attribute]")) ]
                      }
                    | _ ->
                        match Stream.npeek 2 strm__ with
                        [ [("LIDENT", l); ("", ":")] -> do {
                            Stream.junk strm__;
                            Stream.junk strm__;
                            let l = Ploc.VaVal l in
                            match
                              try
                                Some (poly_type_below_alg_attribute_0 strm__)
                              with
                              [ Stream.Failure -> None ]
                            with
                            [ Some t ->
                                match
                                  try Some (item_attributes_0 strm__) with
                                  [ Stream.Failure -> None ]
                                with
                                [ Some item_attrs ->
                                    let loc = P.gloc bp strm__ in
                                    let attrs =
                                      merge_left_auxiliary_attrs
                                        ~{nonterm_name = "csig-inherit"}
                                        ~{left_name = "algebraic attributes"}
                                        ~{right_name = "item attributes"}
                                        alg_attrs item_attrs
                                    in
                                    MLast.CgMth loc (Ploc.VaVal True) l t
                                      attrs
                                | _ ->
                                    raise
                                      (Stream.Error
                                         (P.error "class_sig_item"
                                            "[poly_type_below_alg_attribute]"
                                            "[item_attributes]")) ]
                            | _ ->
                                raise
                                  (Stream.Error
                                     (P.error "class_sig_item"
                                        "lowercase identifier"
                                        "[poly_type_below_alg_attribute]")) ]
                          }
                        | _ -> raise Stream.Failure ] ]
                ? P.error "class_sig_item" "'private'"
                    "'virtual' or lowercase identifier" :] ->
             a
         | [: :] ->
             match Stream.npeek 2 strm__ with
             [ [("LIDENT", l); ("", ":")] -> do {
                 Stream.junk strm__;
                 Stream.junk strm__;
                 let l = Ploc.VaVal l in
                 match
                   try Some (poly_type_below_alg_attribute_0 strm__) with
                   [ Stream.Failure -> None ]
                 with
                 [ Some t ->
                     match
                       try Some (item_attributes_0 strm__) with
                       [ Stream.Failure -> None ]
                     with
                     [ Some item_attrs ->
                         let loc = P.gloc bp strm__ in
                         let attrs =
                           merge_left_auxiliary_attrs
                             ~{nonterm_name = "csig-inherit"}
                             ~{left_name = "algebraic attributes"}
                             ~{right_name = "item attributes"} alg_attrs
                             item_attrs
                         in
                         MLast.CgMth loc (Ploc.VaVal False) l t attrs
                     | _ ->
                         raise
                           (Stream.Error
                              (P.error "class_sig_item"
                                 "[poly_type_below_alg_attribute]"
                                 "[item_attributes]")) ]
                 | _ ->
                     raise
                       (Stream.Error
                          (P.error "class_sig_item" "lowercase identifier"
                             "[poly_type_below_alg_attribute]")) ]
               }
             | _ -> raise Stream.Failure ] ]
         ? P.error "class_sig_item" "[alg_attributes_no_anti]"
             "'virtual' or 'private' or lowercase identifier" :] ->
      a
  | [: `("", "constraint");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "class_sig_item" "'constraint'" "[alg_attributes_no_anti]";
       t1 = ctyp_0
         ? P.error "class_sig_item" "[alg_attributes_no_anti]" "[ctyp]";
       `("", "=") ? P.error "class_sig_item" "[ctyp]" "'='";
       t2 = ctyp_0 ? P.error "class_sig_item" "'='" "[ctyp]";
       item_attrs = item_attributes_0
         ? P.error "class_sig_item" "[ctyp]" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      let attrs =
        merge_left_auxiliary_attrs ~{nonterm_name = "csig-inherit"}
          ~{left_name = "algebraic attributes"}
          ~{right_name = "item attributes"} alg_attrs item_attrs
      in
      MLast.CgCtr loc t1 t2 attrs
  | [: attr = floating_attribute_0; loc = P.gloc bp ! :] ->
      MLast.CgFlAtt loc attr
  | [: e = item_extension_0; loc = P.gloc bp ! :] -> MLast.CgExten loc e ]
and class_self_type_0 =
  parser bp
    [: `("", "("); t = ctyp_0 ? P.error "class_self_type" "'('" "[ctyp]";
       `("", ")") ? P.error "class_self_type" "[ctyp]" "')'" :] ->
      t
and signature_0 =
  parser bp
    [: sg =
         P.vala
           (P.list0
              (parser
                 [: s = sig_item_0; _ = P.option (P.token ("", ";;")) :] ->
                   s)) :] ->
      sg
and mod_expr_ident_0 =
  parser bp
    [: a = mod_expr_ident_1; a = mod_expr_ident_gen_cont bp a ! :] -> a
and mod_expr_ident_0_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", ".");
       j = P.orzero mod_expr_ident_1 mod_expr_ident_0
         ? P.error "mod_expr_ident" "'.'" "[mod_expr_ident]";
       loc = P.gloc bp !;
       let i = a__ in
       a = mod_expr_ident_0_cont bp (MLast.MeAcc loc i j) ! :] ->
      a
  | [: :] -> a__ ]
and mod_expr_ident_1 =
  parser bp
    [: i = P.vala (P.token ("UIDENT", "")); loc = P.gloc bp ! :] ->
      MLast.MeUid loc i
and mod_expr_ident_gen_cont bp a__ strm__ =
  let x = mod_expr_ident_0_cont bp a__ strm__ in
  if x == a__ then x else mod_expr_ident_gen_cont bp x strm__
and paren_module_expr_0 =
  parser bp
    [: `("", "(");
       a =
         parser
         [ [: `("", "val");
              alg_attrs = alg_attributes_no_anti_0
                ? P.error "paren_module_expr" "'val'"
                    "[alg_attributes_no_anti]";
              e = expr_0
                ? P.error "paren_module_expr" "[alg_attributes_no_anti]"
                    "[expr]";
              a =
                parser
                [ [: `("", ":");
                     mt1 = module_type_0
                       ? P.error "paren_module_expr" "':'" "[module_type]";
                     a =
                       parser
                       [ [: `("", ":>");
                            mt2 = module_type_0
                              ? P.error "paren_module_expr" "':>'"
                                  "[module_type]";
                            `("", ")")
                              ? P.error "paren_module_expr" "[module_type]"
                                  "')'";
                            loc = P.gloc bp ! :] ->
                           module_expr_wrap_attrs
                             (MLast.MeUnp loc e (Some mt1) (Some mt2))
                             alg_attrs
                       | [: `("", ")"); loc = P.gloc bp ! :] ->
                           module_expr_wrap_attrs
                             (MLast.MeUnp loc e (Some mt1) None) alg_attrs ]
                       ? P.error "paren_module_expr" "[module_type]"
                           "':>' or ')'" :] ->
                    a
                | [: `("", ")"); loc = P.gloc bp ! :] ->
                    module_expr_wrap_attrs (MLast.MeUnp loc e None None)
                      alg_attrs ]
                ? P.error "paren_module_expr" "[expr]" "':' or ')'" :] ->
             a
         | [: me = module_expr_0;
              a =
                parser
                [ [: `("", ":");
                     mt = module_type_0
                       ? P.error "paren_module_expr" "':'" "[module_type]";
                     `("", ")")
                       ? P.error "paren_module_expr" "[module_type]" "')'";
                     loc = P.gloc bp ! :] ->
                    MLast.MeTyc loc me mt
                | [: `("", ")") :] -> me ]
                ? P.error "paren_module_expr" "[module_expr]"
                    "':' or ')'" :] ->
             a ]
         ? P.error "paren_module_expr" "'('" "'val' or [module_expr]" :] ->
      a
and and_let_binding_0 =
  parser bp
    [: `("", "and");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "and_let_binding" "'and'" "[alg_attributes_no_anti]";
       a =
         parser
         [ [: p = val_ident_0;
              a =
                parser
                [ [: _ = check_and_or_in; loc = P.gloc bp ! :] ->
                    match p with
                    [ MLast.PaLid _ (Ploc.VaVal s) ->
                        (p, MLast.ExLid loc (Ploc.VaVal s),
                         Ploc.VaVal alg_attrs)
                    | _ ->
                        failwith
                          "let punning: binder must be a lowercase ident (variable)" ]
                | [: _ = check_colon;
                     e = fun_binding_0
                       ? P.error "and_let_binding" "[check_colon]"
                           "[fun_binding]";
                     item_attrs = item_attributes_0
                       ? P.error "and_let_binding" "[fun_binding]"
                           "[item_attributes]";
                     loc = P.gloc bp ! :] ->
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "let_binding"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    let (p, e) =
                      match e with
                      [ MLast.ExTyc _ e t -> (MLast.PaTyc loc p t, e)
                      | _ -> (p, e) ]
                    in
                    (p, e, attrs)
                | [: _ = check_not_colon;
                     e = fun_binding_0
                       ? P.error "and_let_binding" "[check_not_colon]"
                           "[fun_binding]";
                     item_attrs = item_attributes_0
                       ? P.error "and_let_binding" "[fun_binding]"
                           "[item_attributes]";
                     loc = P.gloc bp ! :] ->
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "let_binding"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    (p, e, attrs) ]
                ? P.error "and_let_binding" "[val_ident]"
                    "[check_and_or_in] or [check_colon] or [check_not_colon]" :] ->
             a
         | [: p = patt_0;
              a =
                parser
                [ [: `("", "=");
                     e = expr_0 ? P.error "and_let_binding" "'='" "[expr]";
                     item_attrs = item_attributes_0
                       ? P.error "and_let_binding" "[expr]"
                           "[item_attributes]";
                     loc = P.gloc bp ! :] ->
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "let_binding"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    (p, e, attrs)
                | [: `("", ":");
                     t = poly_type_0
                       ? P.error "and_let_binding" "':'" "[poly_type]";
                     `("", "=")
                       ? P.error "and_let_binding" "[poly_type]" "'='";
                     e = expr_0 ? P.error "and_let_binding" "'='" "[expr]";
                     item_attrs = item_attributes_0
                       ? P.error "and_let_binding" "[expr]"
                           "[item_attributes]";
                     loc = P.gloc bp ! :] ->
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "let_binding"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    (MLast.PaTyc loc p t, e, attrs) ]
                ? P.error "and_let_binding" "[patt]" "'=' or ':'" :] ->
             a ]
         ? P.error "and_let_binding" "[alg_attributes_no_anti]"
             "[val_ident] or [patt]" :] ->
      a
and first_let_binding_0 =
  parser bp
  [ [: p = val_ident_0;
       a =
         parser
         [ [: _ = check_and_or_in; loc = P.gloc bp ! :] ->
             match p with
             [ MLast.PaLid _ (Ploc.VaVal s) ->
                 (p, MLast.ExLid loc (Ploc.VaVal s), Ploc.VaVal [])
             | _ ->
                 failwith
                   "let punning: binder must be a lowercase ident (variable)" ]
         | [: _ = check_colon;
              e = fun_binding_0
                ? P.error "first_let_binding" "[check_colon]" "[fun_binding]";
              item_attrs = item_attributes_0
                ? P.error "first_let_binding" "[fun_binding]"
                    "[item_attributes]";
              loc = P.gloc bp ! :] ->
             let (p, e) =
               match e with
               [ MLast.ExTyc _ e t -> (MLast.PaTyc loc p t, e)
               | _ -> (p, e) ]
             in
             (p, e, item_attrs)
         | [: _ = check_not_colon;
              e = fun_binding_0
                ? P.error "first_let_binding" "[check_not_colon]"
                    "[fun_binding]";
              item_attrs = item_attributes_0
                ? P.error "first_let_binding" "[fun_binding]"
                    "[item_attributes]" :] ->
             (p, e, item_attrs) ]
         ? P.error "first_let_binding" "[val_ident]"
             "[check_and_or_in] or [check_colon] or [check_not_colon]" :] ->
      a
  | [: p = patt_0;
       a =
         parser
         [ [: `("", "=");
              e = expr_0 ? P.error "first_let_binding" "'='" "[expr]";
              item_attrs = item_attributes_0
                ? P.error "first_let_binding" "[expr]"
                    "[item_attributes]" :] ->
             (p, e, item_attrs)
         | [: `("", ":");
              t = poly_type_0
                ? P.error "first_let_binding" "':'" "[poly_type]";
              `("", "=") ? P.error "first_let_binding" "[poly_type]" "'='";
              e = expr_0 ? P.error "first_let_binding" "'='" "[expr]";
              item_attrs = item_attributes_0
                ? P.error "first_let_binding" "[expr]" "[item_attributes]";
              loc = P.gloc bp ! :] ->
             (MLast.PaTyc loc p t, e, item_attrs) ]
         ? P.error "first_let_binding" "[patt]" "'=' or ':'" :] ->
      a ]
and type_extension_0 =
  parser bp
    [: tpl = type_parameters_0;
       n = P.vala longident_lident_0
         ? P.error "type_extension" "[type_parameters]" "[longident_lident]";
       `("", "+=") ? P.error "type_extension" "[longident_lident]" "'+='";
       pf = P.vala (P.bool (P.token ("", "private")))
         ? P.error "type_extension" "'+='" "???";
       _ = P.option (parser [: `("", ("|" as x)) :] -> x);
       ecs =
         P.vala (P.list1sep extension_constructor_0 (P.token ("", "|")) False)
         ? P.error "type_extension" "" "???";
       attrs = item_attributes_0
         ? P.error "type_extension" "???" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      {MLast.teNam = n; tePrm = Ploc.VaVal tpl; tePrv = pf;
       teAttributes = attrs; teECs = ecs}
and rest_type_decl_0 =
  parser bp
    [: `("", "and");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "rest_type_decl" "'and'" "[alg_attributes_no_anti]";
       tpl = type_parameters_0
         ? P.error "rest_type_decl" "[alg_attributes_no_anti]"
             "[type_parameters]";
       n = P.vala type_patt_0
         ? P.error "rest_type_decl" "[type_parameters]" "[type_patt]";
       a =
         parser
         [ [: `("", "=");
              pf = P.vala (P.bool (P.token ("", "private")))
                ? P.error "rest_type_decl" "'='" "???";
              tk = type_kind_0 ? P.error "rest_type_decl" "???" "[type_kind]";
              cl = P.vala (P.list0 constrain_0)
                ? P.error "rest_type_decl" "[type_kind]" "???";
              item_attrs = item_attributes_0
                ? P.error "rest_type_decl" "???" "[item_attributes]";
              loc = P.gloc bp ! :] ->
             let attrs =
               merge_left_auxiliary_attrs ~{nonterm_name = "type_decl"}
                 ~{left_name = "algebraic attributes"}
                 ~{right_name = "item attributes"} alg_attrs item_attrs
             in
             {MLast.tdIsDecl = Ploc.VaVal True; MLast.tdNam = n;
              MLast.tdPrm = Ploc.VaVal tpl; MLast.tdPrv = pf;
              MLast.tdDef = tk; MLast.tdCon = cl; MLast.tdAttributes = attrs}
         | [: `("", ":=");
              pf = P.vala (P.bool (P.token ("", "private")))
                ? P.error "rest_type_decl" "':='" "???";
              tk = type_kind_0 ? P.error "rest_type_decl" "???" "[type_kind]";
              cl = P.vala (P.list0 constrain_0)
                ? P.error "rest_type_decl" "[type_kind]" "???";
              item_attrs = item_attributes_0
                ? P.error "rest_type_decl" "???" "[item_attributes]";
              loc = P.gloc bp ! :] ->
             let attrs =
               merge_left_auxiliary_attrs ~{nonterm_name = "type_decl"}
                 ~{left_name = "algebraic attributes"}
                 ~{right_name = "item attributes"} alg_attrs item_attrs
             in
             {MLast.tdIsDecl = Ploc.VaVal False; MLast.tdNam = n;
              MLast.tdPrm = Ploc.VaVal tpl; MLast.tdPrv = pf;
              MLast.tdDef = tk; MLast.tdCon = cl; MLast.tdAttributes = attrs}
         | [: cl = P.vala (P.list0 constrain_0);
              item_attrs = item_attributes_0
                ? P.error "rest_type_decl" "???" "[item_attributes]";
              loc = P.gloc bp ! :] ->
             let tk = MLast.TyQuo loc (Ploc.VaVal (choose_tvar tpl)) in
             let attrs =
               merge_left_auxiliary_attrs ~{nonterm_name = "type_decl"}
                 ~{left_name = "algebraic attributes"}
                 ~{right_name = "item attributes"} alg_attrs item_attrs
             in
             {MLast.tdIsDecl = Ploc.VaVal True; MLast.tdNam = n;
              MLast.tdPrm = Ploc.VaVal tpl; MLast.tdPrv = Ploc.VaVal False;
              MLast.tdDef = tk; MLast.tdCon = cl;
              MLast.tdAttributes = attrs} ]
         ? P.error "rest_type_decl" "[type_patt]" "'=' or ':=' or ???" :] ->
      a
and constrain_0 =
  parser bp
    [: `("", "constraint");
       t1 = ctyp_0 ? P.error "constrain" "'constraint'" "[ctyp]";
       `("", "=") ? P.error "constrain" "[ctyp]" "'='";
       t2 = ctyp_0 ? P.error "constrain" "'='" "[ctyp]" :] ->
      (t1, t2)
and constructor_declaration_0 =
  parser bp
    [: ci = cons_ident_0;
       (sl, tl, rto, alg_attrs) = rest_constructor_declaration_0
         ? P.error "constructor_declaration" "[cons_ident]"
             "[rest_constructor_declaration]";
       loc = P.gloc bp ! :] ->
      (loc, ci, sl, tl, rto, alg_attrs)
and type_kind_0 =
  parser bp
  [ [: `("", ".."); loc = P.gloc bp ! :] -> MLast.TyOpn loc
  | [: `("", "{");
       ldl = P.vala label_declarations_0
         ? P.error "type_kind" "'{'" "[label_declarations]";
       `("", "}") ? P.error "type_kind" "[label_declarations]" "'}'";
       loc = P.gloc bp ! :] ->
      MLast.TyRec loc ldl
  | [: _ = test_constr_decl; _ = P.option (P.token ("", "|"));
       cdl = P.list0sep constructor_declaration_0 (P.token ("", "|")) False;
       loc = P.gloc bp ! :] ->
      MLast.TySum loc (Ploc.VaVal cdl)
  | [: t = ctyp_0;
       a =
         parser
         [ [: `("", "="); pf = P.bool (P.token ("", "private"));
              a =
                parser
                [ [: `("", "{");
                     ldl = P.vala label_declarations_0
                       ? P.error "type_kind" "'{'" "[label_declarations]";
                     `("", "}")
                       ? P.error "type_kind" "[label_declarations]" "'}'";
                     loc = P.gloc bp ! :] ->
                    MLast.TyMan loc t (Ploc.VaVal pf) (MLast.TyRec loc ldl)
                | [: `("", ".."); loc = P.gloc bp ! :] ->
                    MLast.TyMan loc t (Ploc.VaVal pf) (MLast.TyOpn loc)
                | [: _ = P.option (P.token ("", "|"));
                     cdl =
                       P.list1sep constructor_declaration_0
                         (P.token ("", "|")) False
                       ? P.error "type_kind" "" "[constructor_declaration]";
                     loc = P.gloc bp ! :] ->
                    MLast.TyMan loc t (Ploc.VaVal pf)
                      (MLast.TySum loc (Ploc.VaVal cdl))
                | [: :] -> raise (Stream.Error "") ]
                ? P.error "type_kind" ""
                    "'{' or '..' or '|' or [constructor_declaration]" :] ->
             a
         | [: :] -> t ]
         ? P.error "type_kind" "[ctyp]" "'='" :] ->
      a ]
and type_patt_0 =
  parser bp
    [: n = P.vala (P.token ("LIDENT", "")); loc = P.gloc bp ! :] -> (loc, n)
and first_type_decl_0 =
  parser bp
    [: tpl = type_parameters_0;
       n = P.vala type_patt_0
         ? P.error "first_type_decl" "[type_parameters]" "[type_patt]";
       a =
         parser
         [ [: `("", "=");
              pf = P.vala (P.bool (P.token ("", "private")))
                ? P.error "first_type_decl" "'='" "???";
              tk = type_kind_0
                ? P.error "first_type_decl" "???" "[type_kind]";
              cl = P.vala (P.list0 constrain_0)
                ? P.error "first_type_decl" "[type_kind]" "???";
              attrs = item_attributes_0
                ? P.error "first_type_decl" "???" "[item_attributes]";
              loc = P.gloc bp ! :] ->
             {MLast.tdIsDecl = Ploc.VaVal True; MLast.tdNam = n;
              MLast.tdPrm = Ploc.VaVal tpl; MLast.tdPrv = pf;
              MLast.tdDef = tk; MLast.tdCon = cl; MLast.tdAttributes = attrs}
         | [: `("", ":=");
              pf = P.vala (P.bool (P.token ("", "private")))
                ? P.error "first_type_decl" "':='" "???";
              tk = type_kind_0
                ? P.error "first_type_decl" "???" "[type_kind]";
              cl = P.vala (P.list0 constrain_0)
                ? P.error "first_type_decl" "[type_kind]" "???";
              attrs = item_attributes_0
                ? P.error "first_type_decl" "???" "[item_attributes]";
              loc = P.gloc bp ! :] ->
             {MLast.tdIsDecl = Ploc.VaVal False; MLast.tdNam = n;
              MLast.tdPrm = Ploc.VaVal tpl; MLast.tdPrv = pf;
              MLast.tdDef = tk; MLast.tdCon = cl; MLast.tdAttributes = attrs}
         | [: cl = P.vala (P.list0 constrain_0);
              attrs = item_attributes_0
                ? P.error "first_type_decl" "???" "[item_attributes]";
              loc = P.gloc bp ! :] ->
             let tk = MLast.TyQuo loc (Ploc.VaVal (choose_tvar tpl)) in
             {MLast.tdIsDecl = Ploc.VaVal True; MLast.tdNam = n;
              MLast.tdPrm = Ploc.VaVal tpl; MLast.tdPrv = Ploc.VaVal False;
              MLast.tdDef = tk; MLast.tdCon = cl;
              MLast.tdAttributes = attrs} ]
         ? P.error "first_type_decl" "[type_patt]" "'=' or ':=' or ???" :] ->
      a
and rest_mod_binding_0 =
  parser bp
    [: `("", "and");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "rest_mod_binding" "'and'" "[alg_attributes_no_anti]";
       i = P.vala uidopt_0
         ? P.error "rest_mod_binding" "[alg_attributes_no_anti]" "[uidopt]";
       me = mod_fun_binding_0
         ? P.error "rest_mod_binding" "[uidopt]" "[mod_fun_binding]";
       item_attrs = item_attributes_0
         ? P.error "rest_mod_binding" "[mod_fun_binding]" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      let attrs =
        merge_left_auxiliary_attrs ~{nonterm_name = "mod_binding"}
          ~{left_name = "algebraic attributes"}
          ~{right_name = "item attributes"} alg_attrs item_attrs
      in
      (i, me, attrs)
and mod_fun_binding_0 =
  parser bp
  [ [: `("", ":");
       mt = module_type_0 ? P.error "mod_fun_binding" "':'" "[module_type]";
       `("", "=") ? P.error "mod_fun_binding" "[module_type]" "'='";
       me = module_expr_0 ? P.error "mod_fun_binding" "'='" "[module_expr]";
       loc = P.gloc bp ! :] ->
      MLast.MeTyc loc me mt
  | [: `("", "=");
       me = module_expr_0
         ? P.error "mod_fun_binding" "'='" "[module_expr]" :] ->
      me
  | [: arg = P.vala functor_parameter_0;
       mb = mod_fun_binding_0
         ? P.error "mod_fun_binding" "[functor_parameter]"
             "[mod_fun_binding]";
       loc = P.gloc bp ! :] ->
      MLast.MeFun loc arg mb ]
and first_mod_binding_0 =
  parser bp
    [: i = P.vala uidopt_0;
       me = mod_fun_binding_0
         ? P.error "first_mod_binding" "[uidopt]" "[mod_fun_binding]";
       item_attrs = item_attributes_0
         ? P.error "first_mod_binding" "[mod_fun_binding]"
             "[item_attributes]" :] ->
      (i, me, item_attrs)
and type_binder_opt_0 =
  parser bp
  [ [: _ = check_type_binder;
       ls = P.vala (P.list1 typevar_0)
         ? P.error "type_binder_opt" "[check_type_binder]" "???";
       `("", ".") ? P.error "type_binder_opt" "???" "'.'" :] ->
      ls
  | [: loc = P.gloc bp :] -> Ploc.VaVal [] ]
and label_declaration_0 =
  parser bp
    [: :] ->
      match Stream.npeek 3 strm__ with
      [ [("", "mutable"); ("LIDENT", i); ("", ":")] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          Stream.junk strm__;
          match
            try Some (poly_type_below_alg_attribute_0 strm__) with
            [ Stream.Failure -> None ]
          with
          [ Some t ->
              match
                try Some (alg_attributes_0 strm__) with
                [ Stream.Failure -> None ]
              with
              [ Some attrs ->
                  let loc = P.gloc bp strm__ in
                  (loc, i, True, t, attrs)
              | _ ->
                  raise
                    (Stream.Error
                       (P.error "label_declaration"
                          "[poly_type_below_alg_attribute]"
                          "[alg_attributes]")) ]
          | _ ->
              raise
                (Stream.Error
                   (P.error "label_declaration" "'mutable'"
                      "[poly_type_below_alg_attribute]")) ]
        }
      | _ ->
          match Stream.npeek 2 strm__ with
          [ [("LIDENT", i); ("", ":")] -> do {
              Stream.junk strm__;
              Stream.junk strm__;
              match
                try Some (poly_type_below_alg_attribute_0 strm__) with
                [ Stream.Failure -> None ]
              with
              [ Some t ->
                  match
                    try Some (alg_attributes_0 strm__) with
                    [ Stream.Failure -> None ]
                  with
                  [ Some attrs ->
                      let loc = P.gloc bp strm__ in
                      (loc, i, False, t, attrs)
                  | _ ->
                      raise
                        (Stream.Error
                           (P.error "label_declaration"
                              "[poly_type_below_alg_attribute]"
                              "[alg_attributes]")) ]
              | _ ->
                  raise
                    (Stream.Error
                       (P.error "label_declaration" "lowercase identifier"
                          "[poly_type_below_alg_attribute]")) ]
            }
          | _ -> raise Stream.Failure ] ]
and label_declarations_0 =
  parser bp
    [: (a, b, c, d, attrs1) = label_declaration_0;
       a =
         parser
         [ [: `("", ";");
              attrs2 = alg_attributes_no_anti_0
                ? P.error "label_declarations" "';'"
                    "[alg_attributes_no_anti]";
              a =
                parser
                [ [: ldl = label_declarations_0; loc = P.gloc bp ! :] ->
                    let attrs =
                      merge_right_auxiliary_attrs
                        ~{nonterm_name = "label_declarations"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "algebraic attributes"} attrs1 attrs2
                    in
                    [(a, b, c, d, attrs) :: ldl]
                | [: loc = P.gloc bp :] ->
                    let attrs =
                      merge_right_auxiliary_attrs
                        ~{nonterm_name = "label_declarations"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "algebraic attributes"} attrs1 attrs2
                    in
                    [(a, b, c, d, attrs)] ]
                ? P.error "label_declarations" "[alg_attributes_no_anti]"
                    "[label_declarations]" :] ->
             a
         | [: :] -> [(a, b, c, d, attrs1)] ]
         ? P.error "label_declarations" "[label_declaration]" "';'" :] ->
      a
and record_type_0 =
  parser bp
    [: `("", "{");
       ldl = P.vala label_declarations_0
         ? P.error "record_type" "'{'" "[label_declarations]";
       `("", "}") ? P.error "record_type" "[label_declarations]" "'}'";
       loc = P.gloc bp ! :] ->
      MLast.TyRec loc ldl
and rest_constructor_declaration_0 =
  parser bp
  [ [: `("", "of");
       a =
         parser
         [ [: cal = P.vala (P.list1sep ctyp_6 (P.token ("", "*")) False);
              alg_attrs = alg_attributes_0
                ? P.error "rest_constructor_declaration" "???"
                    "[alg_attributes]";
              loc = P.gloc bp ! :] ->
             (Ploc.VaVal [], cal, Ploc.VaVal None, alg_attrs)
         | [: cdrec = record_type_0;
              alg_attrs = alg_attributes_0
                ? P.error "rest_constructor_declaration" "[record_type]"
                    "[alg_attributes]";
              loc = P.gloc bp ! :] ->
             (Ploc.VaVal [], Ploc.VaVal [cdrec], Ploc.VaVal None, alg_attrs) ]
         ? P.error "rest_constructor_declaration" "'of'"
             "??? or [record_type]" :] ->
      a
  | [: `("", ":");
       a =
         parser
         [ [: _ = check_type_binder;
              ls = P.vala (P.list1 typevar_0)
                ? P.error "rest_constructor_declaration" "[check_type_binder]"
                    "???";
              `("", ".") ? P.error "rest_constructor_declaration" "???" "'.'";
              cal = P.vala (P.list1sep ctyp_6 (P.token ("", "*")) False)
                ? P.error "rest_constructor_declaration" "'.'" "???";
              `("", "->")
                ? P.error "rest_constructor_declaration" "???" "'->'";
              t = ctyp_0
                ? P.error "rest_constructor_declaration" "'->'" "[ctyp]";
              alg_attrs = alg_attributes_0
                ? P.error "rest_constructor_declaration" "[ctyp]"
                    "[alg_attributes]";
              loc = P.gloc bp ! :] ->
             (ls, cal, Ploc.VaVal (Some t), alg_attrs)
         | [: cal = P.vala (P.list1sep ctyp_6 (P.token ("", "*")) False);
              a =
                parser
                [ [: `("", "->");
                     t = ctyp_0
                       ? P.error "rest_constructor_declaration" "'->'"
                           "[ctyp]";
                     alg_attrs = alg_attributes_0
                       ? P.error "rest_constructor_declaration" "[ctyp]"
                           "[alg_attributes]";
                     loc = P.gloc bp ! :] ->
                    (Ploc.VaVal [], cal, Ploc.VaVal (Some t), alg_attrs)
                | [: alg_attrs = alg_attributes_0; loc = P.gloc bp ! :] ->
                    let t =
                      match cal with
                      [ Ploc.VaVal [t] -> t
                      | Ploc.VaVal [t :: tl] ->
                          MLast.TyTup loc (Ploc.VaVal [t :: tl])
                      | _ -> assert False ]
                    in
                    (Ploc.VaVal [], Ploc.VaVal [], Ploc.VaVal (Some t),
                     alg_attrs) ]
                ? P.error "rest_constructor_declaration" "???"
                    "'->' or [alg_attributes]" :] ->
             a
         | [: cdrec = record_type_0;
              `("", "->")
                ? P.error "rest_constructor_declaration" "[record_type]"
                    "'->'";
              t = ctyp_0
                ? P.error "rest_constructor_declaration" "'->'" "[ctyp]";
              alg_attrs = alg_attributes_0
                ? P.error "rest_constructor_declaration" "[ctyp]"
                    "[alg_attributes]";
              loc = P.gloc bp ! :] ->
             (Ploc.VaVal [], Ploc.VaVal [cdrec], Ploc.VaVal (Some t),
              alg_attrs) ]
         ? P.error "rest_constructor_declaration" "':'"
             "[check_type_binder] or ??? or [record_type]" :] ->
      a
  | [: alg_attrs = alg_attributes_0; loc = P.gloc bp ! :] ->
      (Ploc.VaVal [], Ploc.VaVal [], Ploc.VaVal None, alg_attrs) ]
and cons_ident_0 =
  parser bp
  [ [: `("UIDENT", "True"); loc = P.gloc bp ! :] -> Ploc.VaVal "True_"
  | [: `("UIDENT", "False"); loc = P.gloc bp ! :] -> Ploc.VaVal "False_"
  | [: `("", "true"); loc = P.gloc bp ! :] -> Ploc.VaVal "True"
  | [: `("", "false"); loc = P.gloc bp ! :] -> Ploc.VaVal "False"
  | [: :] ->
      match Stream.npeek 2 strm__ with
      [ [("", "["); ("", "]")] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          try
            let loc = P.gloc bp strm__ in
            Ploc.VaVal "[]"
          with
          [ Stream.Failure ->
              raise (Stream.Error (P.error "cons_ident" "'['" "???")) ]
        }
      | _ ->
          match strm__ with parser
          [ [: `("", "(");
               a =
                 parser
                 [ [: `("", ")"); loc = P.gloc bp ! :] -> Ploc.VaVal "()"
                 | [: :] ->
                     match Stream.npeek 2 strm__ with
                     [ [("", "::"); ("", ")")] -> do {
                         Stream.junk strm__;
                         Stream.junk strm__;
                         try
                           let loc = P.gloc bp strm__ in
                           Ploc.VaVal "::"
                         with
                         [ Stream.Failure ->
                             raise
                               (Stream.Error
                                  (P.error "cons_ident" "'::'" "???")) ]
                       }
                     | _ -> raise Stream.Failure ] ]
                 ? P.error "cons_ident" "'('" "')' or '::'" :] ->
              a
          | [: i = P.vala (P.token ("UIDENT", "")) :] -> i ] ] ]
and extension_constructor_0 =
  parser bp
    [: attrs = alg_attributes_no_anti_0;
       ci = cons_ident_0
         ? P.error "extension_constructor" "[alg_attributes_no_anti]"
             "[cons_ident]";
       a =
         parser
         [ [: `("", "=");
              b = P.vala longident_0
                ? P.error "extension_constructor" "'='" "[longident]";
              alg_attrs = alg_attributes_0
                ? P.error "extension_constructor" "[longident]"
                    "[alg_attributes]";
              loc = P.gloc bp ! :] ->
             let alg_attrs =
               merge_left_auxiliary_attrs
                 ~{nonterm_name = "extension_constructor"}
                 ~{left_name = "algebraic attributes"}
                 ~{right_name = "item attributes"} attrs alg_attrs
             in
             MLast.EcRebind loc ci b alg_attrs
         | [: (sl, tl, rto, alg_attrs) = rest_constructor_declaration_0;
              loc = P.gloc bp ! :] ->
             let alg_attrs =
               merge_left_auxiliary_attrs
                 ~{nonterm_name = "extension_constructor"}
                 ~{left_name = "algebraic attributes"}
                 ~{right_name = "(right) algebraic attributes"} attrs
                 alg_attrs
             in
             MLast.EcTuple loc (loc, ci, sl, tl, rto, alg_attrs) ]
         ? P.error "extension_constructor" "[cons_ident]"
             "'=' or [rest_constructor_declaration]" :] ->
      a
and class_type_0 =
  parser bp
    [: :] ->
      match Stream.npeek 3 strm__ with
      [ [("", "?"); ("LIDENT", i); ("", ":")] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          Stream.junk strm__;
          let i = Ploc.VaVal i in
          match try Some (ctyp_6 strm__) with [ Stream.Failure -> None ] with
          [ Some t ->
              match strm__ with parser
              [ [: `("", "->");
                   ct = class_type_0
                     ? P.error "class_type" "'->'" "[class_type]";
                   loc = P.gloc bp ! :] ->
                  MLast.CtFun loc (MLast.TyOlb loc i t) ct
              | [: :] ->
                  raise
                    (Stream.Error
                       (P.error "class_type" "[ctyp level apply]" "'->'")) ]
          | _ ->
              raise
                (Stream.Error
                   (P.error "class_type" "'?'" "[ctyp level apply]")) ]
        }
      | _ ->
          match Stream.npeek 2 strm__ with
          [ [("LIDENT", i); ("", ":")] -> do {
              Stream.junk strm__;
              Stream.junk strm__;
              match
                try Some (ctyp_6 strm__) with [ Stream.Failure -> None ]
              with
              [ Some t ->
                  match strm__ with parser
                  [ [: `("", "->");
                       ct = class_type_0
                         ? P.error "class_type" "'->'" "[class_type]";
                       loc = P.gloc bp ! :] ->
                      MLast.CtFun loc (MLast.TyLab loc (Ploc.VaVal i) t) ct
                  | [: :] ->
                      raise
                        (Stream.Error
                           (P.error "class_type" "[ctyp level apply]"
                              "'->'")) ]
              | _ ->
                  raise
                    (Stream.Error
                       (P.error "class_type" "lowercase identifier"
                          "[ctyp level apply]")) ]
            }
          | _ ->
              match
                try
                  Some (P.vala (P.token ("QUESTIONIDENTCOLON", "")) strm__)
                with
                [ Stream.Failure -> None ]
              with
              [ Some i ->
                  match
                    try Some (ctyp_6 strm__) with [ Stream.Failure -> None ]
                  with
                  [ Some t ->
                      match strm__ with parser
                      [ [: `("", "->");
                           ct = class_type_0
                             ? P.error "class_type" "'->'" "[class_type]";
                           loc = P.gloc bp ! :] ->
                          MLast.CtFun loc (MLast.TyOlb loc i t) ct
                      | [: :] ->
                          raise
                            (Stream.Error
                               (P.error "class_type" "[ctyp level apply]"
                                  "'->'")) ]
                  | _ ->
                      raise
                        (Stream.Error
                           (P.error "class_type" "QUESTIONIDENTCOLON"
                              "[ctyp level apply]")) ]
              | _ ->
                  match Stream.npeek 2 strm__ with
                  [ [("QUESTIONIDENT", i); ("", ":")] -> do {
                      Stream.junk strm__;
                      Stream.junk strm__;
                      let i = Ploc.VaVal i in
                      match
                        try Some (ctyp_6 strm__) with
                        [ Stream.Failure -> None ]
                      with
                      [ Some t ->
                          match strm__ with parser
                          [ [: `("", "->");
                               ct = class_type_0
                                 ? P.error "class_type" "'->'" "[class_type]";
                               loc = P.gloc bp ! :] ->
                              MLast.CtFun loc (MLast.TyOlb loc i t) ct
                          | [: :] ->
                              raise
                                (Stream.Error
                                   (P.error "class_type" "[ctyp level apply]"
                                      "'->'")) ]
                      | _ ->
                          raise
                            (Stream.Error
                               (P.error "class_type" "QUESTIONIDENT"
                                  "[ctyp level apply]")) ]
                    }
                  | _ ->
                      match
                        try Some (test_ctyp_minusgreater strm__) with
                        [ Stream.Failure -> None ]
                      with
                      [ Some _ ->
                          match
                            try Some (ctyp_5 strm__) with
                            [ Stream.Failure -> None ]
                          with
                          [ Some t ->
                              match strm__ with parser
                              [ [: `("", "->");
                                   ct = class_type_0
                                     ? P.error "class_type" "'->'"
                                         "[class_type]";
                                   loc = P.gloc bp ! :] ->
                                  MLast.CtFun loc t ct
                              | [: :] ->
                                  raise
                                    (Stream.Error
                                       (P.error "class_type"
                                          "[ctyp level star]" "'->'")) ]
                          | _ ->
                              raise
                                (Stream.Error
                                   (P.error "class_type"
                                      "[test_ctyp_minusgreater]"
                                      "[ctyp level star]")) ]
                      | _ ->
                          match
                            try Some (class_signature_0 strm__) with
                            [ Stream.Failure -> None ]
                          with
                          [ Some cs -> cs
                          | _ -> raise Stream.Failure ] ] ] ] ] ]
and item_extension_0 =
  parser bp
  [ [: `("", "[%%");
       e = P.vala attribute_body_0
         ? P.error "item_extension" "'[%%'" "[attribute_body]";
       `("", "]") ? P.error "item_extension" "[attribute_body]" "']'" :] ->
      e
  | [: `("QUOTEDEXTENSION", s); loc = P.gloc bp ! :] ->
      Ploc.VaVal (make_string_extension loc s) ]
and floating_attribute_0 =
  parser bp
    [: `("", "[@@@");
       attr = P.vala attribute_body_0
         ? P.error "floating_attribute" "'[@@@'" "[attribute_body]";
       `("", "]")
         ? P.error "floating_attribute" "[attribute_body]" "']'" :] ->
      attr
and priv_virt_0 =
  parser bp
  [ [: `("", "private");
       a =
         parser
         [ [: `("", "virtual") :] -> (True, True)
         | [: :] -> (True, False) ]
         ? P.error "priv_virt" "'private'" "'virtual'" :] ->
      a
  | [: `("", "virtual");
       a =
         parser
         [ [: `("", "private") :] -> (True, True)
         | [: :] -> (False, True) ]
         ? P.error "priv_virt" "'virtual'" "'private'" :] ->
      a
  | [: :] -> (False, False) ]
and cvalue_binding_or_ctyp_0 =
  parser bp
  [ [: `("", "=");
       e = expr_0 ? P.error "cvalue_binding_or_ctyp" "'='" "[expr]" :] ->
      Left e
  | [: `("", ":");
       t = ctyp_0 ? P.error "cvalue_binding_or_ctyp" "':'" "[ctyp]";
       a =
         parser
         [ [: `("", "=");
              e = expr_0 ? P.error "cvalue_binding_or_ctyp" "'='" "[expr]";
              loc = P.gloc bp ! :] ->
             Left (MLast.ExTyc loc e t)
         | [: `("", ":>");
              t2 = ctyp_0 ? P.error "cvalue_binding_or_ctyp" "':>'" "[ctyp]";
              `("", "=") ? P.error "cvalue_binding_or_ctyp" "[ctyp]" "'='";
              e = expr_0 ? P.error "cvalue_binding_or_ctyp" "'='" "[expr]";
              loc = P.gloc bp ! :] ->
             Left (MLast.ExCoe loc e (Some t) t2)
         | [: :] -> Right t ]
         ? P.error "cvalue_binding_or_ctyp" "[ctyp]" "'=' or ':>'" :] ->
      a
  | [: `("", ":>");
       t = ctyp_0 ? P.error "cvalue_binding_or_ctyp" "':>'" "[ctyp]";
       `("", "=") ? P.error "cvalue_binding_or_ctyp" "[ctyp]" "'='";
       e = expr_0 ? P.error "cvalue_binding_or_ctyp" "'='" "[expr]";
       loc = P.gloc bp ! :] ->
      Left (MLast.ExCoe loc e None t) ]
and mut_virt_0 =
  parser bp
  [ [: `("", "mutable");
       a =
         parser
         [ [: `("", "virtual") :] -> (True, True)
         | [: :] -> (True, False) ]
         ? P.error "mut_virt" "'mutable'" "'virtual'" :] ->
      a
  | [: `("", "virtual");
       a =
         parser
         [ [: `("", "mutable") :] -> (True, True)
         | [: :] -> (False, True) ]
         ? P.error "mut_virt" "'virtual'" "'mutable'" :] ->
      a
  | [: :] -> (False, False) ]
and class_str_item_0 =
  parser bp
  [ [: `("", "inherit");
       ovf = P.vala (P.bool (P.token ("", "!")))
         ? P.error "class_str_item" "'inherit'" "???";
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "class_str_item" "???" "[alg_attributes_no_anti]";
       ce = class_expr_0
         ? P.error "class_str_item" "[alg_attributes_no_anti]" "[class_expr]";
       pb =
         P.vala
           (P.option
              (parser
                 [: :] ->
                   match Stream.npeek 2 strm__ with
                   [ [("", "as"); ("LIDENT", i)] -> do {
                       Stream.junk strm__;
                       Stream.junk strm__;
                       try i with
                       [ Stream.Failure ->
                           raise
                             (Stream.Error (P.error "implem" "'as'" "???")) ]
                     }
                   | _ -> raise Stream.Failure ]))
         ? P.error "class_str_item" "[class_expr]" "???";
       item_attrs = item_attributes_0
         ? P.error "class_str_item" "???" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      let attrs =
        merge_left_auxiliary_attrs ~{nonterm_name = "cstr-inherit"}
          ~{left_name = "algebraic attributes"}
          ~{right_name = "item attributes"} alg_attrs item_attrs
      in
      MLast.CrInh loc ovf ce pb attrs
  | [: `("", "val");
       ov = P.vala (P.bool (P.token ("", "!")))
         ? P.error "class_str_item" "'val'" "???";
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "class_str_item" "???" "[alg_attributes_no_anti]";
       (mf, vf) = mut_virt_0
         ? P.error "class_str_item" "[alg_attributes_no_anti]" "[mut_virt]";
       lab = P.vala (P.token ("LIDENT", ""))
         ? P.error "class_str_item" "[mut_virt]" "lowercase identifier";
       e = cvalue_binding_or_ctyp_0
         ? P.error "class_str_item" "lowercase identifier"
             "[cvalue_binding_or_ctyp]";
       item_attrs = item_attributes_0
         ? P.error "class_str_item" "[cvalue_binding_or_ctyp]"
             "[item_attributes]";
       loc = P.gloc bp ! :] ->
      let attrs =
        merge_left_auxiliary_attrs ~{nonterm_name = "cstr-inherit"}
          ~{left_name = "algebraic attributes"}
          ~{right_name = "item attributes"} alg_attrs item_attrs
      in
      match (vf, e) with
      [ (False, Left e) -> MLast.CrVal loc ov (Ploc.VaVal mf) lab e attrs
      | (True, Left _) ->
          Ploc.raise loc
            (Stream.Error "val with definition cannot be virtual")
      | (False, Right _) ->
          Ploc.raise loc
            (Stream.Error "val without definition must be virtual")
      | (True, Right t) ->
          if Pcaml.unvala ov then
            Ploc.raise loc (Stream.Error "virtual value cannot override")
          else MLast.CrVav loc (Ploc.VaVal mf) lab t attrs ]
  | [: `("", "method");
       ov = P.vala (P.bool (P.token ("", "!")))
         ? P.error "class_str_item" "'method'" "???";
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "class_str_item" "???" "[alg_attributes_no_anti]";
       (pf, vf) = priv_virt_0
         ? P.error "class_str_item" "[alg_attributes_no_anti]" "[priv_virt]";
       l = P.vala (P.token ("LIDENT", ""))
         ? P.error "class_str_item" "[priv_virt]" "lowercase identifier";
       a =
         parser
         [ [: `("", ":");
              t = poly_type_below_alg_attribute_0
                ? P.error "class_str_item" "':'"
                    "[poly_type_below_alg_attribute]";
              a =
                parser
                [ [: `("", "=");
                     e = expr_0 ? P.error "class_str_item" "'='" "[expr]";
                     item_attrs = item_attributes_0
                       ? P.error "class_str_item" "[expr]"
                           "[item_attributes]";
                     loc = P.gloc bp ! :] ->
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "cstr-inherit"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    MLast.CrMth loc ov (Ploc.VaVal pf) l (Ploc.VaVal (Some t))
                      e attrs
                | [: item_attrs = item_attributes_0; loc = P.gloc bp ! :] ->
                    if Pcaml.unvala ov then
                      Ploc.raise loc
                        (Stream.Error
                           "method without definition is not being overriden!")
                    else if not vf then
                      Ploc.raise loc
                        (Stream.Error
                           "method without definition must be virtual")
                    else
                      let attrs =
                        merge_left_auxiliary_attrs
                          ~{nonterm_name = "cstr-inherit"}
                          ~{left_name = "algebraic attributes"}
                          ~{right_name = "item attributes"} alg_attrs
                          item_attrs
                      in
                      MLast.CrVir loc (Ploc.VaVal pf) l t attrs ]
                ? P.error "class_str_item" "[poly_type_below_alg_attribute]"
                    "'=' or [item_attributes]" :] ->
             a
         | [: sb = fun_binding_0;
              item_attrs = item_attributes_0
                ? P.error "class_str_item" "[fun_binding]"
                    "[item_attributes]";
              loc = P.gloc bp ! :] ->
             let attrs =
               merge_left_auxiliary_attrs ~{nonterm_name = "cstr-inherit"}
                 ~{left_name = "algebraic attributes"}
                 ~{right_name = "item attributes"} alg_attrs item_attrs
             in
             MLast.CrMth loc ov (Ploc.VaVal pf) l (Ploc.VaVal None) sb attrs ]
         ? P.error "class_str_item" "lowercase identifier"
             "':' or [fun_binding]" :] ->
      a
  | [: `("", "constraint");
       t1 = ctyp_0 ? P.error "class_str_item" "'constraint'" "[ctyp]";
       `("", "=") ? P.error "class_str_item" "[ctyp]" "'='";
       t2 = ctyp_0 ? P.error "class_str_item" "'='" "[ctyp]";
       attrs = item_attributes_0
         ? P.error "class_str_item" "[ctyp]" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      MLast.CrCtr loc t1 t2 attrs
  | [: `("", "initializer");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "class_str_item" "'initializer'"
             "[alg_attributes_no_anti]";
       se = expr_0
         ? P.error "class_str_item" "[alg_attributes_no_anti]" "[expr]";
       item_attrs = item_attributes_0
         ? P.error "class_str_item" "[expr]" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      let attrs =
        merge_left_auxiliary_attrs ~{nonterm_name = "cstr-inherit"}
          ~{left_name = "algebraic attributes"}
          ~{right_name = "item attributes"} alg_attrs item_attrs
      in
      MLast.CrIni loc se attrs
  | [: attr = floating_attribute_0; loc = P.gloc bp ! :] ->
      MLast.CrFlAtt loc attr
  | [: e = item_extension_0; loc = P.gloc bp ! :] -> MLast.CrExten loc e ]
and class_structure_0 =
  parser bp
  [ [: cf = P.list0 class_str_item_0 :] -> cf
  | [: :] -> raise (Stream.Error "") ]
and class_self_patt_0 =
  parser bp
    [: `("", "("); p = patt_0 ? P.error "class_self_patt" "'('" "[patt]";
       a =
         parser
         [ [: `("", ")") :] -> p
         | [: `("", ":");
              t = ctyp_0 ? P.error "class_self_patt" "':'" "[ctyp]";
              `("", ")") ? P.error "class_self_patt" "[ctyp]" "')'";
              loc = P.gloc bp ! :] ->
             MLast.PaTyc loc p t ]
         ? P.error "class_self_patt" "[patt]" "')' or ':'" :] ->
      a
and class_expr_simple_0 =
  parser bp
  [ [: `("", "["); ct = ctyp_0 ? P.error "class_expr_simple" "'['" "[ctyp]";
       a =
         parser
         [ [: `("", ",");
              ctcl = P.list1sep ctyp_0 (P.token ("", ",")) False
                ? P.error "class_expr_simple" "','" "[ctyp]";
              `("", "]") ? P.error "class_expr_simple" "" "[ctyp] or ']'";
              cli = P.vala longident_lident_0
                ? P.error "class_expr_simple" "']'" "[longident_lident]";
              loc = P.gloc bp ! :] ->
             MLast.CeCon loc cli (Ploc.VaVal [ct :: ctcl])
         | [: `("", "]");
              cli = P.vala longident_lident_0
                ? P.error "class_expr_simple" "']'" "[longident_lident]";
              loc = P.gloc bp ! :] ->
             MLast.CeCon loc cli (Ploc.VaVal [ct]) ]
         ? P.error "class_expr_simple" "[ctyp]" "',' or ']'" :] ->
      a
  | [: `("", "object");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "class_expr_simple" "'object'" "[alg_attributes_no_anti]";
       cspo = P.vala (P.option class_self_patt_0)
         ? P.error "class_expr_simple" "[alg_attributes_no_anti]" "???";
       cf = P.vala class_structure_0
         ? P.error "class_expr_simple" "???" "[class_structure]";
       `("", "end") ? P.error "class_expr_simple" "[class_structure]" "'end'";
       loc = P.gloc bp ! :] ->
      class_expr_wrap_attrs (MLast.CeStr loc cspo cf) alg_attrs
  | [: `("", "(");
       ce = class_expr_0 ? P.error "class_expr_simple" "'('" "[class_expr]";
       a =
         parser
         [ [: `("", ":");
              ct = class_type_0
                ? P.error "class_expr_simple" "':'" "[class_type]";
              `("", ")") ? P.error "class_expr_simple" "[class_type]" "')'";
              loc = P.gloc bp ! :] ->
             MLast.CeTyc loc ce ct
         | [: `("", ")") :] -> ce ]
         ? P.error "class_expr_simple" "[class_expr]" "':' or ')'" :] ->
      a
  | [: cli = P.vala longident_lident_0; loc = P.gloc bp ! :] ->
      MLast.CeCon loc cli (Ploc.VaVal []) ]
and class_expr_apply_0 =
  parser bp
    [: a = class_expr_apply_1; a = class_expr_apply_gen_cont bp a ! :] -> a
and class_expr_apply_0_cont bp a__ strm__ =
  match try Some (expr_15 strm__) with [ Stream.Failure -> None ] with
  [ Some e ->
      let loc = P.gloc bp strm__ in
      let ce = a__ in
      class_expr_apply_0_cont bp (MLast.CeApp loc ce e) strm__
  | _ -> a__ ]
and class_expr_apply_1 = parser bp [: ce = class_expr_simple_0 :] -> ce
and class_expr_apply_gen_cont bp a__ strm__ =
  let x = class_expr_apply_0_cont bp a__ strm__ in
  if x == a__ then x else class_expr_apply_gen_cont bp x strm__
and alg_extension_0 =
  parser bp
  [ [: `("", "[%");
       e = P.vala attribute_body_0
         ? P.error "alg_extension" "'[%'" "[attribute_body]";
       `("", "]") ? P.error "alg_extension" "[attribute_body]" "']'" :] ->
      e
  | [: `("QUOTEDEXTENSION", s); loc = P.gloc bp ! :] ->
      Ploc.VaVal (make_string_extension loc s) ]
and item_attribute_0 =
  parser bp
    [: `("", "[@@");
       attr = P.vala attribute_body_0
         ? P.error "item_attribute" "'[@@'" "[attribute_body]";
       `("", "]") ? P.error "item_attribute" "[attribute_body]" "']'" :] ->
      attr
and item_attributes_0 =
  parser bp [: l = P.vala (P.list0 item_attribute_0) :] -> l
and poly_type_0 =
  parser bp
  [ [: `("", "type");
       nt = P.list1 (P.token ("LIDENT", ""))
         ? P.error "poly_type" "'type'" "lowercase identifier";
       `("", ".") ? P.error "poly_type" "" "lowercase identifier or '.'";
       ct = ctyp_0 ? P.error "poly_type" "'.'" "[ctyp]";
       loc = P.gloc bp ! :] ->
      MLast.TyPot loc (Ploc.VaVal nt) ct
  | [: _ = test_typevar_list_dot;
       tpl = P.list1 typevar_0
         ? P.error "poly_type" "[test_typevar_list_dot]" "[typevar]";
       `("", ".") ? P.error "poly_type" "" "[typevar] or '.'";
       t2 = ctyp_0 ? P.error "poly_type" "'.'" "[ctyp]";
       loc = P.gloc bp ! :] ->
      MLast.TyPol loc (Ploc.VaVal tpl) t2
  | [: t = ctyp_0 :] -> t ]
and fun_binding_0 =
  parser bp
  [ [: `("", "="); e = expr_0 ? P.error "fun_binding" "'='" "[expr]" :] -> e
  | [: `("", ":");
       t = poly_type_0 ? P.error "fun_binding" "':'" "[poly_type]";
       a =
         parser
         [ [: `("", "="); e = expr_0 ? P.error "fun_binding" "'='" "[expr]";
              loc = P.gloc bp ! :] ->
             MLast.ExTyc loc e t
         | [: `("", ":>");
              t2 = poly_type_0 ? P.error "fun_binding" "':>'" "[poly_type]";
              `("", "=") ? P.error "fun_binding" "[poly_type]" "'='";
              e = expr_0 ? P.error "fun_binding" "'='" "[expr]";
              loc = P.gloc bp ! :] ->
             MLast.ExCoe loc (MLast.ExTyc loc e t) None t2 ]
         ? P.error "fun_binding" "[poly_type]" "'=' or ':>'" :] ->
      a
  | [: `("", ":>");
       t = poly_type_0 ? P.error "fun_binding" "':>'" "[poly_type]";
       `("", "=") ? P.error "fun_binding" "[poly_type]" "'='";
       e = expr_0 ? P.error "fun_binding" "'='" "[expr]";
       loc = P.gloc bp ! :] ->
      MLast.ExCoe loc e None t
  | [: p = labeled_patt_0;
       e = fun_binding_0
         ? P.error "fun_binding" "[labeled_patt]" "[fun_binding]";
       loc = P.gloc bp ! :] ->
      MLast.ExFun loc (Ploc.VaVal [(p, Ploc.VaVal None, e)])
  | [: _ = check_new_type_extended :] ->
      match Stream.npeek 2 strm__ with
      [ [("", "("); ("", "type")] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          match
            try Some (P.list1 (P.token ("LIDENT", "")) strm__) with
            [ Stream.Failure -> None ]
          with
          [ Some l ->
              match strm__ with parser
              [ [: `("", ")");
                   e = fun_binding_0
                     ? P.error "fun_binding" "')'" "[fun_binding]";
                   loc = P.gloc bp ! :] ->
                  List.fold_right
                    (fun id e ->
                       MLast.ExFun loc
                         (Ploc.VaVal
                            [(MLast.PaNty loc (Ploc.VaVal id),
                              Ploc.VaVal None, e)]))
                    l e
              | [: :] ->
                  raise
                    (Stream.Error
                       (P.error "fun_binding" ""
                          "lowercase identifier or ')'")) ]
          | _ ->
              raise
                (Stream.Error
                   (P.error "fun_binding" "'('" "lowercase identifier")) ]
        }
      | _ ->
          raise
            (Stream.Error
               (P.error "fun_binding" "[check_new_type_extended]" "'('")) ]
  | [: _ = check_not_new_type_extended;
       p = patt_8
         ? P.error "fun_binding" "[check_not_new_type_extended]"
             "[patt level simple]";
       e = fun_binding_0
         ? P.error "fun_binding" "[patt level simple]" "[fun_binding]";
       loc = P.gloc bp ! :] ->
      MLast.ExFun loc (Ploc.VaVal [(p, Ploc.VaVal None, e)]) ]
and val_ident_0 =
  parser bp
    [: _ = check_not_part_of_patt;
       a =
         parser
         [ [: `("", "(");
              op = operator_rparen
                ? P.error "val_ident" "'('" "[operator_rparen]";
              loc = P.gloc bp ! :] ->
             if op = "::" then
               MLast.PaLong loc (MLast.LiUid loc (Ploc.VaVal op))
                 (Ploc.VaVal [])
             else MLast.PaLid loc (Ploc.VaVal op)
         | [: `("LIDENT", s); loc = P.gloc bp ! :] ->
             MLast.PaLid loc (Ploc.VaVal s) ]
         ? P.error "val_ident" "[check_not_part_of_patt]"
             "'(' or lowercase identifier" :] ->
      a
and let_binding_0 =
  parser bp
    [: alg_attrs = alg_attributes_no_anti_0;
       a =
         parser
         [ [: p = val_ident_0;
              a =
                parser
                [ [: _ = check_and_or_in; loc = P.gloc bp ! :] ->
                    match p with
                    [ MLast.PaLid _ (Ploc.VaVal s) ->
                        (p, MLast.ExLid loc (Ploc.VaVal s),
                         Ploc.VaVal alg_attrs)
                    | _ ->
                        failwith
                          "let punning: binder must be a lowercase ident (variable)" ]
                | [: _ = check_colon;
                     e = fun_binding_0
                       ? P.error "let_binding" "[check_colon]"
                           "[fun_binding]";
                     item_attrs = item_attributes_0
                       ? P.error "let_binding" "[fun_binding]"
                           "[item_attributes]";
                     loc = P.gloc bp ! :] ->
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "let_binding"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    let (p, e) =
                      match e with
                      [ MLast.ExTyc _ e t -> (MLast.PaTyc loc p t, e)
                      | _ -> (p, e) ]
                    in
                    (p, e, attrs)
                | [: _ = check_not_colon;
                     e = fun_binding_0
                       ? P.error "let_binding" "[check_not_colon]"
                           "[fun_binding]";
                     item_attrs = item_attributes_0
                       ? P.error "let_binding" "[fun_binding]"
                           "[item_attributes]";
                     loc = P.gloc bp ! :] ->
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "let_binding"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    (p, e, attrs) ]
                ? P.error "let_binding" "[val_ident]"
                    "[check_and_or_in] or [check_colon] or [check_not_colon]" :] ->
             a
         | [: p = patt_0;
              a =
                parser
                [ [: `("", "=");
                     e = expr_0 ? P.error "let_binding" "'='" "[expr]";
                     item_attrs = item_attributes_0
                       ? P.error "let_binding" "[expr]" "[item_attributes]";
                     loc = P.gloc bp ! :] ->
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "let_binding"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    (p, e, attrs)
                | [: `("", ":");
                     t = poly_type_0
                       ? P.error "let_binding" "':'" "[poly_type]";
                     `("", "=") ? P.error "let_binding" "[poly_type]" "'='";
                     e = expr_0 ? P.error "let_binding" "'='" "[expr]";
                     item_attrs = item_attributes_0
                       ? P.error "let_binding" "[expr]" "[item_attributes]";
                     loc = P.gloc bp ! :] ->
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "let_binding"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    (p, MLast.ExTyc loc e t, attrs) ]
                ? P.error "let_binding" "[patt]" "'=' or ':'" :] ->
             a ]
         ? P.error "let_binding" "[alg_attributes_no_anti]"
             "[val_ident] or [patt]" :] ->
      a
and labeled_patt_0 =
  parser bp
    [: :] ->
      match Stream.npeek 3 strm__ with
      [ [("", "~"); ("", "("); ("LIDENT", i)] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          Stream.junk strm__;
          try
            match strm__ with parser
            [ [: `("", ")"); loc = P.gloc bp ! :] ->
                MLast.PaLab loc
                  (Ploc.VaVal
                     [(MLast.PaLid loc (Ploc.VaVal i), Ploc.VaVal None)])
            | [: `("", ":");
                 t = ctyp_0 ? P.error "labeled_patt" "':'" "[ctyp]";
                 `("", ")") ? P.error "labeled_patt" "[ctyp]" "')'";
                 loc = P.gloc bp ! :] ->
                MLast.PaLab loc
                  (Ploc.VaVal
                     [(MLast.PaTyc loc (MLast.PaLid loc (Ploc.VaVal i)) t,
                       Ploc.VaVal None)]) ]
          with
          [ Stream.Failure ->
              raise
                (Stream.Error (P.error "labeled_patt" "'~'" "')' or ':'")) ]
        }
      | _ ->
          match Stream.npeek 3 strm__ with
          [ [("", "?"); ("", "("); ("LIDENT", i)] -> do {
              Stream.junk strm__;
              Stream.junk strm__;
              Stream.junk strm__;
              try
                match strm__ with parser
                [ [: `("", "=");
                     e = expr_0 ? P.error "labeled_patt" "'='" "[expr]";
                     `("", ")") ? P.error "labeled_patt" "[expr]" "')'";
                     loc = P.gloc bp ! :] ->
                    MLast.PaOlb loc (MLast.PaLid loc (Ploc.VaVal i))
                      (Ploc.VaVal (Some e))
                | [: `("", ")"); loc = P.gloc bp ! :] ->
                    MLast.PaOlb loc (MLast.PaLid loc (Ploc.VaVal i))
                      (Ploc.VaVal None)
                | [: `("", ":");
                     t = ctyp_0 ? P.error "labeled_patt" "':'" "[ctyp]";
                     a =
                       parser
                       [ [: `("", "=");
                            e = expr_0
                              ? P.error "labeled_patt" "'='" "[expr]";
                            `("", ")")
                              ? P.error "labeled_patt" "[expr]" "')'";
                            loc = P.gloc bp ! :] ->
                           MLast.PaOlb loc
                             (MLast.PaTyc loc (MLast.PaLid loc (Ploc.VaVal i))
                                t)
                             (Ploc.VaVal (Some e))
                       | [: `("", ")"); loc = P.gloc bp ! :] ->
                           MLast.PaOlb loc
                             (MLast.PaTyc loc (MLast.PaLid loc (Ploc.VaVal i))
                                t)
                             (Ploc.VaVal None) ]
                       ? P.error "labeled_patt" "[ctyp]" "'=' or ')'" :] ->
                    a ]
              with
              [ Stream.Failure ->
                  raise
                    (Stream.Error
                       (P.error "labeled_patt" "'?'" "'=' or ')' or ':'")) ]
            }
          | _ ->
              match
                try
                  Some (P.vala (P.token ("TILDEIDENTCOLON", "")) strm__)
                with
                [ Stream.Failure -> None ]
              with
              [ Some i ->
                  match
                    try Some (patt_8 strm__) with [ Stream.Failure -> None ]
                  with
                  [ Some p ->
                      let loc = P.gloc bp strm__ in
                      MLast.PaLab loc
                        (Ploc.VaVal
                           [(MLast.PaLid loc i, Ploc.VaVal (Some p))])
                  | _ ->
                      raise
                        (Stream.Error
                           (P.error "labeled_patt" "TILDEIDENTCOLON"
                              "[patt level simple]")) ]
              | _ ->
                  match
                    try Some (P.vala (P.token ("TILDEIDENT", "")) strm__) with
                    [ Stream.Failure -> None ]
                  with
                  [ Some i ->
                      let loc = P.gloc bp strm__ in
                      MLast.PaLab loc
                        (Ploc.VaVal [(MLast.PaLid loc i, Ploc.VaVal None)])
                  | _ ->
                      match
                        try
                          Some
                            (P.vala (P.token ("QUESTIONIDENTCOLON", ""))
                               strm__)
                        with
                        [ Stream.Failure -> None ]
                      with
                      [ Some i ->
                          try
                            match strm__ with parser
                            [ [: `("", "_"); loc = P.gloc bp ! :] ->
                                MLast.PaOlb loc (MLast.PaLid loc i)
                                  (Ploc.VaVal None)
                            | [: `("", "(");
                                 p = patt_0
                                   ? P.error "labeled_patt" "'('" "[patt]";
                                 a =
                                   parser
                                   [ [: `("", "=");
                                        e = expr_0
                                          ? P.error "labeled_patt" "'='"
                                              "[expr]";
                                        `("", ")")
                                          ? P.error "labeled_patt" "[expr]"
                                              "')'";
                                        loc = P.gloc bp ! :] ->
                                       MLast.PaOlb loc (MLast.PaLid loc i)
                                         (Ploc.VaVal
                                            (Some
                                               (MLast.ExOlb loc p
                                                  (Ploc.VaVal (Some e)))))
                                   | [: `("", ":");
                                        t = ctyp_0
                                          ? P.error "labeled_patt" "':'"
                                              "[ctyp]";
                                        a =
                                          parser
                                          [ [: `("", ")");
                                               loc = P.gloc bp ! :] ->
                                              MLast.PaOlb loc
                                                (MLast.PaLid loc i)
                                                (Ploc.VaVal
                                                   (Some
                                                      (MLast.ExOlb loc
                                                         (MLast.PaTyc loc p t)
                                                         (Ploc.VaVal None))))
                                          | [: `("", "=");
                                               e = expr_0
                                                 ? P.error "labeled_patt"
                                                     "'='" "[expr]";
                                               `("", ")")
                                                 ? P.error "labeled_patt"
                                                     "[expr]" "')'";
                                               loc = P.gloc bp ! :] ->
                                              MLast.PaOlb loc
                                                (MLast.PaLid loc i)
                                                (Ploc.VaVal
                                                   (Some
                                                      (MLast.ExOlb loc
                                                         (MLast.PaTyc loc p t)
                                                         (Ploc.VaVal
                                                            (Some e))))) ]
                                          ? P.error "labeled_patt" "[ctyp]"
                                              "')' or '='" :] ->
                                       a
                                   | [: `("", ")"); loc = P.gloc bp ! :] ->
                                       MLast.PaOlb loc (MLast.PaLid loc i)
                                         (Ploc.VaVal
                                            (Some
                                               (MLast.ExOlb loc p
                                                  (Ploc.VaVal None)))) ]
                                   ? P.error "labeled_patt" "[patt]"
                                       "'=' or ':' or ')'" :] ->
                                a
                            | [: `("LIDENT", j); loc = P.gloc bp ! :] ->
                                MLast.PaOlb loc (MLast.PaLid loc i)
                                  (Ploc.VaVal
                                     (Some
                                        (MLast.ExOlb loc
                                           (MLast.PaLid loc (Ploc.VaVal j))
                                           (Ploc.VaVal None)))) ]
                          with
                          [ Stream.Failure ->
                              raise
                                (Stream.Error
                                   (P.error "labeled_patt"
                                      "QUESTIONIDENTCOLON"
                                      "'_' or '(' or lowercase identifier")) ]
                      | _ ->
                          match
                            try
                              Some
                                (P.vala (P.token ("QUESTIONIDENT", ""))
                                   strm__)
                            with
                            [ Stream.Failure -> None ]
                          with
                          [ Some i ->
                              let loc = P.gloc bp strm__ in
                              MLast.PaOlb loc (MLast.PaLid loc i)
                                (Ploc.VaVal None)
                          | _ -> raise Stream.Failure ] ] ] ] ] ]
and class_fun_def_0 =
  parser bp
  [ [: p = labeled_patt_0;
       a =
         parser
         [ [: `("", "->");
              ce = class_expr_0
                ? P.error "class_fun_def" "'->'" "[class_expr]";
              loc = P.gloc bp ! :] ->
             MLast.CeFun loc p ce
         | [: cfd = class_fun_def_0; loc = P.gloc bp ! :] ->
             MLast.CeFun loc p cfd ]
         ? P.error "class_fun_def" "[labeled_patt]"
             "'->' or [class_fun_def]" :] ->
      a
  | [: p = patt_8;
       a =
         parser
         [ [: `("", "->");
              ce = class_expr_0
                ? P.error "class_fun_def" "'->'" "[class_expr]";
              loc = P.gloc bp ! :] ->
             MLast.CeFun loc p ce
         | [: cfd = class_fun_def_0; loc = P.gloc bp ! :] ->
             MLast.CeFun loc p cfd ]
         ? P.error "class_fun_def" "[patt level simple]"
             "'->' or [class_fun_def]" :] ->
      a ]
and class_expr_0 =
  parser bp
  [ [: `("", "fun");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "class_expr" "'fun'" "[alg_attributes_no_anti]";
       cfd = class_fun_def_0
         ? P.error "class_expr" "[alg_attributes_no_anti]"
             "[class_fun_def]" :] ->
      class_expr_wrap_attrs cfd alg_attrs
  | [: `("", "let");
       a =
         parser
         [ [: `("", "open");
              ovf = P.vala (P.bool (P.token ("", "!")))
                ? P.error "class_expr" "'open'" "???";
              alg_attrs = alg_attributes_no_anti_0
                ? P.error "class_expr" "???" "[alg_attributes_no_anti]";
              i = extended_longident_0
                ? P.error "class_expr" "[alg_attributes_no_anti]"
                    "[extended_longident]";
              `("", "in")
                ? P.error "class_expr" "[extended_longident]" "'in'";
              ce = class_expr_0 ? P.error "class_expr" "'in'" "[class_expr]";
              loc = P.gloc bp ! :] ->
             class_expr_wrap_attrs (MLast.CeLop loc ovf i ce) alg_attrs
         | [: rf = P.vala (P.bool (P.token ("", "rec")));
              lb =
                P.vala (P.list1sep let_binding_0 (P.token ("", "and")) False)
                ? P.error "class_expr" "???" "???";
              `("", "in") ? P.error "class_expr" "???" "'in'";
              ce = class_expr_0 ? P.error "class_expr" "'in'" "[class_expr]";
              loc = P.gloc bp ! :] ->
             MLast.CeLet loc rf lb ce ]
         ? P.error "class_expr" "'let'" "'open' or ???" :] ->
      a
  | [: a = class_expr_1 :] -> a ]
and class_expr_1 =
  parser bp [: a = class_expr_2; a = class_expr_1_cont bp a ! :] -> a
and class_expr_1_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "[@");
       attr = P.vala attribute_body_0
         ? P.error "class_expr" "'[@'" "[attribute_body]";
       `("", "]") ? P.error "class_expr" "[attribute_body]" "']'";
       loc = P.gloc bp !;
       let ct = a__ in
       a = class_expr_1_cont bp (MLast.CeAtt loc ct attr) ! :] ->
      a
  | [: :] -> a__ ]
and class_expr_2 =
  parser bp
  [ [: e = alg_extension_0; loc = P.gloc bp ! :] -> MLast.CeExten loc e
  | [: e = P.orzero class_expr_3 class_expr_0 :] -> e
  | [: a = class_expr_3 :] -> a ]
and class_expr_3 = parser bp [: ce = class_expr_apply_0 :] -> ce
and class_expr_gen_cont bp a__ strm__ =
  let x = class_expr_1_cont bp a__ strm__ in
  if x == a__ then x else class_expr_gen_cont bp x strm__
and class_fun_binding_0 =
  parser bp
  [ [: `("", "=");
       ce = class_expr_0
         ? P.error "class_fun_binding" "'='" "[class_expr]" :] ->
      ce
  | [: `("", ":");
       ct = class_type_0 ? P.error "class_fun_binding" "':'" "[class_type]";
       `("", "=") ? P.error "class_fun_binding" "[class_type]" "'='";
       ce = class_expr_0 ? P.error "class_fun_binding" "'='" "[class_expr]";
       loc = P.gloc bp ! :] ->
      MLast.CeTyc loc ce ct
  | [: p = labeled_patt_0;
       cfb = class_fun_binding_0
         ? P.error "class_fun_binding" "[labeled_patt]" "[class_fun_binding]";
       loc = P.gloc bp ! :] ->
      MLast.CeFun loc p cfb
  | [: p = patt_8;
       cfb = class_fun_binding_0
         ? P.error "class_fun_binding" "[patt level simple]"
             "[class_fun_binding]";
       loc = P.gloc bp ! :] ->
      MLast.CeFun loc p cfb ]
and class_declaration_0 =
  parser bp
    [: alg_attrs = alg_attributes_no_anti_0;
       vf = P.vala (P.bool (P.token ("", "virtual")))
         ? P.error "class_declaration" "[alg_attributes_no_anti]" "???";
       ctp = class_type_parameters_0
         ? P.error "class_declaration" "???" "[class_type_parameters]";
       i = P.vala (P.token ("LIDENT", ""))
         ? P.error "class_declaration" "[class_type_parameters]"
             "lowercase identifier";
       cfb = class_fun_binding_0
         ? P.error "class_declaration" "lowercase identifier"
             "[class_fun_binding]";
       item_attrs = item_attributes_0
         ? P.error "class_declaration" "[class_fun_binding]"
             "[item_attributes]";
       loc = P.gloc bp ! :] ->
      let attrs =
        merge_left_auxiliary_attrs ~{nonterm_name = "class-decl"}
          ~{left_name = "algebraic attributes"}
          ~{right_name = "item attributes"} alg_attrs item_attrs
      in
      {MLast.ciLoc = loc; MLast.ciVir = vf; MLast.ciPrm = ctp;
       MLast.ciNam = i; MLast.ciExp = cfb; MLast.ciAttributes = attrs}
and str_item_0 =
  parser bp
  [ [: `("", "class");
       a =
         parser
         [ [: `("", "type");
              ext = ext_opt_0 ? P.error "str_item" "'type'" "[ext_opt]";
              ctd =
                P.vala
                  (P.list1sep class_type_declaration_0 (P.token ("", "and"))
                     False)
                ? P.error "str_item" "[ext_opt]" "???";
              loc = P.gloc bp ! :] ->
             str_item_to_inline (MLast.StClt loc ctd) ext
         | [: ext = ext_opt_0;
              cd =
                P.vala
                  (P.list1sep class_declaration_0 (P.token ("", "and")) False)
                ? P.error "str_item" "[ext_opt]" "???";
              loc = P.gloc bp ! :] ->
             str_item_to_inline (MLast.StCls loc cd) ext ]
         ? P.error "str_item" "'class'" "'type' or [ext_opt]" :] ->
      a
  | [: `("", "exception");
       ext = ext_opt_0 ? P.error "str_item" "'exception'" "[ext_opt]";
       ec = P.vala extension_constructor_0
         ? P.error "str_item" "[ext_opt]" "[extension_constructor]";
       attrs = item_attributes_0
         ? P.error "str_item" "[extension_constructor]" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      str_item_to_inline (MLast.StExc loc ec attrs) ext
  | [: `("", "external");
       (ext, alg_attrs) = ext_attributes_0
         ? P.error "str_item" "'external'" "[ext_attributes]";
       a =
         parser
         [ [: `("", "(");
              i = operator_rparen
                ? P.error "str_item" "'('" "[operator_rparen]";
              `("", ":") ? P.error "str_item" "[operator_rparen]" "':'";
              ls = type_binder_opt_0
                ? P.error "str_item" "':'" "[type_binder_opt]";
              t = ctyp_0 ? P.error "str_item" "[type_binder_opt]" "[ctyp]";
              `("", "=") ? P.error "str_item" "[ctyp]" "'='";
              pd = P.vala (P.list1 (P.token ("STRING", "")))
                ? P.error "str_item" "'='" "???";
              item_attrs = item_attributes_0
                ? P.error "str_item" "???" "[item_attributes]";
              loc = P.gloc bp ! :] ->
             let attrs =
               merge_left_auxiliary_attrs
                 ~{nonterm_name = "str_item-external"}
                 ~{left_name = "algebraic attributes"}
                 ~{right_name = "item attributes"} alg_attrs item_attrs
             in
             str_item_to_inline (MLast.StExt loc (Ploc.VaVal i) ls t pd attrs)
               ext
         | [: :] ->
             match Stream.npeek 2 strm__ with
             [ [("LIDENT", i); ("", ":")] -> do {
                 Stream.junk strm__;
                 Stream.junk strm__;
                 let i = Ploc.VaVal i in
                 match
                   try Some (type_binder_opt_0 strm__) with
                   [ Stream.Failure -> None ]
                 with
                 [ Some ls ->
                     match
                       try Some (ctyp_0 strm__) with
                       [ Stream.Failure -> None ]
                     with
                     [ Some t ->
                         match strm__ with parser
                         [ [: `("", "=");
                              pd = P.vala (P.list1 (P.token ("STRING", "")))
                                ? P.error "str_item" "'='" "???";
                              item_attrs = item_attributes_0
                                ? P.error "str_item" "???"
                                    "[item_attributes]";
                              loc = P.gloc bp ! :] ->
                             let attrs =
                               merge_left_auxiliary_attrs
                                 ~{nonterm_name = "str_item-external"}
                                 ~{left_name = "algebraic attributes"}
                                 ~{right_name = "item attributes"} alg_attrs
                                 item_attrs
                             in
                             str_item_to_inline
                               (MLast.StExt loc i ls t pd attrs) ext
                         | [: :] ->
                             raise
                               (Stream.Error
                                  (P.error "str_item" "[ctyp]" "'='")) ]
                     | _ ->
                         raise
                           (Stream.Error
                              (P.error "str_item" "[type_binder_opt]"
                                 "[ctyp]")) ]
                 | _ ->
                     raise
                       (Stream.Error
                          (P.error "str_item" "lowercase identifier"
                             "[type_binder_opt]")) ]
               }
             | _ -> raise Stream.Failure ] ]
         ? P.error "str_item" "[ext_attributes]"
             "'(' or lowercase identifier" :] ->
      a
  | [: `("", "include");
       (ext, alg_attrs) = ext_attributes_0
         ? P.error "str_item" "'include'" "[ext_attributes]";
       me = module_expr_0
         ? P.error "str_item" "[ext_attributes]" "[module_expr]";
       item_attrs = item_attributes_0
         ? P.error "str_item" "[module_expr]" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      let attrs =
        merge_left_auxiliary_attrs ~{nonterm_name = "str_item-include"}
          ~{left_name = "algebraic attributes"}
          ~{right_name = "item attributes"} alg_attrs item_attrs
      in
      str_item_to_inline (MLast.StInc loc me attrs) ext
  | [: `("", "module");
       a =
         parser
         [ [: `("", "type");
              (ext, alg_attrs) = ext_attributes_0
                ? P.error "str_item" "'type'" "[ext_attributes]";
              i = P.vala ident_0
                ? P.error "str_item" "[ext_attributes]" "[ident]";
              a =
                parser
                [ [: `("", "=");
                     mt = module_type_0
                       ? P.error "str_item" "'='" "[module_type]";
                     item_attrs = item_attributes_0
                       ? P.error "str_item" "[module_type]"
                           "[item_attributes]";
                     loc = P.gloc bp ! :] ->
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "str_item-module-type"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    str_item_to_inline (MLast.StMty loc i mt attrs) ext
                | [: item_attrs = item_attributes_0; loc = P.gloc bp ! :] ->
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "str_item-module-type"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    str_item_to_inline
                      (MLast.StMty loc i
                         (MLast.MtQuo loc (Ploc.VaVal "abstract")) attrs)
                      ext ]
                ? P.error "str_item" "[ident]"
                    "'=' or [item_attributes]" :] ->
             a
         | [: (ext, alg_attrs) = ext_attributes_0;
              r = P.vala (P.bool (P.token ("", "rec")))
                ? P.error "str_item" "[ext_attributes]" "???";
              h = first_mod_binding_0
                ? P.error "str_item" "???" "[first_mod_binding]";
              t = P.list0 rest_mod_binding_0; loc = P.gloc bp ! :] ->
             let (i, me, attrs) = h in
             let attrs =
               merge_left_auxiliary_attrs ~{nonterm_name = "str_item-module"}
                 ~{left_name = "algebraic attributes"}
                 ~{right_name = "item attributes"} alg_attrs attrs
             in
             let h = (i, me, attrs) in
             str_item_to_inline (MLast.StMod loc r (Ploc.VaVal [h :: t]))
               ext ]
         ? P.error "str_item" "'module'" "'type' or [ext_attributes]" :] ->
      a
  | [: `("", "open");
       (ext, alg_attrs) = ext_attributes_0
         ? P.error "str_item" "'open'" "[ext_attributes]";
       ovf = P.vala (P.bool (P.token ("", "!")))
         ? P.error "str_item" "[ext_attributes]" "???";
       me = module_expr_0 ? P.error "str_item" "???" "[module_expr]";
       item_attrs = item_attributes_0
         ? P.error "str_item" "[module_expr]" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      let attrs =
        merge_left_auxiliary_attrs ~{nonterm_name = "str_item-open"}
          ~{left_name = "algebraic attributes"}
          ~{right_name = "item attributes"} alg_attrs item_attrs
      in
      str_item_to_inline (MLast.StOpn loc ovf me attrs) ext
  | [: `("", "type");
       (ext, attrs) = ext_attributes_0
         ? P.error "str_item" "'type'" "[ext_attributes]";
       a =
         parser
         [ [: _ = check_type_decl; nr = P.bool (P.token ("", "nonrec"));
              htd = first_type_decl_0
                ? P.error "str_item" "" "[first_type_decl]";
              ttd = P.list0 rest_type_decl_0; loc = P.gloc bp ! :] -> do {
             let attrs =
               merge_left_auxiliary_attrs
                 ~{nonterm_name = "str_item-type_decl"}
                 ~{left_name = "algebraic attributes"}
                 ~{right_name = "item attributes"} attrs
                 htd.MLast.tdAttributes
             in
             let htd = {(htd) with MLast.tdAttributes = attrs} in
             let tdl = [htd :: ttd] in
             if List.for_all (fun td -> Pcaml.unvala td.MLast.tdIsDecl) tdl
             then
               ()
             else if
               List.for_all (fun td -> not (Pcaml.unvala td.MLast.tdIsDecl))
                 tdl
             then
               if nr then
                 failwith "type-subst declaration must not specify <<nonrec>>"
               else ()
             else failwith "type-declaration cannot mix decl and subst";
             str_item_to_inline
               (MLast.StTyp loc (Ploc.VaVal nr) (Ploc.VaVal tdl)) ext
           }
         | [: _ = check_type_extension;
              te = type_extension_0
                ? P.error "str_item" "[check_type_extension]"
                    "[type_extension]";
              loc = P.gloc bp ! :] ->
             let attrs =
               merge_left_auxiliary_attrs
                 ~{nonterm_name = "str_item-type_extension"}
                 ~{left_name = "algebraic attributes"}
                 ~{right_name = "item attributes"} attrs te.MLast.teAttributes
             in
             let te = {(te) with MLast.teAttributes = attrs} in
             str_item_to_inline
               (MLast.StTypExten loc
                  {MLast.teNam = te.MLast.teNam; MLast.tePrm = te.MLast.tePrm;
                   MLast.tePrv = te.MLast.tePrv; MLast.teECs = te.MLast.teECs;
                   MLast.teAttributes = te.MLast.teAttributes})
               ext ]
         ? P.error "str_item" "[ext_attributes]"
             "[check_type_decl] or [check_type_extension]" :] ->
      a
  | [: _ = check_let_exception :] ->
      match Stream.npeek 3 strm__ with
      [ [("", "let"); ("", "exception"); ("UIDENT", id)] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          Stream.junk strm__;
          let id = Ploc.VaVal id in
          try
            match strm__ with parser
            [ [: `("", "of");
                 tyl = P.vala (P.list1 ctyp_6)
                   ? P.error "str_item" "'of'" "???";
                 alg_attrs = alg_attributes_0
                   ? P.error "str_item" "???" "[alg_attributes]";
                 `("", "in") ? P.error "str_item" "[alg_attributes]" "'in'";
                 x = expr_0 ? P.error "str_item" "'in'" "[expr]";
                 attrs = item_attributes_0
                   ? P.error "str_item" "[expr]" "[item_attributes]";
                 loc = P.gloc bp ! :] ->
                let e = MLast.ExLEx loc id tyl x alg_attrs in
                MLast.StExp loc e attrs
            | [: alg_attrs = alg_attributes_0;
                 `("", "in") ? P.error "str_item" "[alg_attributes]" "'in'";
                 x = expr_0 ? P.error "str_item" "'in'" "[expr]";
                 attrs = item_attributes_0
                   ? P.error "str_item" "[expr]" "[item_attributes]";
                 loc = P.gloc bp ! :] ->
                let e = MLast.ExLEx loc id (Ploc.VaVal []) x alg_attrs in
                MLast.StExp loc e attrs ]
          with
          [ Stream.Failure ->
              raise
                (Stream.Error
                   (P.error "str_item" "'let'" "'of' or [alg_attributes]")) ]
        }
      | _ ->
          raise
            (Stream.Error
               (P.error "str_item" "[check_let_exception]" "'let'")) ]
  | [: _ = check_let_not_exception;
       `("", "let") ? P.error "str_item" "[check_let_not_exception]" "'let'";
       a =
         parser
         [ [: `("", "module");
              (ext, attrs) = ext_attributes_0
                ? P.error "str_item" "'module'" "[ext_attributes]";
              m = P.vala uidopt_0
                ? P.error "str_item" "[ext_attributes]" "[uidopt]";
              mb = mod_fun_binding_0
                ? P.error "str_item" "[uidopt]" "[mod_fun_binding]";
              `("", "in") ? P.error "str_item" "[mod_fun_binding]" "'in'";
              e = expr_0 ? P.error "str_item" "'in'" "[expr]";
              loc = P.gloc bp ! :] ->
             let e = expr_to_inline (MLast.ExLmd loc m mb e) ext attrs in
             MLast.StExp loc e (Ploc.VaVal [])
         | [: `("", "open");
              ovf = P.vala (P.bool (P.token ("", "!")))
                ? P.error "str_item" "'open'" "???";
              (ext, attrs) = ext_attributes_0
                ? P.error "str_item" "???" "[ext_attributes]";
              m = module_expr_0
                ? P.error "str_item" "[ext_attributes]" "[module_expr]";
              `("", "in") ? P.error "str_item" "[module_expr]" "'in'";
              e = expr_0 ? P.error "str_item" "'in'" "[expr]";
              loc = P.gloc bp ! :] ->
             let e = expr_to_inline (MLast.ExLop loc ovf m e) ext attrs in
             MLast.StExp loc e (Ploc.VaVal [])
         | [: (ext, alg_attrs) = ext_attributes_0;
              r = P.vala (P.bool (P.token ("", "rec")))
                ? P.error "str_item" "[ext_attributes]" "???";
              h = first_let_binding_0
                ? P.error "str_item" "???" "[first_let_binding]";
              t = P.list0 and_let_binding_0;
              a =
                parser
                [ [: `("", "in");
                     x = expr_0 ? P.error "str_item" "'in'" "[expr]";
                     loc = P.gloc bp ! :] ->
                    let (a, b, item_attrs) = h in
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "let_binding"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    let h = (a, b, attrs) in
                    let l = [h :: t] in
                    let e =
                      expr_to_inline (MLast.ExLet loc r (Ploc.VaVal l) x) ext
                        []
                    in
                    MLast.StExp loc e (Ploc.VaVal [])
                | [: loc = P.gloc bp :] ->
                    let (a, b, item_attrs) = h in
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "let_binding"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    let h = (a, b, attrs) in
                    let l = [h :: t] in
                    let si =
                      match l with
                      [ [(p, e, attrs)] ->
                          match p with
                          [ MLast.PaAny _ -> MLast.StExp loc e attrs
                          | _ -> MLast.StVal loc r (Ploc.VaVal l) ]
                      | _ -> MLast.StVal loc r (Ploc.VaVal l) ]
                    in
                    str_item_to_inline si ext ]
                ? P.error "str_item" "" "[and_let_binding] or 'in'" :] ->
             a ]
         ? P.error "str_item" "'let'"
             "'module' or 'open' or [ext_attributes]" :] ->
      a
  | [: e = expr_0;
       attrs = item_attributes_0
         ? P.error "str_item" "[expr]" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      MLast.StExp loc e attrs
  | [: attr = floating_attribute_0; loc = P.gloc bp ! :] ->
      MLast.StFlAtt loc attr
  | [: e = item_extension_0;
       attrs = item_attributes_0
         ? P.error "str_item" "[item_extension]" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      MLast.StExten loc e attrs ]
and structure_0 =
  parser bp
    [: st =
         P.vala
           (P.list0
              (parser
                 [: s = str_item_0; _ = P.option (P.token ("", ";;")) :] ->
                   s)) :] ->
      st
and module_expr_0 =
  parser bp
  [ [: `("", "functor");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "module_expr" "'functor'" "[alg_attributes_no_anti]";
       argl = P.list1 functor_parameter_0
         ? P.error "module_expr" "[alg_attributes_no_anti]"
             "[functor_parameter]";
       `("", "->") ? P.error "module_expr" "" "[functor_parameter] or '->'";
       me = module_expr_0 ? P.error "module_expr" "'->'" "[module_expr]";
       loc = P.gloc bp ! :] ->
      let me =
        List.fold_right (fun arg me -> MLast.MeFun loc (Ploc.VaVal arg) me)
          argl me
      in
      module_expr_wrap_attrs me alg_attrs
  | [: a = module_expr_1 :] -> a ]
and module_expr_1 =
  parser bp [: a = module_expr_2; a = module_expr_1_cont bp a ! :] -> a
and module_expr_1_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "[@");
       attr = P.vala attribute_body_0
         ? P.error "module_expr" "'[@'" "[attribute_body]";
       `("", "]") ? P.error "module_expr" "[attribute_body]" "']'";
       loc = P.gloc bp !;
       let e1 = a__ in
       a = module_expr_1_cont bp (MLast.MeAtt loc e1 attr) ! :] ->
      a
  | [: :] -> a__ ]
and module_expr_2 =
  parser bp
  [ [: `("", "struct");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "module_expr" "'struct'" "[alg_attributes_no_anti]";
       _ = P.option (P.token ("", ";;"));
       st = structure_0 ? P.error "module_expr" "" "[structure]";
       `("", "end") ? P.error "module_expr" "[structure]" "'end'";
       loc = P.gloc bp ! :] ->
      module_expr_wrap_attrs (MLast.MeStr loc st) alg_attrs
  | [: a = module_expr_3 :] -> a ]
and module_expr_3 =
  parser bp [: a = module_expr_4; a = module_expr_3_cont bp a ! :] -> a
and module_expr_3_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", ".");
       me2 = P.orzero module_expr_4 module_expr_0
         ? P.error "module_expr" "'.'" "[module_expr]";
       loc = P.gloc bp !;
       let me1 = a__ in
       a = module_expr_3_cont bp (MLast.MeAcc loc me1 me2) ! :] ->
      a
  | [: :] -> a__ ]
and module_expr_4 =
  parser bp [: a = module_expr_5; a = module_expr_4_cont bp a ! :] -> a
and module_expr_4_cont bp a__ strm__ =
  match Stream.npeek 2 strm__ with
  [ [("", "("); ("", ")")] -> do {
      Stream.junk strm__;
      Stream.junk strm__;
      try
        let loc = P.gloc bp strm__ in
        let me1 = a__ in
        module_expr_4_cont bp
          (MLast.MeApp loc me1 (MLast.MeStr loc (Ploc.VaVal []))) strm__
      with
      [ Stream.Failure ->
          raise (Stream.Error (P.error "module_expr" "'('" "???")) ]
    }
  | _ ->
      match
        try Some (paren_module_expr_0 strm__) with [ Stream.Failure -> None ]
      with
      [ Some me2 ->
          let loc = P.gloc bp strm__ in
          let me1 = a__ in
          module_expr_4_cont bp (MLast.MeApp loc me1 me2) strm__
      | _ -> a__ ] ]
and module_expr_5 =
  parser bp
  [ [: i = mod_expr_ident_0 :] -> i
  | [: me = paren_module_expr_0 :] -> me
  | [: e = alg_extension_0; loc = P.gloc bp ! :] -> MLast.MeExten loc e ]
and module_expr_gen_cont bp a__ strm__ =
  let x = module_expr_1_cont bp a__ strm__ in
  let x = module_expr_3_cont bp x strm__ in
  let x = module_expr_4_cont bp x strm__ in
  if x == a__ then x else module_expr_gen_cont bp x strm__
and longident_0 =
  parser bp
    [: i = P.vala (P.token ("UIDENT", "")); loc = P.gloc bp !;
       a =
         longident_0_cont bp
           (let i = vala_map uident_True_True_ i in
            MLast.LiUid loc i) ! :] ->
      a
and longident_0_cont bp a__ strm__ =
  match try Some (check_dot_uid strm__) with [ Stream.Failure -> None ] with
  [ Some _ ->
      match Stream.npeek 2 strm__ with
      [ [("", "."); ("UIDENT", i)] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          let i = Ploc.VaVal i in
          try
            let loc = P.gloc bp strm__ in
            let me1 = a__ in
            longident_0_cont bp
              (let i = vala_map uident_True_True_ i in
               MLast.LiAcc loc me1 i)
              strm__
          with
          [ Stream.Failure ->
              raise (Stream.Error (P.error "longident" "'.'" "???")) ]
        }
      | _ ->
          raise (Stream.Error (P.error "longident" "[check_dot_uid]" "'.'")) ]
  | _ -> a__ ]
and longident_lident_0 =
  parser bp
  [ [: i = P.vala (P.token ("LIDENT", "")) :] -> (None, i)
  | [: li = P.vala longident_0 :] ->
      match Stream.npeek 2 strm__ with
      [ [("", "."); ("LIDENT", i)] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          let i = Ploc.VaVal i in
          try (Some li, i) with
          [ Stream.Failure ->
              raise (Stream.Error (P.error "longident_lident" "'.'" "???")) ]
        }
      | _ ->
          raise
            (Stream.Error
               (P.error "longident_lident" "[longident]" "'.'")) ] ]
and type_parameters_0 =
  parser bp
  [ [: `("", "(");
       tpl = P.list1sep type_parameter_0 (P.token ("", ",")) False
         ? P.error "type_parameters" "'('" "[type_parameter]";
       `("", ")")
         ? P.error "type_parameters" "" "[type_parameter] or ')'" :] ->
      tpl
  | [: tp = type_parameter_0 :] -> [tp]
  | [: :] -> [] ]
and with_constr_0 =
  parser bp
  [ [: `("", "type");
       tpl = P.vala type_parameters_0
         ? P.error "with_constr" "'type'" "[type_parameters]";
       i = P.vala longident_lident_0
         ? P.error "with_constr" "[type_parameters]" "[longident_lident]";
       a =
         parser
         [ [: `("", "=");
              pf = P.vala (P.bool (P.token ("", "private")))
                ? P.error "with_constr" "'='" "???";
              t = ctyp_1
                ? P.error "with_constr" "???"
                    "[ctyp level below_alg_attribute]";
              loc = P.gloc bp ! :] ->
             MLast.WcTyp loc i tpl pf t
         | [: `("", ":=");
              t = ctyp_1
                ? P.error "with_constr" "':='"
                    "[ctyp level below_alg_attribute]";
              loc = P.gloc bp ! :] ->
             MLast.WcTys loc i tpl t ]
         ? P.error "with_constr" "[longident_lident]" "'=' or ':='" :] ->
      a
  | [: `("", "module");
       a =
         parser
         [ [: `("", "type");
              li = P.vala longident_0
                ? P.error "with_constr" "'type'" "[longident]";
              a =
                parser
                [ [: `("", "=");
                     mt = module_type_0
                       ? P.error "with_constr" "'='" "[module_type]";
                     loc = P.gloc bp ! :] ->
                    MLast.WcMty loc li mt
                | [: `("", ":=");
                     mt = module_type_0
                       ? P.error "with_constr" "':='" "[module_type]";
                     loc = P.gloc bp ! :] ->
                    MLast.WcMts loc li mt ]
                ? P.error "with_constr" "[longident]" "'=' or ':='" :] ->
             a
         | [: i = P.vala longident_0;
              a =
                parser
                [ [: `("", "=");
                     me = module_expr_0
                       ? P.error "with_constr" "'='" "[module_expr]";
                     loc = P.gloc bp ! :] ->
                    MLast.WcMod loc i me
                | [: `("", ":=");
                     me = module_expr_0
                       ? P.error "with_constr" "':='" "[module_expr]";
                     loc = P.gloc bp ! :] ->
                    MLast.WcMos loc i me ]
                ? P.error "with_constr" "[longident]" "'=' or ':='" :] ->
             a ]
         ? P.error "with_constr" "'module'" "'type' or [longident]" :] ->
      a ]
and uidopt_0 =
  parser bp
  [ [: `("", "_") :] -> None
  | [: m = P.vala (P.token ("UIDENT", "")) :] -> Some m ]
and functor_parameter_0 =
  parser bp
    [: `("", "(");
       a =
         parser
         [ [: `("", ")") :] -> None
         | [: i = P.vala uidopt_0;
              `("", ":") ? P.error "functor_parameter" "[uidopt]" "':'";
              t = module_type_0
                ? P.error "functor_parameter" "':'" "[module_type]";
              `("", ")")
                ? P.error "functor_parameter" "[module_type]" "')'" :] ->
             Some (i, t) ]
         ? P.error "functor_parameter" "'('" "')' or [uidopt]" :] ->
      a
and module_type_0 =
  parser bp
  [ [: `("", "functor");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "module_type" "'functor'" "[alg_attributes_no_anti]";
       argl = P.list1 functor_parameter_0
         ? P.error "module_type" "[alg_attributes_no_anti]"
             "[functor_parameter]";
       `("", "->") ? P.error "module_type" "" "[functor_parameter] or '->'";
       mt = module_type_0 ? P.error "module_type" "'->'" "[module_type]";
       loc = P.gloc bp ! :] ->
      let mt =
        List.fold_right (fun arg mt -> MLast.MtFun loc (Ploc.VaVal arg) mt)
          argl mt
      in
      module_type_wrap_attrs mt alg_attrs
  | [: a = module_type_1 :] -> a ]
and module_type_1 =
  parser bp [: a = module_type_2; a = module_type_1_cont bp a ! :] -> a
and module_type_1_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "->");
       mt2 = P.orzero module_type_1 module_type_0
         ? P.error "module_type" "'->'" "[module_type]";
       let mt1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.MtFun loc (Ploc.VaVal (Some (Ploc.VaVal None, mt1))) mt2
  | [: :] -> a__ ]
and module_type_2 =
  parser bp [: a = module_type_3; a = module_type_2_cont bp a ! :] -> a
and module_type_2_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "[@");
       attr = P.vala attribute_body_0
         ? P.error "module_type" "'[@'" "[attribute_body]";
       `("", "]") ? P.error "module_type" "[attribute_body]" "']'";
       loc = P.gloc bp !;
       let e1 = a__ in
       a = module_type_2_cont bp (MLast.MtAtt loc e1 attr) ! :] ->
      a
  | [: :] -> a__ ]
and module_type_3 =
  parser bp [: a = module_type_4; a = module_type_3_cont bp a ! :] -> a
and module_type_3_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "with");
       wcl = P.vala (P.list1sep with_constr_0 (P.token ("", "and")) False)
         ? P.error "module_type" "'with'" "???";
       loc = P.gloc bp !;
       let mt = a__ in
       a = module_type_3_cont bp (MLast.MtWit loc mt wcl) ! :] ->
      a
  | [: :] -> a__ ]
and module_type_4 =
  parser bp
  [ [: `("", "sig");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "module_type" "'sig'" "[alg_attributes_no_anti]";
       sg = signature_0
         ? P.error "module_type" "[alg_attributes_no_anti]" "[signature]";
       `("", "end") ? P.error "module_type" "[signature]" "'end'";
       loc = P.gloc bp ! :] ->
      module_type_wrap_attrs (MLast.MtSig loc sg) alg_attrs
  | [: :] ->
      match Stream.npeek 3 strm__ with
      [ [("", "module"); ("", "type"); ("", "of")] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          Stream.junk strm__;
          match
            try Some (alg_attributes_no_anti_0 strm__) with
            [ Stream.Failure -> None ]
          with
          [ Some alg_attrs ->
              match
                try Some (module_expr_0 strm__) with
                [ Stream.Failure -> None ]
              with
              [ Some me ->
                  let loc = P.gloc bp strm__ in
                  module_type_wrap_attrs (MLast.MtTyo loc me) alg_attrs
              | _ ->
                  raise
                    (Stream.Error
                       (P.error "module_type" "[alg_attributes_no_anti]"
                          "[module_expr]")) ]
          | _ ->
              raise
                (Stream.Error
                   (P.error "module_type" "'module'"
                      "[alg_attributes_no_anti]")) ]
        }
      | _ ->
          match strm__ with parser
          [ [: `("", "(");
               mt = module_type_0
                 ? P.error "module_type" "'('" "[module_type]";
               `("", ")") ? P.error "module_type" "[module_type]" "')'" :] ->
              mt
          | [: i = P.vala (P.token ("LIDENT", "")); loc = P.gloc bp ! :] ->
              MLast.MtLid loc i
          | [: li = extended_longident_0;
               a =
                 parser
                   [: :] ->
                     match Stream.npeek 2 strm__ with
                     [ [("", "."); ("LIDENT", i)] -> do {
                         Stream.junk strm__;
                         Stream.junk strm__;
                         let i = Ploc.VaVal i in
                         try
                           let loc = P.gloc bp strm__ in
                           MLast.MtLongLid loc li i
                         with
                         [ Stream.Failure ->
                             raise
                               (Stream.Error
                                  (P.error "module_type" "'.'" "???")) ]
                       }
                     | _ ->
                         let loc = P.gloc bp strm__ in
                         MLast.MtLong loc li ]
                 ? P.error "module_type" "[extended_longident]" "'.'" :] ->
              a
          | [: e = alg_extension_0; loc = P.gloc bp ! :] ->
              MLast.MtExten loc e ] ] ]
and module_type_gen_cont bp a__ strm__ =
  let x = module_type_1_cont bp a__ strm__ in
  let x = module_type_2_cont bp x strm__ in
  let x = module_type_3_cont bp x strm__ in
  if x == a__ then x else module_type_gen_cont bp x strm__
and ext_attributes_0 =
  parser bp
    [: e = ext_opt_0;
       l = alg_attributes_no_anti_0
         ? P.error "ext_attributes" "[ext_opt]"
             "[alg_attributes_no_anti]" :] ->
      (e, l)
and typevar_0 =
  parser bp
    [: `("", "'"); i = ident_0 ? P.error "typevar" "'''" "[ident]" :] -> i
and poly_type_below_alg_attribute_0 =
  parser bp
  [ [: `("", "type");
       nt = P.list1 (P.token ("LIDENT", ""))
         ? P.error "poly_type_below_alg_attribute" "'type'"
             "lowercase identifier";
       `("", ".")
         ? P.error "poly_type_below_alg_attribute" ""
             "lowercase identifier or '.'";
       ct = ctyp_0 ? P.error "poly_type_below_alg_attribute" "'.'" "[ctyp]";
       loc = P.gloc bp ! :] ->
      MLast.TyPot loc (Ploc.VaVal nt) ct
  | [: _ = test_typevar_list_dot;
       tpl = P.list1 typevar_0
         ? P.error "poly_type_below_alg_attribute" "[test_typevar_list_dot]"
             "[typevar]";
       `("", ".")
         ? P.error "poly_type_below_alg_attribute" "" "[typevar] or '.'";
       t2 = ctyp_0 ? P.error "poly_type_below_alg_attribute" "'.'" "[ctyp]";
       loc = P.gloc bp ! :] ->
      MLast.TyPol loc (Ploc.VaVal tpl) t2
  | [: t = ctyp_1 :] -> t ]
and field_0 =
  parser bp
  [ [: _ = check_lident_colon :] ->
      match Stream.npeek 2 strm__ with
      [ [("LIDENT", lab); ("", ":")] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          match
            try Some (poly_type_below_alg_attribute_0 strm__) with
            [ Stream.Failure -> None ]
          with
          [ Some t ->
              match
                try Some (alg_attributes_0 strm__) with
                [ Stream.Failure -> None ]
              with
              [ Some alg_attrs -> (Some lab, t, alg_attrs)
              | _ ->
                  raise
                    (Stream.Error
                       (P.error "field" "[poly_type_below_alg_attribute]"
                          "[alg_attributes]")) ]
          | _ ->
              raise
                (Stream.Error
                   (P.error "field" "lowercase identifier"
                      "[poly_type_below_alg_attribute]")) ]
        }
      | _ ->
          raise
            (Stream.Error
               (P.error "field" "[check_lident_colon]"
                  "lowercase identifier")) ]
  | [: _ = check_not_lident_colon;
       t = poly_type_below_alg_attribute_0
         ? P.error "field" "[check_not_lident_colon]"
             "[poly_type_below_alg_attribute]";
       alg_attrs = alg_attributes_0
         ? P.error "field" "[poly_type_below_alg_attribute]"
             "[alg_attributes]" :] ->
      (None, t, alg_attrs) ]
and meth_list_0 =
  parser bp
    [: f = field_0;
       a =
         parser
         [ [: `("", ";");
              a =
                parser
                [ [: ml = meth_list_0 :] -> [f :: ml]
                | [: :] -> [f] ]
                ? P.error "meth_list" "';'" "[meth_list]" :] ->
             a
         | [: :] -> [f] ]
         ? P.error "meth_list" "[field]" "';'" :] ->
      a
and extended_longident_lident_0 =
  parser bp
  [ [: i = P.vala (P.token ("LIDENT", "")) :] -> (None, i)
  | [: li = P.vala extended_longident_0 :] ->
      match Stream.npeek 2 strm__ with
      [ [("", "."); ("LIDENT", i)] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          let i = Ploc.VaVal i in
          try (Some li, i) with
          [ Stream.Failure ->
              raise
                (Stream.Error
                   (P.error "extended_longident_lident" "'.'" "???")) ]
        }
      | _ ->
          raise
            (Stream.Error
               (P.error "extended_longident_lident" "[extended_longident]"
                  "'.'")) ] ]
and name_tag_0 =
  parser bp
    [: `("", "`"); i = ident_0 ? P.error "name_tag" "'`'" "[ident]" :] -> i
and alg_attributes_0 =
  parser bp [: l = P.vala alg_attributes_no_anti_0 :] -> l
and poly_variant_0 =
  parser bp
  [ [: `("", "`");
       i = P.vala ident_0 ? P.error "poly_variant" "'`'" "[ident]";
       a =
         parser
         [ [: `("", "of");
              ao = P.vala (P.bool (P.token ("", "&")))
                ? P.error "poly_variant" "'of'" "???";
              l = P.vala (P.list1sep ctyp_1 (P.token ("", "&")) False)
                ? P.error "poly_variant" "???" "???";
              alg_attrs = alg_attributes_0
                ? P.error "poly_variant" "???" "[alg_attributes]";
              loc = P.gloc bp ! :] ->
             MLast.PvTag loc i ao l alg_attrs
         | [: alg_attrs = alg_attributes_0; loc = P.gloc bp ! :] ->
             MLast.PvTag loc i (Ploc.VaVal True) (Ploc.VaVal []) alg_attrs ]
         ? P.error "poly_variant" "[ident]" "'of' or [alg_attributes]" :] ->
      a
  | [: t = ctyp_0; loc = P.gloc bp ! :] -> MLast.PvInh loc t ]
and ctyp_ident_0 =
  parser bp
  [ [: i = P.vala (P.token ("LIDENT", "")); loc = P.gloc bp ! :] ->
      MLast.TyLid loc i
  | [: me1 = extended_longident_0 :] ->
      match Stream.npeek 2 strm__ with
      [ [("", "."); ("LIDENT", i)] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          let i = Ploc.VaVal i in
          try
            let loc = P.gloc bp strm__ in
            MLast.TyAcc loc me1 i
          with
          [ Stream.Failure ->
              raise (Stream.Error (P.error "ctyp_ident" "'.'" "???")) ]
        }
      | _ ->
          raise
            (Stream.Error
               (P.error "ctyp_ident" "[extended_longident]" "'.'")) ] ]
and ctyp_0 = parser bp [: a = ctyp_1; a = ctyp_gen_cont bp a ! :] -> a
and ctyp_0_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "[@");
       attr = P.vala attribute_body_0
         ? P.error "ctyp" "'[@'" "[attribute_body]";
       `("", "]") ? P.error "ctyp" "[attribute_body]" "']'";
       loc = P.gloc bp !;
       let ct = a__ in
       a = ctyp_0_cont bp (MLast.TyAtt loc ct attr) ! :] ->
      a
  | [: :] -> a__ ]
and ctyp_1 =
  parser bp
  [ [: t = P.orzero ctyp_2 ctyp_0 :] -> t
  | [: a = ctyp_2 :] -> a ]
and ctyp_2 = parser bp [: a = ctyp_3; a = ctyp_2_cont bp a ! :] -> a
and ctyp_2_cont bp a__ strm__ =
  match Stream.npeek 2 strm__ with
  [ [("", "as"); ("", "'")] -> do {
      Stream.junk strm__;
      Stream.junk strm__;
      match try Some (ident_0 strm__) with [ Stream.Failure -> None ] with
      [ Some i ->
          let loc = P.gloc bp strm__ in
          let t1 = a__ in
          ctyp_2_cont bp (MLast.TyAli loc t1 (MLast.TyQuo loc (Ploc.VaVal i)))
            strm__
      | _ -> raise (Stream.Error (P.error "ctyp" "'as'" "[ident]")) ]
    }
  | _ -> a__ ]
and ctyp_3 = parser bp [: a = ctyp_4; a = ctyp_3_cont bp a ! :] -> a
and ctyp_3_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "->");
       t2 = P.orzero ctyp_3 ctyp_0 ? P.error "ctyp" "'->'" "[ctyp]";
       let t1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.TyArr loc t1 t2
  | [: :] -> a__ ]
and ctyp_4 =
  parser bp
    [: :] ->
      match Stream.npeek 3 strm__ with
      [ [("", "?"); ("LIDENT", i); ("", ":")] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          Stream.junk strm__;
          let i = Ploc.VaVal i in
          match
            try Some (P.orzero ctyp_4 ctyp_0 strm__) with
            [ Stream.Failure -> None ]
          with
          [ Some t ->
              let loc = P.gloc bp strm__ in
              MLast.TyOlb loc i t
          | _ -> raise (Stream.Error (P.error "ctyp" "'?'" "[ctyp]")) ]
        }
      | _ ->
          match Stream.npeek 2 strm__ with
          [ [("LIDENT", i); ("", ":")] -> do {
              Stream.junk strm__;
              Stream.junk strm__;
              let i = Ploc.VaVal i in
              match
                try Some (P.orzero ctyp_4 ctyp_0 strm__) with
                [ Stream.Failure -> None ]
              with
              [ Some t ->
                  let loc = P.gloc bp strm__ in
                  MLast.TyLab loc i t
              | _ ->
                  raise
                    (Stream.Error
                       (P.error "ctyp" "lowercase identifier" "[ctyp]")) ]
            }
          | _ ->
              match
                try
                  Some (P.vala (P.token ("QUESTIONIDENTCOLON", "")) strm__)
                with
                [ Stream.Failure -> None ]
              with
              [ Some i ->
                  match
                    try Some (P.orzero ctyp_4 ctyp_0 strm__) with
                    [ Stream.Failure -> None ]
                  with
                  [ Some t ->
                      let loc = P.gloc bp strm__ in
                      MLast.TyOlb loc i t
                  | _ ->
                      raise
                        (Stream.Error
                           (P.error "ctyp" "QUESTIONIDENTCOLON" "[ctyp]")) ]
              | _ ->
                  match Stream.npeek 2 strm__ with
                  [ [("QUESTIONIDENT", i); ("", ":")] -> do {
                      Stream.junk strm__;
                      Stream.junk strm__;
                      let i = Ploc.VaVal i in
                      match
                        try Some (P.orzero ctyp_4 ctyp_0 strm__) with
                        [ Stream.Failure -> None ]
                      with
                      [ Some t ->
                          let loc = P.gloc bp strm__ in
                          MLast.TyOlb loc i t
                      | _ ->
                          raise
                            (Stream.Error
                               (P.error "ctyp" "QUESTIONIDENT" "[ctyp]")) ]
                    }
                  | _ -> ctyp_5 strm__ ] ] ] ]
and ctyp_5 = parser bp [: a = ctyp_6; a = ctyp_5_cont bp a ! :] -> a
and ctyp_5_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "*");
       tl = P.list1sep ctyp_6 (P.token ("", "*")) False
         ? P.error "ctyp" "'*'" "[ctyp level apply]";
       loc = P.gloc bp !;
       let t = a__ in
       a = ctyp_5_cont bp (MLast.TyTup loc (Ploc.VaVal [t :: tl])) ! :] ->
      a
  | [: :] -> a__ ]
and ctyp_6 = parser bp [: a = ctyp_7; a = ctyp_6_cont bp a ! :] -> a
and ctyp_6_cont bp a__ strm__ =
  match try Some (ctyp_7 strm__) with [ Stream.Failure -> None ] with
  [ Some t2 ->
      let loc = P.gloc bp strm__ in
      let t1 = a__ in
      ctyp_6_cont bp (MLast.TyApp loc t2 t1) strm__
  | _ -> a__ ]
and ctyp_7 =
  parser bp
  [ [: t = ctyp_ident_0 :] -> t
  | [: a = ctyp_8 :] -> a ]
and ctyp_8 =
  parser bp
  [ [: `("", "[");
       a =
         parser
         [ [: `("", ">");
              a =
                parser
                [ [: `("", "]"); loc = P.gloc bp ! :] ->
                    MLast.TyVrn loc (Ploc.VaVal []) (Some None)
                | [: _ = P.option (P.token ("", "|"));
                     rfl =
                       P.vala
                         (P.list1sep poly_variant_0 (P.token ("", "|")) False)
                       ? P.error "ctyp" "" "???";
                     `("", "]") ? P.error "ctyp" "???" "']'";
                     loc = P.gloc bp ! :] ->
                    MLast.TyVrn loc rfl (Some None)
                | [: :] -> raise (Stream.Error "") ]
                ? P.error "ctyp" "'>'" "']' or '|' or ???" :] ->
             a
         | [: _ = P.option (P.token ("", "|"));
              rfl =
                P.vala (P.list1sep poly_variant_0 (P.token ("", "|")) False)
                ? P.error "ctyp" "" "???";
              `("", "]") ? P.error "ctyp" "???" "']'"; loc = P.gloc bp ! :] ->
             MLast.TyVrn loc rfl None
         | [: :] -> raise (Stream.Error "") ]
         ? P.error "ctyp" "'['" "'>' or '|' or ???" :] ->
      a
  | [: `("", "[<"); _ = P.option (P.token ("", "|"));
       rfl = P.vala (P.list1sep poly_variant_0 (P.token ("", "|")) False)
         ? P.error "ctyp" "" "???";
       a =
         parser
         [ [: `("", "]"); loc = P.gloc bp ! :] ->
             MLast.TyVrn loc rfl (Some (Some (Ploc.VaVal [])))
         | [: `("", ">");
              ntl = P.vala (P.list1 name_tag_0) ? P.error "ctyp" "'>'" "???";
              `("", "]") ? P.error "ctyp" "???" "']'"; loc = P.gloc bp ! :] ->
             MLast.TyVrn loc rfl (Some (Some ntl)) ]
         ? P.error "ctyp" "???" "']' or '>'" :] ->
      a
  | [: `("", "#");
       cli = P.vala extended_longident_lident_0
         ? P.error "ctyp" "'#'" "[extended_longident_lident]";
       loc = P.gloc bp ! :] ->
      MLast.TyCls loc cli
  | [: `("", "<");
       a =
         parser
           [: :] ->
             match Stream.npeek 2 strm__ with
             [ [("", ".."); ("", ">")] -> do {
                 Stream.junk strm__;
                 Stream.junk strm__;
                 try
                   let loc = P.gloc bp strm__ in
                   MLast.TyObj loc (Ploc.VaVal []) (Ploc.VaVal True)
                 with
                 [ Stream.Failure ->
                     raise (Stream.Error (P.error "ctyp" "'..'" "???")) ]
               }
             | _ ->
                 match strm__ with parser
                 [ [: `("", ">"); loc = P.gloc bp ! :] ->
                     MLast.TyObj loc (Ploc.VaVal []) (Ploc.VaVal False)
                 | [: ml = P.vala meth_list_0;
                      v = P.vala (P.bool (P.token ("", "..")))
                        ? P.error "ctyp" "[meth_list]" "???";
                      `("", ">") ? P.error "ctyp" "???" "'>'";
                      loc = P.gloc bp ! :] ->
                     MLast.TyObj loc ml v ] ]
         ? P.error "ctyp" "'<'" "'..' or '>' or [meth_list]" :] ->
      a
  | [: `("", "'"); i = P.vala ident_0 ? P.error "ctyp" "'''" "[ident]";
       loc = P.gloc bp ! :] ->
      MLast.TyQuo loc i
  | [: `("", "_"); loc = P.gloc bp ! :] -> MLast.TyAny loc
  | [: `("", "(");
       a =
         parser
         [ [: `("", "module");
              (ext, attrs) = ext_attributes_0
                ? P.error "ctyp" "'module'" "[ext_attributes]";
              mt = module_type_0
                ? P.error "ctyp" "[ext_attributes]" "[module_type]";
              `("", ")") ? P.error "ctyp" "[module_type]" "')'";
              loc = P.gloc bp ! :] ->
             let mt = module_type_wrap_attrs mt attrs in
             let ct = MLast.TyPck loc mt in
             ctyp_to_inline ct ext []
         | [: t = ctyp_0;
              a =
                parser
                [ [: `("", ",");
                     tl = P.list1sep ctyp_0 (P.token ("", ",")) False
                       ? P.error "ctyp" "','" "[ctyp]";
                     `("", ")") ? P.error "ctyp" "" "[ctyp] or ')'";
                     i = ctyp_7 ? P.error "ctyp" "')'" "[ctyp level ctyp2]";
                     loc = P.gloc bp ! :] ->
                    List.fold_left (fun c a -> MLast.TyApp loc c a) i
                      [t :: tl]
                | [: `("", ")") :] -> t ]
                ? P.error "ctyp" "[ctyp]" "',' or ')'" :] ->
             a ]
         ? P.error "ctyp" "'('" "'module' or [ctyp]" :] ->
      a
  | [: e = alg_extension_0; loc = P.gloc bp ! :] -> MLast.TyExten loc e ]
and ctyp_gen_cont bp a__ strm__ =
  let x = ctyp_0_cont bp a__ strm__ in
  let x = ctyp_2_cont bp x strm__ in
  let x = ctyp_3_cont bp x strm__ in
  let x = ctyp_5_cont bp x strm__ in
  let x = ctyp_6_cont bp x strm__ in
  if x == a__ then x else ctyp_gen_cont bp x strm__
and extended_longident_0 =
  parser bp
    [: a = extended_longident_1; a = extended_longident_gen_cont bp a ! :] ->
      a
and extended_longident_0_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "(");
       me2 = extended_longident_0
         ? P.error "extended_longident" "'('" "[extended_longident]";
       `("", ")") ? P.error "extended_longident" "[extended_longident]" "')'";
       loc = P.gloc bp !;
       let me1 = a__ in
       a = extended_longident_0_cont bp (MLast.LiApp loc me1 me2) ! :] ->
      a
  | [: _ = check_dot_uid :] ->
      match Stream.npeek 2 strm__ with
      [ [("", "."); ("UIDENT", i)] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          let i = Ploc.VaVal i in
          try
            let loc = P.gloc bp strm__ in
            let me1 = a__ in
            extended_longident_0_cont bp
              (let i = vala_map uident_True_True_ i in
               MLast.LiAcc loc me1 i)
              strm__
          with
          [ Stream.Failure ->
              raise
                (Stream.Error (P.error "extended_longident" "'.'" "???")) ]
        }
      | _ ->
          raise
            (Stream.Error
               (P.error "extended_longident" "[check_dot_uid]" "'.'")) ]
  | [: :] -> a__ ]
and extended_longident_1 =
  parser bp
    [: i = P.vala (P.token ("UIDENT", "")); loc = P.gloc bp ! :] ->
      let i = vala_map uident_True_True_ i in
      MLast.LiUid loc i
and extended_longident_gen_cont bp a__ strm__ =
  let x = extended_longident_0_cont bp a__ strm__ in
  if x == a__ then x else extended_longident_gen_cont bp x strm__
and class_signature_0 =
  parser bp
    [: :] ->
      match Stream.npeek 2 strm__ with
      [ [("", "let"); ("", "open")] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          match
            try Some (P.vala (P.bool (P.token ("", "!"))) strm__) with
            [ Stream.Failure -> None ]
          with
          [ Some ovf ->
              match
                try Some (extended_longident_0 strm__) with
                [ Stream.Failure -> None ]
              with
              [ Some i ->
                  match strm__ with parser
                  [ [: `("", "in");
                       ce = class_signature_0
                         ? P.error "class_signature" "'in'"
                             "[class_signature]";
                       loc = P.gloc bp !;
                       a =
                         class_signature_0_cont bp
                           (MLast.CtLop loc ovf i ce) ! :] ->
                      a
                  | [: :] ->
                      raise
                        (Stream.Error
                           (P.error "class_signature" "[extended_longident]"
                              "'in'")) ]
              | _ ->
                  raise
                    (Stream.Error
                       (P.error "class_signature" "???"
                          "[extended_longident]")) ]
          | _ ->
              raise (Stream.Error (P.error "class_signature" "'let'" "???")) ]
        }
      | _ ->
          let a = class_signature_1 strm__ in
          class_signature_0_cont bp a strm__ ]
and class_signature_0_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "[@");
       attr = P.vala attribute_body_0
         ? P.error "class_signature" "'[@'" "[attribute_body]";
       `("", "]") ? P.error "class_signature" "[attribute_body]" "']'";
       loc = P.gloc bp !;
       let ct = a__ in
       a = class_signature_0_cont bp (MLast.CtAtt loc ct attr) ! :] ->
      a
  | [: :] -> a__ ]
and class_signature_1 =
  parser bp
  [ [: `("", "[");
       tl = P.list1sep ctyp_0 (P.token ("", ",")) False
         ? P.error "class_signature" "'['" "[ctyp]";
       `("", "]") ? P.error "class_signature" "" "[ctyp] or ']'";
       id = P.orzero class_signature_1 class_signature_0
         ? P.error "class_signature" "']'" "[class_signature]";
       loc = P.gloc bp ! :] ->
      MLast.CtCon loc id (Ploc.VaVal tl)
  | [: `("", "object");
       alg_attrs = alg_attributes_no_anti_0
         ? P.error "class_signature" "'object'" "[alg_attributes_no_anti]";
       cst = P.vala (P.option class_self_type_0)
         ? P.error "class_signature" "[alg_attributes_no_anti]" "???";
       csf = P.vala (P.list0 class_sig_item_0)
         ? P.error "class_signature" "???" "???";
       `("", "end") ? P.error "class_signature" "???" "'end'";
       loc = P.gloc bp ! :] ->
      class_type_wrap_attrs (MLast.CtSig loc cst csf) alg_attrs
  | [: a = class_signature_2 :] -> a ]
and class_signature_2 =
  parser bp
  [ [: i = P.vala (P.token ("LIDENT", "")); loc = P.gloc bp ! :] ->
      MLast.CtLid loc i
  | [: li = extended_longident_0 :] ->
      match Stream.npeek 2 strm__ with
      [ [("", "."); ("LIDENT", i)] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          let i = Ploc.VaVal i in
          try
            let loc = P.gloc bp strm__ in
            MLast.CtLongLid loc li i
          with
          [ Stream.Failure ->
              raise (Stream.Error (P.error "class_signature" "'.'" "???")) ]
        }
      | _ ->
          raise
            (Stream.Error
               (P.error "class_signature" "[extended_longident]" "'.'")) ] ]
and class_signature_gen_cont bp a__ strm__ =
  let x = class_signature_0_cont bp a__ strm__ in
  if x == a__ then x else class_signature_gen_cont bp x strm__
and ident_0 =
  parser bp
  [ [: `("LIDENT", i) :] -> i
  | [: `("UIDENT", i) :] -> i ]
and simple_type_parameter_0 =
  parser bp
  [ [: `("", "'");
       i = ident_0 ? P.error "simple_type_parameter" "'''" "[ident]" :] ->
      Some i
  | [: `("", "_") :] -> None ]
and type_parameter_0 =
  parser bp
  [ [: `("", "+");
       a =
         parser
         [ [: `("", "!");
              p = P.vala simple_type_parameter_0
                ? P.error "type_parameter" "'!'"
                    "[simple_type_parameter]" :] ->
             (p, (Some True, True))
         | [: p = P.vala simple_type_parameter_0 :] ->
             (p, (Some True, False)) ]
         ? P.error "type_parameter" "'+'"
             "'!' or [simple_type_parameter]" :] ->
      a
  | [: `("", "-");
       a =
         parser
         [ [: `("", "!");
              p = P.vala simple_type_parameter_0
                ? P.error "type_parameter" "'!'"
                    "[simple_type_parameter]" :] ->
             (p, (Some False, True))
         | [: p = P.vala simple_type_parameter_0 :] ->
             (p, (Some False, False)) ]
         ? P.error "type_parameter" "'-'"
             "'!' or [simple_type_parameter]" :] ->
      a
  | [: `("", "!");
       a =
         parser
         [ [: `("", "+");
              p = P.vala simple_type_parameter_0
                ? P.error "type_parameter" "'+'"
                    "[simple_type_parameter]" :] ->
             (p, (Some True, True))
         | [: `("", "-");
              p = P.vala simple_type_parameter_0
                ? P.error "type_parameter" "'-'"
                    "[simple_type_parameter]" :] ->
             (p, (Some False, True))
         | [: p = P.vala simple_type_parameter_0 :] -> (p, (None, True)) ]
         ? P.error "type_parameter" "'!'"
             "'+' or '-' or [simple_type_parameter]" :] ->
      a
  | [: `("", "!+");
       p = P.vala simple_type_parameter_0
         ? P.error "type_parameter" "'!+'" "[simple_type_parameter]" :] ->
      (p, (Some True, True))
  | [: `("", "+!");
       p = P.vala simple_type_parameter_0
         ? P.error "type_parameter" "'+!'" "[simple_type_parameter]" :] ->
      (p, (Some True, True))
  | [: `("", "!-");
       p = P.vala simple_type_parameter_0
         ? P.error "type_parameter" "'!-'" "[simple_type_parameter]" :] ->
      (p, (Some False, True))
  | [: `("", "-!");
       p = P.vala simple_type_parameter_0
         ? P.error "type_parameter" "'-!'" "[simple_type_parameter]" :] ->
      (p, (Some False, True))
  | [: p = P.vala simple_type_parameter_0 :] -> (p, (None, False)) ]
and class_type_parameters_0 =
  parser bp
  [ [: `("", "[");
       tpl = P.vala (P.list1sep type_parameter_0 (P.token ("", ",")) False)
         ? P.error "class_type_parameters" "'['" "???";
       `("", "]") ? P.error "class_type_parameters" "???" "']'";
       loc = P.gloc bp ! :] ->
      (loc, tpl)
  | [: loc = P.gloc bp :] -> (loc, Ploc.VaVal []) ]
and alg_attribute_0 =
  parser bp
    [: `("", "[@");
       attr = P.vala attribute_body_0
         ? P.error "alg_attribute" "'[@'" "[attribute_body]";
       `("", "]") ? P.error "alg_attribute" "[attribute_body]" "']'" :] ->
      attr
and alg_attributes_no_anti_0 =
  parser bp
  [ [: l = P.list0 alg_attribute_0 :] -> l
  | [: :] -> raise (Stream.Error "") ]
and class_type_declaration_0 =
  parser bp
    [: alg_attrs = alg_attributes_no_anti_0;
       vf = P.vala (P.bool (P.token ("", "virtual")))
         ? P.error "class_type_declaration" "[alg_attributes_no_anti]" "???";
       ctp = class_type_parameters_0
         ? P.error "class_type_declaration" "???"
             "[class_type_parameters]" :] ->
      match Stream.npeek 2 strm__ with
      [ [("LIDENT", n); ("", "=")] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          let n = Ploc.VaVal n in
          match
            try Some (class_signature_0 strm__) with
            [ Stream.Failure -> None ]
          with
          [ Some cs ->
              match
                try Some (item_attributes_0 strm__) with
                [ Stream.Failure -> None ]
              with
              [ Some item_attrs ->
                  let loc = P.gloc bp strm__ in
                  let attrs =
                    merge_left_auxiliary_attrs
                      ~{nonterm_name = "class-type-decl"}
                      ~{left_name = "algebraic attributes"}
                      ~{right_name = "item attributes"} alg_attrs item_attrs
                  in
                  {MLast.ciLoc = loc; MLast.ciVir = vf; MLast.ciPrm = ctp;
                   MLast.ciNam = n; MLast.ciExp = cs;
                   MLast.ciAttributes = attrs}
              | _ ->
                  raise
                    (Stream.Error
                       (P.error "class_type_declaration" "[class_signature]"
                          "[item_attributes]")) ]
          | _ ->
              raise
                (Stream.Error
                   (P.error "class_type_declaration" "lowercase identifier"
                      "[class_signature]")) ]
        }
      | _ ->
          raise
            (Stream.Error
               (P.error "class_type_declaration" "[class_type_parameters]"
                  "lowercase identifier")) ]
and ext_opt_0 =
  parser bp
  [ [: ext =
         P.option
           (parser
              [: `("", "%");
                 id = attribute_id_0
                   ? P.error "implem" "'%'" "[attribute_id]" :] ->
                id) :] ->
      ext
  | [: :] -> raise (Stream.Error "") ]
and sig_item_0 =
  parser bp
  [ [: `("", "class");
       a =
         parser
         [ [: `("", "type");
              ext = ext_opt_0 ? P.error "sig_item" "'type'" "[ext_opt]";
              ctd =
                P.vala
                  (P.list1sep class_type_declaration_0 (P.token ("", "and"))
                     False)
                ? P.error "sig_item" "[ext_opt]" "???";
              loc = P.gloc bp ! :] ->
             sig_item_to_inline (MLast.SgClt loc ctd) ext
         | [: ext = ext_opt_0;
              cd =
                P.vala
                  (P.list1sep class_description_0 (P.token ("", "and")) False)
                ? P.error "sig_item" "[ext_opt]" "???";
              loc = P.gloc bp ! :] ->
             sig_item_to_inline (MLast.SgCls loc cd) ext ]
         ? P.error "sig_item" "'class'" "'type' or [ext_opt]" :] ->
      a
  | [: `("", "exception");
       (ext, alg_attrs0) = ext_attributes_0
         ? P.error "sig_item" "'exception'" "[ext_attributes]";
       gc = constructor_declaration_0
         ? P.error "sig_item" "[ext_attributes]" "[constructor_declaration]";
       item_attrs = item_attributes_0
         ? P.error "sig_item" "[constructor_declaration]" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      let (x1, x2, x3, x4, x5, alg_attrs1) = gc in
      let alg_attrs =
        merge_left_auxiliary_attrs ~{nonterm_name = "sig_item-exception"}
          ~{left_name = "algebraic attributes"}
          ~{right_name = "item attributes"} alg_attrs0 alg_attrs1
      in
      let gc = (x1, x2, x3, x4, x5, alg_attrs) in
      sig_item_to_inline (MLast.SgExc loc gc item_attrs) ext
  | [: `("", "external");
       (ext, alg_attrs) = ext_attributes_0
         ? P.error "sig_item" "'external'" "[ext_attributes]";
       a =
         parser
         [ [: `("", "(");
              i = operator_rparen
                ? P.error "sig_item" "'('" "[operator_rparen]";
              `("", ":") ? P.error "sig_item" "[operator_rparen]" "':'";
              ls = type_binder_opt_0
                ? P.error "sig_item" "':'" "[type_binder_opt]";
              t = ctyp_0 ? P.error "sig_item" "[type_binder_opt]" "[ctyp]";
              `("", "=") ? P.error "sig_item" "[ctyp]" "'='";
              pd = P.vala (P.list1 (P.token ("STRING", "")))
                ? P.error "sig_item" "'='" "???";
              item_attrs = item_attributes_0
                ? P.error "sig_item" "???" "[item_attributes]";
              loc = P.gloc bp ! :] ->
             let attrs =
               merge_left_auxiliary_attrs
                 ~{nonterm_name = "sig_item-external"}
                 ~{left_name = "algebraic attributes"}
                 ~{right_name = "item attributes"} alg_attrs item_attrs
             in
             sig_item_to_inline (MLast.SgExt loc (Ploc.VaVal i) ls t pd attrs)
               ext
         | [: :] ->
             match Stream.npeek 2 strm__ with
             [ [("LIDENT", i); ("", ":")] -> do {
                 Stream.junk strm__;
                 Stream.junk strm__;
                 let i = Ploc.VaVal i in
                 match
                   try Some (type_binder_opt_0 strm__) with
                   [ Stream.Failure -> None ]
                 with
                 [ Some ls ->
                     match
                       try Some (ctyp_0 strm__) with
                       [ Stream.Failure -> None ]
                     with
                     [ Some t ->
                         match strm__ with parser
                         [ [: `("", "=");
                              pd = P.vala (P.list1 (P.token ("STRING", "")))
                                ? P.error "sig_item" "'='" "???";
                              item_attrs = item_attributes_0
                                ? P.error "sig_item" "???"
                                    "[item_attributes]";
                              loc = P.gloc bp ! :] ->
                             let attrs =
                               merge_left_auxiliary_attrs
                                 ~{nonterm_name = "sig_item-external"}
                                 ~{left_name = "algebraic attributes"}
                                 ~{right_name = "item attributes"} alg_attrs
                                 item_attrs
                             in
                             sig_item_to_inline
                               (MLast.SgExt loc i ls t pd attrs) ext
                         | [: :] ->
                             raise
                               (Stream.Error
                                  (P.error "sig_item" "[ctyp]" "'='")) ]
                     | _ ->
                         raise
                           (Stream.Error
                              (P.error "sig_item" "[type_binder_opt]"
                                 "[ctyp]")) ]
                 | _ ->
                     raise
                       (Stream.Error
                          (P.error "sig_item" "lowercase identifier"
                             "[type_binder_opt]")) ]
               }
             | _ -> raise Stream.Failure ] ]
         ? P.error "sig_item" "[ext_attributes]"
             "'(' or lowercase identifier" :] ->
      a
  | [: `("", "include");
       (ext, alg_attrs) = ext_attributes_0
         ? P.error "sig_item" "'include'" "[ext_attributes]";
       mt = module_type_0
         ? P.error "sig_item" "[ext_attributes]" "[module_type]";
       item_attrs = item_attributes_0
         ? P.error "sig_item" "[module_type]" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      let attrs =
        merge_left_auxiliary_attrs ~{nonterm_name = "sig_item-include"}
          ~{left_name = "algebraic attributes"}
          ~{right_name = "item attributes"} alg_attrs item_attrs
      in
      sig_item_to_inline (MLast.SgInc loc mt attrs) ext
  | [: `("", "module");
       a =
         parser
         [ [: `("", "type");
              (ext, alg_attrs) = ext_attributes_0
                ? P.error "sig_item" "'type'" "[ext_attributes]";
              i = P.vala ident_0
                ? P.error "sig_item" "[ext_attributes]" "[ident]";
              a =
                parser
                [ [: `("", "=");
                     mt = module_type_0
                       ? P.error "sig_item" "'='" "[module_type]";
                     item_attrs = item_attributes_0
                       ? P.error "sig_item" "[module_type]"
                           "[item_attributes]";
                     loc = P.gloc bp ! :] ->
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "sig_item-module-type"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    sig_item_to_inline (MLast.SgMty loc i mt attrs) ext
                | [: `("", ":=");
                     mt = module_type_0
                       ? P.error "sig_item" "':='" "[module_type]";
                     item_attrs = item_attributes_0
                       ? P.error "sig_item" "[module_type]"
                           "[item_attributes]";
                     loc = P.gloc bp ! :] ->
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "sig_item-module-type"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    sig_item_to_inline (MLast.SgMtySubst loc i mt attrs) ext
                | [: item_attrs = item_attributes_0; loc = P.gloc bp ! :] ->
                    let attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "sig_item-module-type"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    sig_item_to_inline
                      (MLast.SgMty loc i
                         (MLast.MtQuo loc (Ploc.VaVal "abstract")) attrs)
                      ext ]
                ? P.error "sig_item" "[ident]"
                    "'=' or ':=' or [item_attributes]" :] ->
             a
         | [: _ = check_uident_coloneq :] ->
             match Stream.npeek 2 strm__ with
             [ [("UIDENT", i); ("", ":=")] -> do {
                 Stream.junk strm__;
                 Stream.junk strm__;
                 let i = Ploc.VaVal i in
                 match
                   try Some (extended_longident_0 strm__) with
                   [ Stream.Failure -> None ]
                 with
                 [ Some li ->
                     match
                       try Some (item_attributes_0 strm__) with
                       [ Stream.Failure -> None ]
                     with
                     [ Some attrs ->
                         let loc = P.gloc bp strm__ in
                         MLast.SgModSubst loc i li attrs
                     | _ ->
                         raise
                           (Stream.Error
                              (P.error "sig_item" "[extended_longident]"
                                 "[item_attributes]")) ]
                 | _ ->
                     raise
                       (Stream.Error
                          (P.error "sig_item" "uppercase identifier"
                             "[extended_longident]")) ]
               }
             | _ ->
                 raise
                   (Stream.Error
                      (P.error "sig_item" "[check_uident_coloneq]"
                         "uppercase identifier")) ]
         | [: (ext, alg_attrs) = ext_attributes_0;
              a =
                parser
                [ [: _ = check_module_not_alias;
                     rf = P.bool (P.token ("", "rec"));
                     h = first_mod_decl_binding_0
                       ? P.error "sig_item" "" "[first_mod_decl_binding]";
                     t = P.list0 rest_mod_decl_binding_0;
                     loc = P.gloc bp ! :] ->
                    let (i, mt, item_attrs) = h in
                    let item_attrs =
                      merge_left_auxiliary_attrs
                        ~{nonterm_name = "sig_item-module"}
                        ~{left_name = "algebraic attributes"}
                        ~{right_name = "item attributes"} alg_attrs item_attrs
                    in
                    let h = (i, mt, item_attrs) in
                    sig_item_to_inline
                      (MLast.SgMod loc (Ploc.VaVal rf) (Ploc.VaVal [h :: t]))
                      ext
                | [: _ = check_module_alias :] ->
                    match Stream.npeek 2 strm__ with
                    [ [("UIDENT", i); ("", "=")] -> do {
                        Stream.junk strm__;
                        Stream.junk strm__;
                        let i = Ploc.VaVal i in
                        match
                          try Some (longident_0 strm__) with
                          [ Stream.Failure -> None ]
                        with
                        [ Some li ->
                            match
                              try Some (item_attributes_0 strm__) with
                              [ Stream.Failure -> None ]
                            with
                            [ Some item_attrs ->
                                let loc = P.gloc bp strm__ in
                                let attrs =
                                  merge_left_auxiliary_attrs
                                    ~{nonterm_name = "sig_item-module"}
                                    ~{left_name = "algebraic attributes"}
                                    ~{right_name = "item attributes"}
                                    alg_attrs item_attrs
                                in
                                sig_item_to_inline
                                  (MLast.SgMtyAlias loc i (Ploc.VaVal li)
                                     attrs)
                                  ext
                            | _ ->
                                raise
                                  (Stream.Error
                                     (P.error "sig_item" "[longident]"
                                        "[item_attributes]")) ]
                        | _ ->
                            raise
                              (Stream.Error
                                 (P.error "sig_item" "uppercase identifier"
                                    "[longident]")) ]
                      }
                    | _ ->
                        raise
                          (Stream.Error
                             (P.error "sig_item" "[check_module_alias]"
                                "uppercase identifier")) ] ]
                ? P.error "sig_item" "[ext_attributes]"
                    "[check_module_not_alias] or [check_module_alias]" :] ->
             a ]
         ? P.error "sig_item" "'module'"
             "'type' or [check_uident_coloneq] or [ext_attributes]" :] ->
      a
  | [: `("", "open");
       (ext, alg_attrs) = ext_attributes_0
         ? P.error "sig_item" "'open'" "[ext_attributes]";
       i = extended_longident_0
         ? P.error "sig_item" "[ext_attributes]" "[extended_longident]";
       item_attrs = item_attributes_0
         ? P.error "sig_item" "[extended_longident]" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      let attrs =
        merge_left_auxiliary_attrs ~{nonterm_name = "sig_item-open"}
          ~{left_name = "algebraic attributes"}
          ~{right_name = "item attributes"} alg_attrs item_attrs
      in
      sig_item_to_inline (MLast.SgOpn loc i attrs) ext
  | [: `("", "type");
       (ext, attrs) = ext_attributes_0
         ? P.error "sig_item" "'type'" "[ext_attributes]";
       a =
         parser
         [ [: _ = check_type_decl;
              nr = P.vala (P.bool (P.token ("", "nonrec")))
                ? P.error "sig_item" "[check_type_decl]" "???";
              htd = first_type_decl_0
                ? P.error "sig_item" "???" "[first_type_decl]";
              ttd = P.list0 rest_type_decl_0; loc = P.gloc bp ! :] ->
             let attrs =
               merge_left_auxiliary_attrs
                 ~{nonterm_name = "str_item-type_decl"}
                 ~{left_name = "algebraic attributes"}
                 ~{right_name = "item attributes"} attrs
                 htd.MLast.tdAttributes
             in
             let htd = {(htd) with MLast.tdAttributes = attrs} in
             sig_item_to_inline (MLast.SgTyp loc nr (Ploc.VaVal [htd :: ttd]))
               ext
         | [: _ = check_type_extension;
              te = type_extension_0
                ? P.error "sig_item" "[check_type_extension]"
                    "[type_extension]";
              loc = P.gloc bp ! :] ->
             let attrs =
               merge_left_auxiliary_attrs
                 ~{nonterm_name = "str_item-type_extension"}
                 ~{left_name = "algebraic attributes"}
                 ~{right_name = "item attributes"} attrs te.MLast.teAttributes
             in
             let te = {(te) with MLast.teAttributes = attrs} in
             sig_item_to_inline
               (MLast.SgTypExten loc
                  {MLast.teNam = te.MLast.teNam; MLast.tePrm = te.MLast.tePrm;
                   MLast.tePrv = te.MLast.tePrv; MLast.teECs = te.MLast.teECs;
                   MLast.teAttributes = te.MLast.teAttributes})
               ext ]
         ? P.error "sig_item" "[ext_attributes]"
             "[check_type_decl] or [check_type_extension]" :] ->
      a
  | [: `("", "val");
       (ext, attrs1) = ext_attributes_0
         ? P.error "sig_item" "'val'" "[ext_attributes]";
       a =
         parser
         [ [: `("", "(");
              i = operator_rparen
                ? P.error "sig_item" "'('" "[operator_rparen]";
              `("", ":") ? P.error "sig_item" "[operator_rparen]" "':'";
              ls = type_binder_opt_0
                ? P.error "sig_item" "':'" "[type_binder_opt]";
              t = ctyp_0 ? P.error "sig_item" "[type_binder_opt]" "[ctyp]";
              attrs2 = item_attributes_0
                ? P.error "sig_item" "[ctyp]" "[item_attributes]";
              loc = P.gloc bp ! :] ->
             let attrs =
               merge_left_auxiliary_attrs ~{nonterm_name = "sig_item"}
                 ~{left_name = "algebraic attributes"}
                 ~{right_name = "item attributes"} attrs1 attrs2
             in
             let t =
               match Pcaml.unvala ls with
               [ [] -> t
               | _ -> MLast.TyPol loc ls t ]
             in
             sig_item_to_inline (MLast.SgVal loc (Ploc.VaVal i) t attrs) ext
         | [: :] ->
             match Stream.npeek 2 strm__ with
             [ [("LIDENT", i); ("", ":")] -> do {
                 Stream.junk strm__;
                 Stream.junk strm__;
                 let i = Ploc.VaVal i in
                 match
                   try Some (type_binder_opt_0 strm__) with
                   [ Stream.Failure -> None ]
                 with
                 [ Some ls ->
                     match
                       try Some (ctyp_0 strm__) with
                       [ Stream.Failure -> None ]
                     with
                     [ Some t ->
                         match
                           try Some (item_attributes_0 strm__) with
                           [ Stream.Failure -> None ]
                         with
                         [ Some attrs2 ->
                             let loc = P.gloc bp strm__ in
                             let attrs =
                               merge_left_auxiliary_attrs
                                 ~{nonterm_name = "sig_item"}
                                 ~{left_name = "algebraic attributes"}
                                 ~{right_name = "item attributes"} attrs1
                                 attrs2
                             in
                             let t =
                               match Pcaml.unvala ls with
                               [ [] -> t
                               | _ -> MLast.TyPol loc ls t ]
                             in
                             sig_item_to_inline (MLast.SgVal loc i t attrs)
                               ext
                         | _ ->
                             raise
                               (Stream.Error
                                  (P.error "sig_item" "[ctyp]"
                                     "[item_attributes]")) ]
                     | _ ->
                         raise
                           (Stream.Error
                              (P.error "sig_item" "[type_binder_opt]"
                                 "[ctyp]")) ]
                 | _ ->
                     raise
                       (Stream.Error
                          (P.error "sig_item" "lowercase identifier"
                             "[type_binder_opt]")) ]
               }
             | _ -> raise Stream.Failure ] ]
         ? P.error "sig_item" "[ext_attributes]"
             "'(' or lowercase identifier" :] ->
      a
  | [: attr = floating_attribute_0; loc = P.gloc bp ! :] ->
      MLast.SgFlAtt loc attr
  | [: e = item_extension_0;
       attrs = item_attributes_0
         ? P.error "sig_item" "[item_extension]" "[item_attributes]";
       loc = P.gloc bp ! :] ->
      MLast.SgExten loc e attrs ]
and attribute_signature_0 =
  parser bp
    [: st =
         P.vala
           (P.list1
              (parser
                 [: s = sig_item_0; _ = P.option (P.token ("", ";;")) :] ->
                   s)) :] ->
      st
and kwd_attribute_id_0 =
  parser bp
    [: s =
         parser
         [ [: `("", ("and" as x)) :] -> x
         | [: `("", ("as" as x)) :] -> x
         | [: `("", ("assert" as x)) :] -> x
         | [: `("", ("begin" as x)) :] -> x
         | [: `("", ("class" as x)) :] -> x
         | [: `("", ("constraint" as x)) :] -> x
         | [: `("", ("do" as x)) :] -> x
         | [: `("", ("done" as x)) :] -> x
         | [: `("", ("downto" as x)) :] -> x
         | [: `("", ("else" as x)) :] -> x
         | [: `("", ("end" as x)) :] -> x
         | [: `("", ("exception" as x)) :] -> x
         | [: `("", ("external" as x)) :] -> x
         | [: `("", ("false" as x)) :] -> x
         | [: `("", ("for" as x)) :] -> x
         | [: `("", ("fun" as x)) :] -> x
         | [: `("", ("function" as x)) :] -> x
         | [: `("", ("functor" as x)) :] -> x
         | [: `("", ("if" as x)) :] -> x
         | [: `("", ("in" as x)) :] -> x
         | [: `("", ("include" as x)) :] -> x
         | [: `("", ("inherit" as x)) :] -> x
         | [: `("", ("initializer" as x)) :] -> x
         | [: `("", ("lazy" as x)) :] -> x
         | [: `("", ("let" as x)) :] -> x
         | [: `("", ("match" as x)) :] -> x
         | [: `("", ("method" as x)) :] -> x
         | [: `("", ("module" as x)) :] -> x
         | [: `("", ("mutable" as x)) :] -> x
         | [: `("", ("new" as x)) :] -> x
         | [: `("", ("nonrec" as x)) :] -> x
         | [: `("", ("object" as x)) :] -> x
         | [: `("", ("of" as x)) :] -> x
         | [: `("", ("open" as x)) :] -> x
         | [: `("", ("or" as x)) :] -> x
         | [: `("", ("private" as x)) :] -> x
         | [: `("", ("rec" as x)) :] -> x
         | [: `("", ("sig" as x)) :] -> x
         | [: `("", ("struct" as x)) :] -> x
         | [: `("", ("then" as x)) :] -> x
         | [: `("", ("to" as x)) :] -> x
         | [: `("", ("true" as x)) :] -> x
         | [: `("", ("try" as x)) :] -> x
         | [: `("", ("type" as x)) :] -> x
         | [: `("", ("val" as x)) :] -> x
         | [: `("", ("virtual" as x)) :] -> x
         | [: `("", ("when" as x)) :] -> x
         | [: `("", ("while" as x)) :] -> x
         | [: `("", ("with" as x)) :] -> x ] :] ->
      s
and attribute_id_0 =
  parser bp
  [ [: l =
         P.list1sep
           (parser
            [ [: `("LIDENT", i) :] -> i
            | [: `("UIDENT", i) :] -> i ])
           (P.token ("", ".")) False;
       loc = P.gloc bp ! :] ->
      (loc, String.concat "." l)
  | [: s = kwd_attribute_id_0; loc = P.gloc bp ! :] -> (loc, s) ]
and attribute_body_0 =
  parser bp
    [: id = P.vala attribute_id_0;
       a =
         parser
         [ [: `("", ":");
              a =
                parser
                [ [: si = attribute_signature_0; loc = P.gloc bp ! :] ->
                    (id, MLast.SiAttr loc si)
                | [: ty = P.vala ctyp_0; loc = P.gloc bp ! :] ->
                    (id, MLast.TyAttr loc ty) ]
                ? P.error "attribute_body" "':'"
                    "[attribute_signature] or [ctyp]" :] ->
             a
         | [: `("", "?");
              p = P.vala patt_0 ? P.error "attribute_body" "'?'" "[patt]";
              a =
                parser
                [ [: `("", "when");
                     e = P.vala expr_0
                       ? P.error "attribute_body" "'when'" "[expr]";
                     loc = P.gloc bp ! :] ->
                    (id, MLast.PaAttr loc p (Some e))
                | [: loc = P.gloc bp :] -> (id, MLast.PaAttr loc p None) ]
                ? P.error "attribute_body" "[patt]" "'when'" :] ->
             a
         | [: st = attribute_structure_0; loc = P.gloc bp ! :] ->
             (id, MLast.StAttr loc st)
         | [: loc = P.gloc bp :] -> (id, MLast.StAttr loc (Ploc.VaVal [])) ]
         ? P.error "attribute_body" "[attribute_id]"
             "':' or '?' or [attribute_structure]" :] ->
      a
and patt_0 = parser bp [: a = patt_1; a = patt_gen_cont bp a ! :] -> a
and patt_0_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "as");
       a =
         parser
         [ [: `("", "(");
              i = operator_rparen ? P.error "patt" "'('" "[operator_rparen]";
              loc = P.gloc bp !;
              let p1 = a__ in
              a =
                patt_0_cont bp
                  (MLast.PaAli loc p1
                     (MLast.PaLid loc (Ploc.VaVal i))) ! :] ->
             a
         | [: `("LIDENT", i); loc = P.gloc bp !;
              let p1 = a__ in
              a =
                patt_0_cont bp
                  (MLast.PaAli loc p1
                     (MLast.PaLid loc (Ploc.VaVal i))) ! :] ->
             a ]
         ? P.error "patt" "'as'" "'(' or lowercase identifier" :] ->
      a
  | [: :] -> a__ ]
and patt_1 = parser bp [: a = patt_2; a = patt_1_cont bp a ! :] -> a
and patt_1_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "|");
       p2 = P.orzero patt_2 patt_0 ? P.error "patt" "'|'" "[patt]";
       loc = P.gloc bp !;
       let p1 = a__ in
       a = patt_1_cont bp (MLast.PaOrp loc p1 p2) ! :] ->
      a
  | [: :] -> a__ ]
and patt_2 = parser bp [: a = patt_3; a = patt_2_cont bp a ! :] -> a
and patt_2_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", ",");
       pl = P.list1sep (P.orzero patt_3 patt_0) (P.token ("", ",")) False
         ? P.error "patt" "','" "[patt]";
       loc = P.gloc bp !;
       let p = a__ in
       a = patt_2_cont bp (MLast.PaTup loc (Ploc.VaVal [p :: pl])) ! :] ->
      a
  | [: :] -> a__ ]
and patt_3 = parser bp [: a = patt_4; a = patt_3_cont bp a ! :] -> a
and patt_3_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "[@");
       attr = P.vala attribute_body_0
         ? P.error "patt" "'[@'" "[attribute_body]";
       `("", "]") ? P.error "patt" "[attribute_body]" "']'";
       loc = P.gloc bp !;
       let p = a__ in
       a = patt_3_cont bp (MLast.PaAtt loc p attr) ! :] ->
      a
  | [: :] -> a__ ]
and patt_4 =
  parser bp
  [ [: `("", "exception");
       (ext, attrs) = ext_attributes_0
         ? P.error "patt" "'exception'" "[ext_attributes]";
       p = P.orzero patt_4 patt_0
         ? P.error "patt" "[ext_attributes]" "[patt]";
       loc = P.gloc bp ! :] ->
      patt_to_inline (MLast.PaExc loc p) ext attrs
  | [: a = patt_5 :] -> a ]
and patt_5 = parser bp [: a = patt_6; a = patt_5_cont bp a ! :] -> a
and patt_5_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "..");
       p2 = P.orzero patt_6 patt_0 ? P.error "patt" "'..'" "[patt]";
       let p1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.PaRng loc p1 p2
  | [: :] -> a__ ]
and patt_6 = parser bp [: a = patt_7; a = patt_6_cont bp a ! :] -> a
and patt_6_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "::");
       p2 = P.orzero patt_6 patt_0 ? P.error "patt" "'::'" "[patt]";
       let p1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.PaApp loc
        (MLast.PaApp loc
           (MLast.PaLong loc (MLast.LiUid loc (Ploc.VaVal "::"))
              (Ploc.VaVal []))
           p1)
        p2
  | [: :] -> a__ ]
and patt_7 =
  parser bp
  [ [: `("", "lazy");
       (ext, attrs) = ext_attributes_0
         ? P.error "patt" "'lazy'" "[ext_attributes]";
       p = P.orzero patt_7 patt_0
         ? P.error "patt" "[ext_attributes]" "[patt]";
       loc = P.gloc bp !;
       a =
         patt_7_cont bp (patt_to_inline (MLast.PaLaz loc p) ext attrs) ! :] ->
      a
  | [: a = patt_8; a = patt_7_cont bp a ! :] -> a ]
and patt_7_cont bp a__ strm__ =
  match try Some (patt_8 strm__) with [ Stream.Failure -> None ] with
  [ Some p2 ->
      let loc = P.gloc bp strm__ in
      let p1 = a__ in
      patt_7_cont bp
        (let (p1, p2) =
           match p1 with
           [ MLast.PaApp _ p11 p12 -> (p11, MLast.PaApp loc p12 p2)
           | _ -> (p1, p2) ]
         in
         match constr_patt_arity loc p1 with
         [ 1 -> MLast.PaApp loc p1 p2
         | n ->
             let p2 =
               match p2 with
               [ MLast.PaAny _ when n > 1 ->
                   let pl =
                     loop n where rec loop n =
                       if n = 0 then [] else [MLast.PaAny loc :: loop (n - 1)]
                   in
                   MLast.PaTup loc (Ploc.VaVal pl)
               | _ -> p2 ]
             in
             match p2 with
             [ MLast.PaTup _ (Ploc.VaVal pl) ->
                 List.fold_left (fun p1 p2 -> MLast.PaApp loc p1 p2) p1 pl
             | _ -> MLast.PaApp loc p1 p2 ] ])
        strm__
  | _ -> a__ ]
and patt_8 =
  parser bp
  [ [: `("", "`"); s = P.vala ident_0 ? P.error "patt" "'`'" "[ident]";
       loc = P.gloc bp ! :] ->
      MLast.PaVrn loc s
  | [: `("", "#");
       lili = P.vala extended_longident_lident_0
         ? P.error "patt" "'#'" "[extended_longident_lident]";
       loc = P.gloc bp ! :] ->
      MLast.PaTyp loc lili
  | [: `("", "+");
       a =
         parser
         [ [: s = P.vala (P.token ("INT", "")); loc = P.gloc bp ! :] ->
             MLast.PaInt loc s ""
         | [: s = P.vala (P.token ("FLOAT", "")); loc = P.gloc bp ! :] ->
             MLast.PaFlo loc s ]
         ? P.error "patt" "'+'" "integer or float" :] ->
      a
  | [: `("", "-");
       a =
         parser
         [ [: `("INT", s); loc = P.gloc bp ! :] ->
             MLast.PaInt loc (Ploc.VaVal ("-" ^ s)) ""
         | [: `("INT_l", s); loc = P.gloc bp ! :] ->
             MLast.PaInt loc (Ploc.VaVal ("-" ^ s)) "l"
         | [: `("INT_L", s); loc = P.gloc bp ! :] ->
             MLast.PaInt loc (Ploc.VaVal ("-" ^ s)) "L"
         | [: `("INT_n", s); loc = P.gloc bp ! :] ->
             MLast.PaInt loc (Ploc.VaVal ("-" ^ s)) "n"
         | [: `("FLOAT", s); loc = P.gloc bp ! :] ->
             MLast.PaFlo loc (Ploc.VaVal ("-" ^ s)) ]
         ? P.error "patt" "'-'"
             "integer or INT_l or INT_L or INT_n or float" :] ->
      a
  | [: `("UIDENT", "True"); loc = P.gloc bp ! :] ->
      MLast.PaLong loc (MLast.LiUid loc (Ploc.VaVal "True_")) (Ploc.VaVal [])
  | [: `("UIDENT", "False"); loc = P.gloc bp ! :] ->
      MLast.PaLong loc (MLast.LiUid loc (Ploc.VaVal "False_")) (Ploc.VaVal [])
  | [: `("", "false"); loc = P.gloc bp ! :] ->
      MLast.PaLong loc (MLast.LiUid loc (Ploc.VaVal "False")) (Ploc.VaVal [])
  | [: `("", "true"); loc = P.gloc bp ! :] ->
      MLast.PaLong loc (MLast.LiUid loc (Ploc.VaVal "True")) (Ploc.VaVal [])
  | [: `("", "[");
       a =
         parser
         [ [: `("", "]"); loc = P.gloc bp ! :] ->
             MLast.PaLong loc (MLast.LiUid loc (Ploc.VaVal "[]"))
               (Ploc.VaVal [])
         | [: pl = patt_semi_list_0;
              `("", "]") ? P.error "patt" "[patt_semi_list]" "']'";
              loc = P.gloc bp ! :] ->
             mklistpat loc None pl ]
         ? P.error "patt" "'['" "']' or [patt_semi_list]" :] ->
      a
  | [: `("", "[|");
       a =
         parser
         [ [: `("", "|]"); loc = P.gloc bp ! :] ->
             MLast.PaArr loc (Ploc.VaVal [])
         | [: pl = P.vala patt_semi_list_0;
              `("", "|]") ? P.error "patt" "[patt_semi_list]" "'|]'";
              loc = P.gloc bp ! :] ->
             MLast.PaArr loc pl ]
         ? P.error "patt" "'[|'" "'|]' or [patt_semi_list]" :] ->
      a
  | [: `("", "{");
       lpl = P.vala lbl_patt_list_0 ? P.error "patt" "'{'" "[lbl_patt_list]";
       `("", "}") ? P.error "patt" "[lbl_patt_list]" "'}'";
       loc = P.gloc bp ! :] ->
      MLast.PaRec loc lpl
  | [: `("", "(");
       a =
         parser
         [ [: `("", ")"); loc = P.gloc bp ! :] ->
             MLast.PaLong loc (MLast.LiUid loc (Ploc.VaVal "()"))
               (Ploc.VaVal [])
         | [: :] ->
             match Stream.npeek 3 strm__ with
             [ [("", "type"); ("LIDENT", s); ("", ")")] -> do {
                 Stream.junk strm__;
                 Stream.junk strm__;
                 Stream.junk strm__;
                 let s = Ploc.VaVal s in
                 try
                   let loc = P.gloc bp strm__ in
                   MLast.PaNty loc s
                 with
                 [ Stream.Failure ->
                     raise (Stream.Error (P.error "patt" "'type'" "???")) ]
               }
             | _ ->
                 match strm__ with parser
                 [ [: `("", "module");
                      s = P.vala uidopt_0
                        ? P.error "patt" "'module'" "[uidopt]";
                      a =
                        parser
                        [ [: `("", ":");
                             mt = module_type_0
                               ? P.error "patt" "':'" "[module_type]";
                             `("", ")")
                               ? P.error "patt" "[module_type]" "')'";
                             loc = P.gloc bp ! :] ->
                            MLast.PaUnp loc s (Some mt)
                        | [: `("", ")"); loc = P.gloc bp ! :] ->
                            MLast.PaUnp loc s None ]
                        ? P.error "patt" "[uidopt]" "':' or ')'" :] ->
                     a
                 | [: op = operator_rparen; loc = P.gloc bp ! :] ->
                     if op = "::" then
                       MLast.PaLong loc (MLast.LiUid loc (Ploc.VaVal op))
                         (Ploc.VaVal [])
                     else MLast.PaLid loc (Ploc.VaVal op)
                 | [: pl = P.vala p_phony;
                      `("", ")") ? P.error "patt" "[p_phony]" "')'";
                      loc = P.gloc bp ! :] ->
                     MLast.PaTup loc pl
                 | [: p = patt_0;
                      a =
                        parser
                        [ [: `("", ":");
                             t = ctyp_0 ? P.error "patt" "':'" "[ctyp]";
                             `("", ")") ? P.error "patt" "[ctyp]" "')'";
                             loc = P.gloc bp ! :] ->
                            MLast.PaTyc loc p t
                        | [: `("", ")") :] -> p ]
                        ? P.error "patt" "[patt]" "':' or ')'" :] ->
                     a ] ] ]
         ? P.error "patt" "'('"
             "')' or 'type' or 'module' or [operator_rparen] or [p_phony] or [patt]" :] ->
      a
  | [: `("", "_"); loc = P.gloc bp ! :] -> MLast.PaAny loc
  | [: s = P.vala (P.token ("INT", "")); loc = P.gloc bp ! :] ->
      MLast.PaInt loc s ""
  | [: s = P.vala (P.token ("INT_l", "")); loc = P.gloc bp ! :] ->
      MLast.PaInt loc s "l"
  | [: s = P.vala (P.token ("INT_L", "")); loc = P.gloc bp ! :] ->
      MLast.PaInt loc s "L"
  | [: s = P.vala (P.token ("INT_n", "")); loc = P.gloc bp ! :] ->
      MLast.PaInt loc s "n"
  | [: s = P.vala (P.token ("FLOAT", "")); loc = P.gloc bp ! :] ->
      MLast.PaFlo loc s
  | [: s = P.vala (P.token ("STRING", "")); loc = P.gloc bp ! :] ->
      MLast.PaStr loc s
  | [: s = P.vala (P.token ("CHAR", "")); loc = P.gloc bp ! :] ->
      MLast.PaChr loc s
  | [: `("QUOTATION", x); loc = P.gloc bp ! :] ->
      let con = quotation_content x in
      Pcaml.handle_patt_quotation loc con
  | [: p = labeled_patt_0 :] -> p
  | [: p = patt_ident_0 :] -> p
  | [: e = alg_extension_0; loc = P.gloc bp ! :] -> MLast.PaExten loc e ]
and patt_gen_cont bp a__ strm__ =
  let x = patt_0_cont bp a__ strm__ in
  let x = patt_1_cont bp x strm__ in
  let x = patt_2_cont bp x strm__ in
  let x = patt_3_cont bp x strm__ in
  let x = patt_5_cont bp x strm__ in
  let x = patt_6_cont bp x strm__ in
  let x = patt_7_cont bp x strm__ in
  if x == a__ then x else patt_gen_cont bp x strm__
and stream_patt_comp_0 =
  parser bp
  [ [: `("", "'"); p = patt_0 ? P.error "stream_patt_comp" "'''" "[patt]";
       eo =
         P.vala
           (P.option
              (parser
                 [: `("", "when");
                    e = expr_1
                      ? P.error "implem" "'when'" "[expr level expr1]" :] ->
                   e))
         ? P.error "stream_patt_comp" "[patt]" "???";
       loc = P.gloc bp ! :] ->
      SpTrm loc p eo
  | [: `("", "?=");
       pll = P.list1sep lookahead_0 (P.token ("", "|")) False
         ? P.error "stream_patt_comp" "'?='" "[lookahead]";
       loc = P.gloc bp ! :] ->
      SpLhd loc pll
  | [: p = patt_0;
       a =
         parser
         [ [: `("", "=");
              e = expr_1
                ? P.error "stream_patt_comp" "'='" "[expr level expr1]";
              loc = P.gloc bp ! :] ->
             SpNtr loc p e
         | [: loc = P.gloc bp :] -> SpStr loc p ]
         ? P.error "stream_patt_comp" "[patt]" "'='" :] ->
      a ]
and stream_patt_0 =
  parser bp
  [ [: spc = stream_patt_comp_0;
       a =
         parser
         [ [: `("", ";");
              a =
                parser
                [ [: sp = stream_patt_kont_0 :] -> [(spc, SpoNoth) :: sp]
                | [: :] -> [(spc, SpoNoth)] ]
                ? P.error "stream_patt" "';'" "[stream_patt_kont]" :] ->
             a
         | [: :] -> [(spc, SpoNoth)] ]
         ? P.error "stream_patt" "[stream_patt_comp]" "';'" :] ->
      a
  | [: spc = stream_patt_let_0;
       sp = stream_patt_0
         ? P.error "stream_patt" "[stream_patt_let]" "[stream_patt]" :] ->
      [spc :: sp]
  | [: :] -> [] ]
and parser_case_0 =
  parser bp
    [: `("", "[<");
       sp = stream_patt_0 ? P.error "parser_case" "'[<'" "[stream_patt]";
       `("", ">]") ? P.error "parser_case" "[stream_patt]" "'>]'";
       po = P.option ipatt_0; `("", "->") ? P.error "parser_case" "" "'->'";
       e = expr_0 ? P.error "parser_case" "'->'" "[expr]" :] ->
      (sp, po, e)
and ipatt_0 =
  parser bp
  [ [: `("", "_"); loc = P.gloc bp ! :] -> MLast.PaAny loc
  | [: `("LIDENT", i); loc = P.gloc bp ! :] ->
      MLast.PaLid loc (Ploc.VaVal i) ]
and expr_0 =
  parser bp
  [ [: el = P.vala e_phony; loc = P.gloc bp !;
       a = expr_0_cont bp (MLast.ExSeq loc el) ! :] ->
      a
  | [: a = expr_1; a = expr_0_cont bp a ! :] -> a ]
and expr_0_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", ";");
       a =
         parser
         [ [: e2 = expr_0;
              let e1 = a__ in
              loc = P.gloc bp ! :] ->
             MLast.ExSeq loc (Ploc.VaVal [e1 :: get_seq e2])
         | [: :] ->
             let e1 = a__ in
             e1 ]
         ? P.error "expr" "';'" "[expr]" :] ->
      a
  | [: :] -> a__ ]
and expr_1 =
  parser bp
  [ [: `("", "parser"); po = P.option ipatt_0;
       _ = P.option (P.token ("", "|"));
       pcl = P.list1sep parser_case_0 (P.token ("", "|")) False
         ? P.error "expr" "" "[parser_case]";
       loc = P.gloc bp ! :] ->
      cparser loc (po, pcl)
  | [: `("", "function");
       (ext, attrs) = ext_attributes_0
         ? P.error "expr" "'function'" "[ext_attributes]";
       _ = P.option (P.token ("", "|"));
       l = P.vala (P.list1sep match_case_0 (P.token ("", "|")) False)
         ? P.error "expr" "" "???";
       loc = P.gloc bp ! :] ->
      expr_to_inline (MLast.ExFun loc l) ext attrs
  | [: `("", "fun");
       (ext, attrs) = ext_attributes_0
         ? P.error "expr" "'fun'" "[ext_attributes]";
       a =
         parser
         [ [: p = labeled_patt_0;
              (eo, e) = fun_def_0
                ? P.error "expr" "[labeled_patt]" "[fun_def]";
              loc = P.gloc bp ! :] ->
             expr_to_inline
               (MLast.ExFun loc (Ploc.VaVal [(p, Ploc.VaVal eo, e)])) ext
               attrs
         | [: _ = check_new_type_extended :] ->
             match Stream.npeek 2 strm__ with
             [ [("", "("); ("", "type")] -> do {
                 Stream.junk strm__;
                 Stream.junk strm__;
                 match
                   try Some (P.list1 (P.token ("LIDENT", "")) strm__) with
                   [ Stream.Failure -> None ]
                 with
                 [ Some l ->
                     match strm__ with parser
                     [ [: `("", ")");
                          (eo, e) = fun_def_0
                            ? P.error "expr" "')'" "[fun_def]";
                          loc = P.gloc bp ! :] ->
                         if eo <> None then
                           failwith "new-type cannot have when-clause"
                         else
                           let e =
                             List.fold_right
                               (fun id e ->
                                  MLast.ExFun loc
                                    (Ploc.VaVal
                                       [(MLast.PaNty loc (Ploc.VaVal id),
                                         Ploc.VaVal None, e)]))
                               l e
                           in
                           expr_to_inline e ext attrs
                     | [: :] ->
                         raise
                           (Stream.Error
                              (P.error "expr" ""
                                 "lowercase identifier or ')'")) ]
                 | _ ->
                     raise
                       (Stream.Error
                          (P.error "expr" "'('" "lowercase identifier")) ]
               }
             | _ ->
                 raise
                   (Stream.Error
                      (P.error "expr" "[check_new_type_extended]" "'('")) ]
         | [: _ = check_not_new_type_extended;
              p = patt_8
                ? P.error "expr" "[check_not_new_type_extended]"
                    "[patt level simple]";
              tyopt =
                P.option
                  (parser
                     [: `("", ":");
                        t = ctyp_6
                          ? P.error "implem" "':'" "[ctyp level apply]" :] ->
                       t);
              (eo, e) = fun_def_0 ? P.error "expr" "" "[fun_def]";
              loc = P.gloc bp ! :] ->
             let e =
               match tyopt with
               [ None -> e
               | Some ty -> MLast.ExTyc loc e ty ]
             in
             expr_to_inline
               (MLast.ExFun loc (Ploc.VaVal [(p, Ploc.VaVal eo, e)])) ext
               attrs ]
         ? P.error "expr" "[ext_attributes]"
             "[labeled_patt] or [check_new_type_extended] or [check_not_new_type_extended]" :] ->
      a
  | [: `("", "match");
       (ext, attrs) = ext_attributes_0
         ? P.error "expr" "'match'" "[ext_attributes]";
       e = expr_0 ? P.error "expr" "[ext_attributes]" "[expr]";
       `("", "with") ? P.error "expr" "[expr]" "'with'";
       a =
         parser
         [ [: `("", "parser"); po = P.option ipatt_0;
              _ = P.option (P.token ("", "|"));
              pcl = P.list1sep parser_case_0 (P.token ("", "|")) False
                ? P.error "expr" "" "[parser_case]";
              loc = P.gloc bp ! :] ->
             expr_to_inline (cparser_match loc e (po, pcl)) ext attrs
         | [: _ = P.option (P.token ("", "|"));
              l = P.vala (P.list1sep match_case_0 (P.token ("", "|")) False)
                ? P.error "expr" "" "???";
              loc = P.gloc bp ! :] ->
             expr_to_inline (MLast.ExMat loc e l) ext attrs
         | [: :] -> raise (Stream.Error "") ]
         ? P.error "expr" "'with'" "'parser' or '|' or ???" :] ->
      a
  | [: `("", "try");
       (ext, attrs) = ext_attributes_0
         ? P.error "expr" "'try'" "[ext_attributes]";
       e = expr_0 ? P.error "expr" "[ext_attributes]" "[expr]";
       `("", "with") ? P.error "expr" "[expr]" "'with'";
       _ = P.option (P.token ("", "|"));
       l = P.vala (P.list1sep match_case_0 (P.token ("", "|")) False)
         ? P.error "expr" "" "???";
       loc = P.gloc bp ! :] ->
      expr_to_inline (MLast.ExTry loc e l) ext attrs
  | [: `("", "if");
       (ext, attrs) = ext_attributes_0
         ? P.error "expr" "'if'" "[ext_attributes]";
       e1 = expr_0 ? P.error "expr" "[ext_attributes]" "[expr]";
       `("", "then") ? P.error "expr" "[expr]" "'then'";
       e2 = expr_1 ? P.error "expr" "'then'" "[expr level expr1]";
       a =
         parser
         [ [: `("", "else");
              e3 = expr_1 ? P.error "expr" "'else'" "[expr level expr1]";
              loc = P.gloc bp ! :] ->
             expr_to_inline (MLast.ExIfe loc e1 e2 e3) ext attrs
         | [: loc = P.gloc bp :] ->
             expr_to_inline
               (MLast.ExIfe loc e1 e2
                  (MLast.ExLong loc (MLast.LiUid loc (Ploc.VaVal "()"))))
               ext attrs ]
         ? P.error "expr" "[expr level expr1]" "'else'" :] ->
      a
  | [: `("", "for");
       (ext, attrs) = ext_attributes_0
         ? P.error "expr" "'for'" "[ext_attributes]";
       a =
         parser
         [ [: `("", "(");
              i = operator_rparen ? P.error "expr" "'('" "[operator_rparen]";
              `("", "=") ? P.error "expr" "[operator_rparen]" "'='";
              e1 = expr_0 ? P.error "expr" "'='" "[expr]";
              df = P.vala direction_flag_0
                ? P.error "expr" "[expr]" "[direction_flag]";
              e2 = expr_0 ? P.error "expr" "[direction_flag]" "[expr]";
              `("", "do") ? P.error "expr" "[expr]" "'do'";
              e = P.vala expr_0 ? P.error "expr" "'do'" "[expr]";
              `("", "done") ? P.error "expr" "[expr]" "'done'";
              loc = P.gloc bp ! :] ->
             let i = Ploc.VaVal i in
             let el = Pcaml.vala_map get_seq e in
             expr_to_inline (MLast.ExFor loc (MLast.PaLid loc i) e1 e2 df el)
               ext attrs
         | [: i = patt_0; `("", "=") ? P.error "expr" "[patt]" "'='";
              e1 = expr_0 ? P.error "expr" "'='" "[expr]";
              df = P.vala direction_flag_0
                ? P.error "expr" "[expr]" "[direction_flag]";
              e2 = expr_0 ? P.error "expr" "[direction_flag]" "[expr]";
              `("", "do") ? P.error "expr" "[expr]" "'do'";
              e = P.vala expr_0 ? P.error "expr" "'do'" "[expr]";
              `("", "done") ? P.error "expr" "[expr]" "'done'";
              loc = P.gloc bp ! :] ->
             let el = Pcaml.vala_map get_seq e in
             expr_to_inline (MLast.ExFor loc i e1 e2 df el) ext attrs ]
         ? P.error "expr" "[ext_attributes]" "'(' or [patt]" :] ->
      a
  | [: `("", "while");
       (ext, attrs) = ext_attributes_0
         ? P.error "expr" "'while'" "[ext_attributes]";
       e1 = expr_0 ? P.error "expr" "[ext_attributes]" "[expr]";
       `("", "do") ? P.error "expr" "[expr]" "'do'";
       e2 = P.vala expr_0 ? P.error "expr" "'do'" "[expr]";
       `("", "done") ? P.error "expr" "[expr]" "'done'";
       loc = P.gloc bp ! :] ->
      let el = Pcaml.vala_map get_seq e2 in
      expr_to_inline (MLast.ExWhi loc e1 el) ext attrs
  | [: _ = check_let_exception :] ->
      match Stream.npeek 3 strm__ with
      [ [("", "let"); ("", "exception"); ("UIDENT", id)] -> do {
          Stream.junk strm__;
          Stream.junk strm__;
          Stream.junk strm__;
          let id = Ploc.VaVal id in
          try
            match strm__ with parser
            [ [: `("", "of");
                 tyl = P.vala (P.list1 ctyp_6) ? P.error "expr" "'of'" "???";
                 alg_attrs = alg_attributes_0
                   ? P.error "expr" "???" "[alg_attributes]";
                 `("", "in") ? P.error "expr" "[alg_attributes]" "'in'";
                 x = P.orzero expr_1 expr_0 ? P.error "expr" "'in'" "[expr]";
                 loc = P.gloc bp ! :] ->
                MLast.ExLEx loc id tyl x alg_attrs
            | [: alg_attrs = alg_attributes_0;
                 `("", "in") ? P.error "expr" "[alg_attributes]" "'in'";
                 x = P.orzero expr_1 expr_0 ? P.error "expr" "'in'" "[expr]";
                 loc = P.gloc bp ! :] ->
                MLast.ExLEx loc id (Ploc.VaVal []) x alg_attrs ]
          with
          [ Stream.Failure ->
              raise
                (Stream.Error
                   (P.error "expr" "'let'" "'of' or [alg_attributes]")) ]
        }
      | _ ->
          raise
            (Stream.Error (P.error "expr" "[check_let_exception]" "'let'")) ]
  | [: _ = check_let_not_exception;
       `("", "let") ? P.error "expr" "[check_let_not_exception]" "'let'";
       a =
         parser
         [ [: `("", "module");
              (ext, attrs) = ext_attributes_0
                ? P.error "expr" "'module'" "[ext_attributes]";
              m = P.vala uidopt_0
                ? P.error "expr" "[ext_attributes]" "[uidopt]";
              mb = mod_fun_binding_0
                ? P.error "expr" "[uidopt]" "[mod_fun_binding]";
              `("", "in") ? P.error "expr" "[mod_fun_binding]" "'in'";
              e = expr_0 ? P.error "expr" "'in'" "[expr level top]";
              loc = P.gloc bp ! :] ->
             expr_to_inline (MLast.ExLmd loc m mb e) ext attrs
         | [: `("", "open");
              ovf = P.vala (P.bool (P.token ("", "!")))
                ? P.error "expr" "'open'" "???";
              (ext, attrs) = ext_attributes_0
                ? P.error "expr" "???" "[ext_attributes]";
              m = module_expr_0
                ? P.error "expr" "[ext_attributes]" "[module_expr]";
              `("", "in") ? P.error "expr" "[module_expr]" "'in'";
              e = expr_0 ? P.error "expr" "'in'" "[expr level top]";
              loc = P.gloc bp ! :] ->
             expr_to_inline (MLast.ExLop loc ovf m e) ext attrs
         | [: (ext, alg_attrs) = ext_attributes_0;
              o = P.vala (P.bool (P.token ("", "rec")))
                ? P.error "expr" "[ext_attributes]" "???";
              h = first_let_binding_0
                ? P.error "expr" "???" "[first_let_binding]";
              t = P.list0 and_let_binding_0;
              `("", "in") ? P.error "expr" "" "[and_let_binding] or 'in'";
              x = expr_0 ? P.error "expr" "'in'" "[expr level top]";
              loc = P.gloc bp ! :] ->
             let (a, b, item_attrs) = h in
             let attrs =
               merge_left_auxiliary_attrs ~{nonterm_name = "let_binding"}
                 ~{left_name = "algebraic attributes"}
                 ~{right_name = "item attributes"} alg_attrs item_attrs
             in
             let h = (a, b, attrs) in
             let l = [h :: t] in
             expr_to_inline (MLast.ExLet loc o (Ploc.VaVal l) x) ext [] ]
         ? P.error "expr" "'let'"
             "'module' or 'open' or [ext_attributes]" :] ->
      a
  | [: letop = letop;
       b = letop_binding_0 ? P.error "expr" "[letop]" "[letop_binding]";
       l = P.list0 andop_binding_0;
       `("", "in") ? P.error "expr" "" "[andop_binding] or 'in'";
       x = expr_0 ? P.error "expr" "'in'" "[expr level top]";
       loc = P.gloc bp ! :] ->
      build_letop_binder loc letop b l x
  | [: a = expr_2 :] -> a ]
and expr_2 = parser bp [: a = expr_3; a = expr_2_cont bp a ! :] -> a
and expr_2_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", ",");
       el = P.list1sep (P.orzero expr_3 expr_0) (P.token ("", ",")) False
         ? P.error "expr" "','" "[expr]";
       loc = P.gloc bp !;
       let e = a__ in
       a = expr_2_cont bp (MLast.ExTup loc (Ploc.VaVal [e :: el])) ! :] ->
      a
  | [: :] -> a__ ]
and expr_3 = parser bp [: a = expr_4; a = expr_3_cont bp a ! :] -> a
and expr_3_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", ":="); e2 = expr_1 ? P.error "expr" "':='" "[expr level expr1]";
       let e1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.ExAss loc
        (MLast.ExFle loc e1 (Ploc.VaVal (None, Ploc.VaVal "val"))) e2
  | [: `("", "<-"); e2 = expr_1 ? P.error "expr" "'<-'" "[expr level expr1]";
       let e1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.ExAss loc e1 e2
  | [: :] -> a__ ]
and expr_4 = parser bp [: a = expr_5; a = expr_4_cont bp a ! :] -> a
and expr_4_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "or");
       e2 = P.orzero expr_4 expr_0 ? P.error "expr" "'or'" "[expr]";
       let e1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "or")) e1)
        e2
  | [: `("", "||");
       e2 = P.orzero expr_4 expr_0 ? P.error "expr" "'||'" "[expr]";
       let e1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "||")) e1)
        e2
  | [: :] -> a__ ]
and expr_5 = parser bp [: a = expr_6; a = expr_5_cont bp a ! :] -> a
and expr_5_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "&");
       e2 = P.orzero expr_5 expr_0 ? P.error "expr" "'&'" "[expr]";
       let e1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "&")) e1)
        e2
  | [: `("", "&&");
       e2 = P.orzero expr_5 expr_0 ? P.error "expr" "'&&'" "[expr]";
       let e1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "&&")) e1)
        e2
  | [: :] -> a__ ]
and expr_6 = parser bp [: a = expr_7; a = expr_6_cont bp a ! :] -> a
and expr_6_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "<");
       e2 = P.orzero expr_7 expr_0 ? P.error "expr" "'<'" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_6_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "<")) e1)
              e2) ! :] ->
      a
  | [: `("", ">");
       e2 = P.orzero expr_7 expr_0 ? P.error "expr" "'>'" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_6_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal ">")) e1)
              e2) ! :] ->
      a
  | [: `("", "<=");
       e2 = P.orzero expr_7 expr_0 ? P.error "expr" "'<='" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_6_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "<=")) e1)
              e2) ! :] ->
      a
  | [: `("", ">=");
       e2 = P.orzero expr_7 expr_0 ? P.error "expr" "'>='" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_6_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal ">=")) e1)
              e2) ! :] ->
      a
  | [: `("", "=");
       e2 = P.orzero expr_7 expr_0 ? P.error "expr" "'='" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_6_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "=")) e1)
              e2) ! :] ->
      a
  | [: `("", "<>");
       e2 = P.orzero expr_7 expr_0 ? P.error "expr" "'<>'" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_6_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "<>")) e1)
              e2) ! :] ->
      a
  | [: `("", "==");
       e2 = P.orzero expr_7 expr_0 ? P.error "expr" "'=='" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_6_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "==")) e1)
              e2) ! :] ->
      a
  | [: `("", "!=");
       e2 = P.orzero expr_7 expr_0 ? P.error "expr" "'!='" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_6_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "!=")) e1)
              e2) ! :] ->
      a
  | [: op = infixop0;
       e2 = P.orzero expr_7 expr_0 ? P.error "expr" "[infixop0]" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_6_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal op)) e1)
              e2) ! :] ->
      a
  | [: :] -> a__ ]
and expr_7 = parser bp [: a = expr_8; a = expr_7_cont bp a ! :] -> a
and expr_7_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "^");
       e2 = P.orzero expr_7 expr_0 ? P.error "expr" "'^'" "[expr]";
       let e1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "^")) e1)
        e2
  | [: `("", "@");
       e2 = P.orzero expr_7 expr_0 ? P.error "expr" "'@'" "[expr]";
       let e1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "@")) e1)
        e2
  | [: op = infixop1;
       e2 = P.orzero expr_7 expr_0 ? P.error "expr" "[infixop1]" "[expr]";
       let e1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal op)) e1)
        e2
  | [: :] -> a__ ]
and expr_8 = parser bp [: a = expr_9; a = expr_8_cont bp a ! :] -> a
and expr_8_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "[@");
       attr = P.vala attribute_body_0
         ? P.error "expr" "'[@'" "[attribute_body]";
       `("", "]") ? P.error "expr" "[attribute_body]" "']'";
       loc = P.gloc bp !;
       let e1 = a__ in
       a = expr_8_cont bp (MLast.ExAtt loc e1 attr) ! :] ->
      a
  | [: :] -> a__ ]
and expr_9 = parser bp [: a = expr_10; a = expr_9_cont bp a ! :] -> a
and expr_9_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "::");
       e2 = P.orzero expr_9 expr_0 ? P.error "expr" "'::'" "[expr]";
       let e1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc
        (MLast.ExApp loc
           (MLast.ExLong loc (MLast.LiUid loc (Ploc.VaVal "::"))) e1)
        e2
  | [: :] -> a__ ]
and expr_10 = parser bp [: a = expr_11; a = expr_10_cont bp a ! :] -> a
and expr_10_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "+");
       e2 = P.orzero expr_11 expr_0 ? P.error "expr" "'+'" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_10_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "+")) e1)
              e2) ! :] ->
      a
  | [: `("", "-");
       e2 = P.orzero expr_11 expr_0 ? P.error "expr" "'-'" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_10_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "-")) e1)
              e2) ! :] ->
      a
  | [: op = infixop2;
       e2 = P.orzero expr_11 expr_0 ? P.error "expr" "[infixop2]" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_10_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal op)) e1)
              e2) ! :] ->
      a
  | [: :] -> a__ ]
and expr_11 = parser bp [: a = expr_12; a = expr_11_cont bp a ! :] -> a
and expr_11_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "*");
       e2 = P.orzero expr_12 expr_0 ? P.error "expr" "'*'" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_11_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "*")) e1)
              e2) ! :] ->
      a
  | [: `("", "/");
       e2 = P.orzero expr_12 expr_0 ? P.error "expr" "'/'" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_11_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "/")) e1)
              e2) ! :] ->
      a
  | [: `("", "%");
       e2 = P.orzero expr_12 expr_0 ? P.error "expr" "'%'" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_11_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "%")) e1)
              e2) ! :] ->
      a
  | [: `("", "land");
       e2 = P.orzero expr_12 expr_0 ? P.error "expr" "'land'" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_11_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "land")) e1)
              e2) ! :] ->
      a
  | [: `("", "lor");
       e2 = P.orzero expr_12 expr_0 ? P.error "expr" "'lor'" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_11_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "lor")) e1)
              e2) ! :] ->
      a
  | [: `("", "lxor");
       e2 = P.orzero expr_12 expr_0 ? P.error "expr" "'lxor'" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_11_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "lxor")) e1)
              e2) ! :] ->
      a
  | [: `("", "mod");
       e2 = P.orzero expr_12 expr_0 ? P.error "expr" "'mod'" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_11_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "mod")) e1)
              e2) ! :] ->
      a
  | [: op = infixop3;
       e2 = P.orzero expr_12 expr_0 ? P.error "expr" "[infixop3]" "[expr]";
       loc = P.gloc bp !;
       let e1 = a__ in
       a =
         expr_11_cont bp
           (MLast.ExApp loc
              (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal op)) e1)
              e2) ! :] ->
      a
  | [: :] -> a__ ]
and expr_12 = parser bp [: a = expr_13; a = expr_12_cont bp a ! :] -> a
and expr_12_cont bp a__ strm__ =
  match strm__ with parser
  [ [: `("", "**");
       e2 = P.orzero expr_12 expr_0 ? P.error "expr" "'**'" "[expr]";
       let e1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "**")) e1)
        e2
  | [: `("", "asr");
       e2 = P.orzero expr_12 expr_0 ? P.error "expr" "'asr'" "[expr]";
       let e1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc
        (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "asr")) e1) e2
  | [: `("", "lsl");
       e2 = P.orzero expr_12 expr_0 ? P.error "expr" "'lsl'" "[expr]";
       let e1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc
        (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "lsl")) e1) e2
  | [: `("", "lsr");
       e2 = P.orzero expr_12 expr_0 ? P.error "expr" "'lsr'" "[expr]";
       let e1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc
        (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "lsr")) e1) e2
  | [: op = infixop4;
       e2 = P.orzero expr_12 expr_0 ? P.error "expr" "[infixop4]" "[expr]";
       let e1 = a__ in
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal op)) e1)
        e2
  | [: :] -> a__ ]
and expr_13 =
  parser bp
  [ [: `("", "-");
       e = P.orzero expr_13 expr_0 ? P.error "expr" "'-'" "[expr]";
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "-")) e
  | [: `("", "-.");
       e = P.orzero expr_13 expr_0 ? P.error "expr" "'-.'" "[expr]";
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "-.")) e
  | [: `("", "+");
       e = P.orzero expr_13 expr_0 ? P.error "expr" "'+'" "[expr]";
       loc = P.gloc bp ! :] ->
      match e with
      [ MLast.ExInt _ (Ploc.VaVal _) "" -> e
      | _ -> MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "~+")) e ]
  | [: `("", "+.");
       e = P.orzero expr_13 expr_0 ? P.error "expr" "'+.'" "[expr]";
       loc = P.gloc bp ! :] ->
      match e with
      [ MLast.ExFlo _ (Ploc.VaVal _) -> e
      | _ -> MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "~+.")) e ]
  | [: a = expr_14 :] -> a ]
and expr_14 =
  parser bp
  [ [: `("", "assert");
       (ext, attrs) = ext_attributes_0
         ? P.error "expr" "'assert'" "[ext_attributes]";
       e = P.orzero expr_14 expr_0
         ? P.error "expr" "[ext_attributes]" "[expr]";
       loc = P.gloc bp !;
       a =
         expr_14_cont bp
           (expr_to_inline (MLast.ExAsr loc e) ext attrs) ! :] ->
      a
  | [: `("", "lazy");
       (ext, attrs) = ext_attributes_0
         ? P.error "expr" "'lazy'" "[ext_attributes]";
       e = P.orzero expr_14 expr_0
         ? P.error "expr" "[ext_attributes]" "[expr]";
       loc = P.gloc bp !;
       a =
         expr_14_cont bp
           (expr_to_inline (MLast.ExLaz loc e) ext attrs) ! :] ->
      a
  | [: a = expr_15; a = expr_14_cont bp a ! :] -> a ]
and expr_14_cont bp a__ strm__ =
  match try Some (expr_15 strm__) with [ Stream.Failure -> None ] with
  [ Some e2 ->
      let loc = P.gloc bp strm__ in
      let e1 = a__ in
      expr_14_cont bp
        (let (e1, e2) =
           if is_expr_constr_call e1 then
             match e1 with
             [ MLast.ExApp _ e11 e12 -> (e11, MLast.ExApp loc e12 e2)
             | _ -> (e1, e2) ]
           else (e1, e2)
         in
         match constr_expr_arity loc e1 with
         [ 1 -> MLast.ExApp loc e1 e2
         | _ ->
             match e2 with
             [ MLast.ExTup _ (Ploc.VaVal el) ->
                 List.fold_left (fun e1 e2 -> MLast.ExApp loc e1 e2) e1 el
             | _ -> MLast.ExApp loc e1 e2 ] ])
        strm__
  | _ -> a__ ]
and expr_15 =
  parser bp
  [ [: i = P.vala (P.token ("TILDEIDENTCOLON", ""));
       e = P.orzero expr_15 expr_0
         ? P.error "expr" "TILDEIDENTCOLON" "[expr]";
       loc = P.gloc bp ! :] ->
      MLast.ExLab loc (Ploc.VaVal [(MLast.PaLid loc i, Ploc.VaVal (Some e))])
  | [: i = P.vala (P.token ("TILDEIDENT", "")); loc = P.gloc bp ! :] ->
      MLast.ExLab loc (Ploc.VaVal [(MLast.PaLid loc i, Ploc.VaVal None)])
  | [: i = P.vala (P.token ("QUESTIONIDENTCOLON", ""));
       e = P.orzero expr_15 expr_0
         ? P.error "expr" "QUESTIONIDENTCOLON" "[expr]";
       loc = P.gloc bp ! :] ->
      MLast.ExOlb loc (MLast.PaLid loc i) (Ploc.VaVal (Some e))
  | [: i = P.vala (P.token ("QUESTIONIDENT", "")); loc = P.gloc bp ! :] ->
      MLast.ExOlb loc (MLast.PaLid loc i) (Ploc.VaVal None)
  | [: a = expr_16 :] -> a ]
and expr_16 = parser bp [: a = expr_17; a = expr_16_cont bp a ! :] -> a
and expr_16_cont bp a__ strm__ =
  match Stream.npeek 2 strm__ with
  [ [("", "#"); ("LIDENT", lab)] -> do {
      Stream.junk strm__;
      Stream.junk strm__;
      let lab = Ploc.VaVal lab in
      try
        let loc = P.gloc bp strm__ in
        let e = a__ in
        expr_16_cont bp (MLast.ExSnd loc e lab) strm__
      with
      [ Stream.Failure -> raise (Stream.Error (P.error "expr" "'#'" "???")) ]
    }
  | _ ->
      match strm__ with parser
      [ [: `("", ".");
           a =
             parser
             [ [: `("", "(");
                  a =
                    parser
                    [ [: op = operator_rparen; loc = P.gloc bp !;
                         let e1 = a__ in
                         a =
                           expr_16_cont bp
                             (if op = "::" then
                                Ploc.raise loc
                                  (Failure
                                     ".(::) (dot paren colon colon paren) cannot appear except after longident")
                              else
                                MLast.ExFle loc e1
                                  (Ploc.VaVal (None, Ploc.VaVal op))) ! :] ->
                        a
                    | [: e2 = expr_0;
                         `("", ")") ? P.error "expr" "[expr]" "')'";
                         loc = P.gloc bp !;
                         let e1 = a__ in
                         a =
                           expr_16_cont bp
                             (if expr_last_is_uid e1 then
                                failwith
                                  "internal error in original-syntax parser at dot-lparen"
                              else
                                MLast.ExAre loc (Ploc.VaVal ".") e1
                                  (Ploc.VaVal [e2])) ! :] ->
                        a ]
                    ? P.error "expr" "'('" "[operator_rparen] or [expr]" :] ->
                 a
             | [: `("", "["); e2 = expr_0 ? P.error "expr" "'['" "[expr]";
                  `("", "]") ? P.error "expr" "[expr]" "']'";
                  loc = P.gloc bp !;
                  let e1 = a__ in
                  a =
                    expr_16_cont bp
                      (MLast.ExSte loc (Ploc.VaVal ".") e1
                         (Ploc.VaVal [e2])) ! :] ->
                 a
             | [: `("", "{");
                  el = P.list1sep expr_10 (P.token ("", ",")) False
                    ? P.error "expr" "'{'" "[expr level +]";
                  `("", "}") ? P.error "expr" "" "[expr level +] or '}'";
                  loc = P.gloc bp !;
                  let e1 = a__ in
                  a =
                    expr_16_cont bp
                      (MLast.ExBae loc (Ploc.VaVal ".") e1
                         (Ploc.VaVal el)) ! :] ->
                 a
             | [: lili = P.vala longident_lident_0; loc = P.gloc bp !;
                  let e1 = a__ in
                  a = expr_16_cont bp (MLast.ExFle loc e1 lili) ! :] ->
                 a ]
             ? P.error "expr" "'.'"
                 "'(' or '[' or '{' or [longident_lident]" :] ->
          a
      | [: op = hashop;
           e2 = P.orzero expr_17 expr_0 ? P.error "expr" "[hashop]" "[expr]";
           loc = P.gloc bp !;
           let e = a__ in
           a =
             expr_16_cont bp
               (MLast.ExApp loc
                  (MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal op)) e)
                  e2) ! :] ->
          a
      | [: op = P.vala dotop;
           a =
             parser
             [ [: `("", "(");
                  el = P.list1sep expr_10 (P.token ("", ";")) False
                    ? P.error "expr" "'('" "[expr level +]";
                  `("", ")") ? P.error "expr" "" "[expr level +] or ')'";
                  loc = P.gloc bp !;
                  let e1 = a__ in
                  a =
                    expr_16_cont bp
                      (MLast.ExAre loc op e1 (Ploc.VaVal el)) ! :] ->
                 a
             | [: `("", "[");
                  el = P.list1sep expr_10 (P.token ("", ";")) False
                    ? P.error "expr" "'['" "[expr level +]";
                  `("", "]") ? P.error "expr" "" "[expr level +] or ']'";
                  loc = P.gloc bp !;
                  let e1 = a__ in
                  a =
                    expr_16_cont bp
                      (MLast.ExSte loc op e1 (Ploc.VaVal el)) ! :] ->
                 a
             | [: `("", "{");
                  el = P.list1sep expr_10 (P.token ("", ";")) False
                    ? P.error "expr" "'{'" "[expr level +]";
                  `("", "}") ? P.error "expr" "" "[expr level +] or '}'";
                  loc = P.gloc bp !;
                  let e1 = a__ in
                  a =
                    expr_16_cont bp
                      (MLast.ExBae loc op e1 (Ploc.VaVal el)) ! :] ->
                 a ]
             ? P.error "expr" "[dotop]" "'(' or '[' or '{'" :] ->
          a
      | [: :] -> a__ ] ]
and expr_17 =
  parser bp
  [ [: `("", "!");
       e = P.orzero expr_17 expr_0 ? P.error "expr" "'!'" "[expr]";
       loc = P.gloc bp ! :] ->
      MLast.ExFle loc e (Ploc.VaVal (None, Ploc.VaVal "val"))
  | [: `("", "~-");
       e = P.orzero expr_17 expr_0 ? P.error "expr" "'~-'" "[expr]";
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "~-")) e
  | [: `("", "~-.");
       e = P.orzero expr_17 expr_0 ? P.error "expr" "'~-.'" "[expr]";
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal "~-.")) e
  | [: f = prefixop;
       e = P.orzero expr_17 expr_0 ? P.error "expr" "[prefixop]" "[expr]";
       loc = P.gloc bp ! :] ->
      MLast.ExApp loc (MLast.ExLid loc (Ploc.VaVal f)) e
  | [: a = expr_18 :] -> a ]
and expr_18 =
  parser bp
  [ [: `("", "[<");
       a =
         parser
         [ [: `("", ">]"); loc = P.gloc bp ! :] -> cstream loc []
         | [: sel = stream_expr_comp_list_0;
              `("", ">]") ? P.error "expr" "[stream_expr_comp_list]" "'>]'";
              loc = P.gloc bp ! :] ->
             cstream loc sel ]
         ? P.error "expr" "'[<'" "'>]' or [stream_expr_comp_list]" :] ->
      a
  | [: `("", "`"); s = P.vala ident_0 ? P.error "expr" "'`'" "[ident]";
       loc = P.gloc bp ! :] ->
      MLast.ExVrn loc s
  | [: `("", "{<");
       a =
         parser
         [ [: `("", ">}"); loc = P.gloc bp ! :] ->
             MLast.ExOvr loc (Ploc.VaVal [])
         | [: fel = P.vala field_expr_list_0;
              `("", ">}") ? P.error "expr" "[field_expr_list]" "'>}'";
              loc = P.gloc bp ! :] ->
             MLast.ExOvr loc fel ]
         ? P.error "expr" "'{<'" "'>}' or [field_expr_list]" :] ->
      a
  | [: `("", "new");
       (ext, attrs) = ext_attributes_0
         ? P.error "expr" "'new'" "[ext_attributes]";
       cli = P.vala longident_lident_0
         ? P.error "expr" "[ext_attributes]" "[longident_lident]";
       loc = P.gloc bp ! :] ->
      expr_to_inline (MLast.ExNew loc cli) ext attrs
  | [: `("", "object");
       (ext, attrs) = ext_attributes_0
         ? P.error "expr" "'object'" "[ext_attributes]";
       cspo = P.vala (P.option class_self_patt_0)
         ? P.error "expr" "[ext_attributes]" "???";
       cf = P.vala class_structure_0
         ? P.error "expr" "???" "[class_structure]";
       `("", "end") ? P.error "expr" "[class_structure]" "'end'";
       loc = P.gloc bp ! :] ->
      expr_to_inline (MLast.ExObj loc cspo cf) ext attrs
  | [: `("UIDENT", "True"); loc = P.gloc bp ! :] ->
      MLast.ExLong loc (MLast.LiUid loc (Ploc.VaVal "True_"))
  | [: `("UIDENT", "False"); loc = P.gloc bp ! :] ->
      MLast.ExLong loc (MLast.LiUid loc (Ploc.VaVal "False_"))
  | [: `("", "true"); loc = P.gloc bp ! :] ->
      MLast.ExLong loc (MLast.LiUid loc (Ploc.VaVal "True"))
  | [: `("", "false"); loc = P.gloc bp ! :] ->
      MLast.ExLong loc (MLast.LiUid loc (Ploc.VaVal "False"))
  | [: `("", "[");
       a =
         parser
         [ [: `("", "]"); loc = P.gloc bp ! :] ->
             MLast.ExLong loc (MLast.LiUid loc (Ploc.VaVal "[]"))
         | [: el = expr1_semi_list_0;
              `("", "]") ? P.error "expr" "[expr1_semi_list]" "']'";
              loc = P.gloc bp ! :] ->
             mklistexp loc None el ]
         ? P.error "expr" "'['" "']' or [expr1_semi_list]" :] ->
      a
  | [: `("", "[|");
       a =
         parser
         [ [: `("", "|]"); loc = P.gloc bp ! :] ->
             MLast.ExArr loc (Ploc.VaVal [])
         | [: el = P.vala expr1_semi_list_0;
              `("", "|]") ? P.error "expr" "[expr1_semi_list]" "'|]'";
              loc = P.gloc bp ! :] ->
             MLast.ExArr loc el ]
         ? P.error "expr" "'[|'" "'|]' or [expr1_semi_list]" :] ->
      a
  | [: `("", "{");
       a =
         parser
         [ [: _ = test_label_eq;
              lel = P.vala lbl_expr_list_0
                ? P.error "expr" "[test_label_eq]" "[lbl_expr_list]";
              `("", "}") ? P.error "expr" "[lbl_expr_list]" "'}'";
              loc = P.gloc bp ! :] ->
             MLast.ExRec loc lel None
         | [: e = expr_14;
              `("", "with") ? P.error "expr" "[expr level apply]" "'with'";
              lel = P.vala lbl_expr_list_0
                ? P.error "expr" "'with'" "[lbl_expr_list]";
              `("", "}") ? P.error "expr" "[lbl_expr_list]" "'}'";
              loc = P.gloc bp ! :] ->
             MLast.ExRec loc lel (Some e) ]
         ? P.error "expr" "'{'" "[test_label_eq] or [expr level apply]" :] ->
      a
  | [: `("", "(");
       a =
         parser
         [ [: `("", ")"); loc = P.gloc bp ! :] ->
             MLast.ExLong loc (MLast.LiUid loc (Ploc.VaVal "()"))
         | [: `("", "module");
              me = module_expr_0 ? P.error "expr" "'module'" "[module_expr]";
              a =
                parser
                [ [: `("", ":");
                     mt = module_type_0
                       ? P.error "expr" "':'" "[module_type]";
                     `("", ")") ? P.error "expr" "[module_type]" "')'";
                     loc = P.gloc bp ! :] ->
                    MLast.ExPck loc me (Some mt)
                | [: `("", ")"); loc = P.gloc bp ! :] ->
                    MLast.ExPck loc me None ]
                ? P.error "expr" "[module_expr]" "':' or ')'" :] ->
             a
         | [: op = operator_rparen; loc = P.gloc bp ! :] ->
             if op = "::" then
               MLast.ExLong loc (MLast.LiUid loc (Ploc.VaVal op))
             else MLast.ExLid loc (Ploc.VaVal op)
         | [: el = P.vala e_phony;
              `("", ")") ? P.error "expr" "[e_phony]" "')'";
              loc = P.gloc bp ! :] ->
             MLast.ExTup loc el
         | [: e = expr_0;
              a =
                parser
                [ [: `("", ":>"); t = ctyp_0 ? P.error "expr" "':>'" "[ctyp]";
                     `("", ")") ? P.error "expr" "[ctyp]" "')'";
                     loc = P.gloc bp ! :] ->
                    MLast.ExCoe loc e None t
                | [: `("", ":"); t = ctyp_0 ? P.error "expr" "':'" "[ctyp]";
                     a =
                       parser
                       [ [: `("", ":>");
                            t2 = ctyp_0 ? P.error "expr" "':>'" "[ctyp]";
                            `("", ")") ? P.error "expr" "[ctyp]" "')'";
                            loc = P.gloc bp ! :] ->
                           MLast.ExCoe loc e (Some t) t2
                       | [: `("", ")"); loc = P.gloc bp ! :] ->
                           MLast.ExTyc loc e t ]
                       ? P.error "expr" "[ctyp]" "':>' or ')'" :] ->
                    a
                | [: `("", ")"); loc = P.gloc bp ! :] -> concat_comm loc e ]
                ? P.error "expr" "[expr]" "':>' or ':' or ')'" :] ->
             a ]
         ? P.error "expr" "'('"
             "')' or 'module' or [operator_rparen] or [e_phony] or [expr]" :] ->
      a
  | [: `("", "begin");
       (ext, attrs) = ext_attributes_0
         ? P.error "expr" "'begin'" "[ext_attributes]";
       a =
         parser
         [ [: `("", "end"); loc = P.gloc bp ! :] ->
             expr_to_inline
               (MLast.ExLong loc (MLast.LiUid loc (Ploc.VaVal "()"))) ext
               attrs
         | [: e = expr_0; `("", "end") ? P.error "expr" "[expr]" "'end'";
              loc = P.gloc bp ! :] ->
             expr_to_inline (concat_comm loc e) ext attrs ]
         ? P.error "expr" "[ext_attributes]" "'end' or [expr]" :] ->
      a
  | [: s = P.vala (P.token ("INT", "")); loc = P.gloc bp ! :] ->
      MLast.ExInt loc s ""
  | [: s = P.vala (P.token ("INT_l", "")); loc = P.gloc bp ! :] ->
      MLast.ExInt loc s "l"
  | [: s = P.vala (P.token ("INT_L", "")); loc = P.gloc bp ! :] ->
      MLast.ExInt loc s "L"
  | [: s = P.vala (P.token ("INT_n", "")); loc = P.gloc bp ! :] ->
      MLast.ExInt loc s "n"
  | [: s = P.vala (P.token ("FLOAT", "")); loc = P.gloc bp ! :] ->
      MLast.ExFlo loc s
  | [: s = P.vala (P.token ("STRING", "")); loc = P.gloc bp ! :] ->
      MLast.ExStr loc s
  | [: c = P.vala (P.token ("CHAR", "")); loc = P.gloc bp ! :] ->
      MLast.ExChr loc c
  | [: i = P.vala (P.token ("LIDENT", "")); loc = P.gloc bp ! :] ->
      MLast.ExLid loc i
  | [: `("QUOTATION", x); loc = P.gloc bp ! :] ->
      let con = quotation_content x in
      Pcaml.handle_expr_quotation loc con
  | [: e = alg_extension_0; loc = P.gloc bp ! :] -> MLast.ExExten loc e
  | [: i = expr_longident_0 :] -> i ]
and expr_gen_cont bp a__ strm__ =
  let x = expr_0_cont bp a__ strm__ in
  let x = expr_2_cont bp x strm__ in
  let x = expr_3_cont bp x strm__ in
  let x = expr_4_cont bp x strm__ in
  let x = expr_5_cont bp x strm__ in
  let x = expr_6_cont bp x strm__ in
  let x = expr_7_cont bp x strm__ in
  let x = expr_8_cont bp x strm__ in
  let x = expr_9_cont bp x strm__ in
  let x = expr_10_cont bp x strm__ in
  let x = expr_11_cont bp x strm__ in
  let x = expr_12_cont bp x strm__ in
  let x = expr_14_cont bp x strm__ in
  let x = expr_16_cont bp x strm__ in
  if x == a__ then x else expr_gen_cont bp x strm__
and str_item_semi_0 =
  parser bp
    [: si = str_item_0; _ = P.option (P.token ("", ";;"));
       loc = P.gloc bp ! :] ->
      (si, loc)
;

let lexer = Lazy.force P.lexer in
List.iter (fun kw -> lexer.Plexing.tok_using ("", kw))
  ["!"; "!+"; "!-"; "!="; "#"; "%"; "&"; "&&"; "'"; "("; ")"; "*"; "**"; "+";
   "+!"; "+."; "+="; ","; "-"; "-!"; "-."; "->"; "."; ".."; "/"; ":"; "::";
   ":="; ":>"; ";"; ";;"; "<"; "<-"; "<="; "<>"; "="; "=="; ">"; ">="; ">]";
   ">}"; "?"; "?!"; "?="; "??"; "@"; "["; "[%"; "[%%"; "[<"; "[@"; "[@@";
   "[@@@"; "[|"; "]"; "^"; "_"; "`"; "and"; "as"; "asr"; "assert"; "begin";
   "class"; "constraint"; "do"; "done"; "downto"; "else"; "end"; "exception";
   "external"; "false"; "for"; "fun"; "function"; "functor"; "if"; "in";
   "include"; "inherit"; "initializer"; "land"; "lazy"; "let"; "lor"; "lsl";
   "lsr"; "lxor"; "match"; "method"; "mod"; "module"; "mutable"; "new";
   "nonrec"; "object"; "of"; "open"; "or"; "parser"; "private"; "rec"; "sig";
   "struct"; "then"; "to"; "true"; "try"; "type"; "val"; "virtual"; "when";
   "while"; "with"; "{"; "{<"; "|"; "|]"; "||"; "}"; "~"; "~-"; "~-."];
(****************************************)

value interf_p =
  Grammar.Entry.of_parser Pcaml.gram "interf" interf_0
;

value implem_p =
  Grammar.Entry.of_parser Pcaml.gram "implem" implem_0
;

value top_phrase_p =
  Grammar.Entry.of_parser Pcaml.gram "top_phrase" top_phrase_0
;

value use_file_p =
  Grammar.Entry.of_parser Pcaml.gram "use_file" use_file_0
;

EXTEND
  interf:
    [ [ x = interf_p -> x ] ]
  ;
  implem:
    [ [ x = implem_p -> x ] ]
  ;
  top_phrase:
    [ [ x = top_phrase_p -> x ] ]
  ;
  use_file:
    [ [ x = use_file_p -> x ] ]
  ;
END;
